/*
 * NFADeterminizer.cpp
 *
 *  Created on: 16. 1. 2014
 *	  Author: Jan Travnicek
 */

#include "common/RHDPDACommon.h"

#include <automaton/PDA/RealTimeHeightDeterministicNPDA.h>
#include "automaton/common/State.h"
#include "alphabet/Symbol.h"
#include "alphabet/LabeledSymbol.h"
#include "label/Label.h"
#include <set>
#include <global/GlobalData.h>

namespace automaton {

namespace determinize {

void addRetTransition(const automaton::State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const alphabet::Symbol& dvpdaSymbol, const automaton::State& to, automaton::RealTimeHeightDeterministicDPDA& deterministic) {
	deterministic.addState(from);
	deterministic.addState(to);
	deterministic.addStackSymbol(dvpdaSymbol);

	deterministic.addReturnTransition(from, input, dvpdaSymbol, to);
}

void retInitial(const automaton::State& state, const alphabet::Symbol& pdaSymbol, const std::variant<string::Epsilon, alphabet::Symbol>& input, const automaton::RealTimeHeightDeterministicNPDA& nondeterministic, automaton::RealTimeHeightDeterministicDPDA& deterministic, std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, State>& rubbishReturnTransitions) {
	std::set<std::pair<label::Label, label::Label>> S = unpackFromStateLabel(state.getName());

	std::set<std::pair<label::Label, label::Label>> S1;
	for(const auto& entry : S) {
		const label::Label& q = entry.first;
		const label::Label& q2 = entry.second;

		for(const auto& transition : nondeterministic.getReturnTransitions()) {
			if(q2 != std::get<0>(transition.first).getName()) continue;
			if(input != std::get<1>(transition.first)) continue;
			if(nondeterministic.getBottomOfTheStackSymbol() != std::get<2>(transition.first)) continue;

			for(const auto& to : transition.second) {
				const label::Label& q1 = to.getName();
				S1.insert(std::make_pair(q, q1));
			}
		}
	}

	automaton::State to(packToStateLabel(std::move(S1)));
	if(S1.empty()) {
		std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol> key(state, input, pdaSymbol);
		rubbishReturnTransitions.insert(std::make_pair(key, to));
	} else {
		addRetTransition(state, input, pdaSymbol, to, deterministic);
	}
}

void ret(const automaton::State& state, const alphabet::Symbol& pdaSymbol, const std::variant<string::Epsilon, alphabet::Symbol>& input, const automaton::RealTimeHeightDeterministicNPDA& nondeterministic, automaton::RealTimeHeightDeterministicDPDA& deterministic, std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, State>& rubbishReturnTransitions) {
	std::set<std::pair<label::Label, label::Label>> S = unpackFromStateLabel(state.getName());
	std::pair<std::set<std::pair<label::Label, label::Label>>, std::variant<string::Epsilon, alphabet::Symbol>> pdaSymbolUnpack = unpackFromDRHDPDAStackSymbol(pdaSymbol);
	const std::set<std::pair<label::Label, label::Label>>& S1 = pdaSymbolUnpack.first;

	std::set<std::pair<label::Label, label::Label>> update;

	for(const auto& transition : nondeterministic.getCallTransitions()) {
		if(pdaSymbolUnpack.second != std::get<1>(transition.first)) continue;

		const label::Label& q = std::get<0>(transition.first).getName();
		for(const auto& to : transition.second) {
			const label::Label q1 = to.first.getName();
			const alphabet::Symbol& gamma = to.second;

			for(const auto& entry : S) {
				if(q1 != entry.first) continue;
				const label::Label& q2 = entry.second;

				for(const auto& transition2 : nondeterministic.getReturnTransitions()) {
					if(q2 != std::get<0>(transition2.first).getName()) continue;
					if(input != std::get<1>(transition2.first)) continue;
					if(gamma != std::get<2>(transition2.first)) continue;

					for(const auto& to2 : transition2.second) {
						const label::Label& qI = to2.getName();

						update.insert(std::make_pair(q, qI));
					}
				}
			}
		}
	}

	std::set<std::pair<label::Label, label::Label>> S2;
	for(const auto& entry : S1) {
		const label::Label& q = entry.first;
		const label::Label& q3 = entry.second;
		for(const auto& entry2 : update) {
			if(q3 != entry2.first) continue;

			const label::Label& qI = entry2.second;

			S2.insert(std::make_pair(q, qI));
		}
	}

	automaton::State to(packToStateLabel(std::move(S2)));
	if(S2.empty()) {
		std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol> key(state, input, pdaSymbol);
		rubbishReturnTransitions.insert(std::make_pair(key, to));
	} else {
		addRetTransition(state, input, pdaSymbol, to, deterministic);
	}
}

void addCallTransition(const automaton::State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const automaton::State& to, const alphabet::Symbol& dvpdaSymbol, automaton::RealTimeHeightDeterministicDPDA& deterministic) {
	deterministic.addState(from);
	deterministic.addState(to);
	deterministic.addStackSymbol(dvpdaSymbol);

	deterministic.addCallTransition(from, input, to, dvpdaSymbol);
}

void call(const automaton::State& state, const std::variant<string::Epsilon, alphabet::Symbol>& input, const automaton::RealTimeHeightDeterministicNPDA& nondeterministic, automaton::RealTimeHeightDeterministicDPDA& deterministic, std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, std::pair<State, alphabet::Symbol> >& rubbishCallTransitions) {
	std::set<std::pair<label::Label, label::Label>> S = unpackFromStateLabel(state.getName());

	std::set<label::Label> R = retrieveDSubSet(S);
	std::set<label::Label> R1;

	for(const label::Label& q : R) {
		for(const auto& transition : nondeterministic.getCallTransitions()) {
			if(q != transition.first.first.getName()) continue;
			if(input != transition.first.second) continue;

			for(const auto& to : transition.second) {
				const label::Label& q1 = to.first.getName();

				R1.insert(q1);
			}
		}
	}

	automaton::State to(packToStateLabel(createIdentity(std::move(R1))));
	alphabet::Symbol push(alphabet::LabeledSymbol(packToStackSymbolLabel(std::make_pair(std::move(S), input))));
	if(R1.empty()) {
		rubbishCallTransitions.insert(std::make_pair(std::make_pair(state, input), std::make_pair(to, push)));
	} else {
		addCallTransition(state, input, to, push, deterministic);
	}
}

void addLocalTransition(const automaton::State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const automaton::State& to, automaton::RealTimeHeightDeterministicDPDA& deterministic) {
	deterministic.addState(from);
	deterministic.addState(to);

	deterministic.addLocalTransition(from, input, to);
}

void local(const automaton::State& state, const std::variant<string::Epsilon, alphabet::Symbol>& input, const automaton::RealTimeHeightDeterministicNPDA& nondeterministic, automaton::RealTimeHeightDeterministicDPDA& deterministic, std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, State>& rubbishLocalTransitions ) {
	std::set<std::pair<label::Label, label::Label>> S = unpackFromStateLabel(state.getName());
	std::set<std::pair<label::Label, label::Label>> S1;

	for(const auto& entry : S) {
		label::Label q = entry.first;
		label::Label q2 = entry.second;
		for(const auto& transition : nondeterministic.getLocalTransitions()) {
			if(q2 != transition.first.first.getName()) continue;
			if(input != transition.first.second) continue;

			for(const auto& to : transition.second) {
				label::Label q1 = to.getName();

				S1.insert(std::make_pair(q, q1));
			}
		}
	}

	automaton::State to(packToStateLabel(std::move(S1)));
	if(S1.empty()) {
		rubbishLocalTransitions.insert(std::make_pair(std::make_pair(state, input), to));
	} else {
		addLocalTransition(state, input, to, deterministic);
	}
}

std::tuple<std::set<std::variant<string::Epsilon, alphabet::Symbol>>, std::set<std::variant<string::Epsilon, alphabet::Symbol>>, std::set<std::variant<string::Epsilon, alphabet::Symbol>>> getLocalCallRetPartitioning(const automaton::RealTimeHeightDeterministicNPDA& n, const automaton::State& state) {
	std::set<std::variant<string::Epsilon, alphabet::Symbol>> local;
	std::set<std::variant<string::Epsilon, alphabet::Symbol>> call;
	std::set<std::variant<string::Epsilon, alphabet::Symbol>> ret;

	const std::set<label::Label> dSubSet = retrieveDSubSet(unpackFromStateLabel(state.getName()));

	for(const auto& transition : n.getLocalTransitions()) {
		if(dSubSet.count(transition.first.first.getName()))
			local.insert(transition.first.second);
	}

	for(const auto& transition : n.getCallTransitions()) {
		if(dSubSet.count(transition.first.first.getName()))
			call.insert(transition.first.second);
	}

	for(const auto& transition : n.getReturnTransitions()) {
		if(dSubSet.count(std::get<0>(transition.first).getName()))
			ret.insert(std::get<1>(transition.first));
	}

	return std::make_tuple(std::move(local), std::move(call), std::move(ret));
}

automaton::RealTimeHeightDeterministicDPDA Determinize::determinize(const automaton::RealTimeHeightDeterministicNPDA& n) {
	label::Label initialLabel = packToStateLabel(createIdentity(retrieveLabels(n.getInitialStates())));

	automaton::RealTimeHeightDeterministicDPDA d(automaton::State(initialLabel), n.getBottomOfTheStackSymbol());
	d.setInputAlphabet(n.getInputAlphabet());

	std::set<automaton::State> rubbishStates = {automaton::State(packToStateLabel({}))};
	std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, State> rubbishReturnTransitions;
	std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, std::pair<State, alphabet::Symbol> > rubbishCallTransitions;
	std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, State> rubbishLocalTransitions;

	for(;;) {
		std::set<std::pair<automaton::State, alphabet::Symbol>> stateSymbols = existsDirtyStateSymbol(d, rubbishStates, rubbishCallTransitions, rubbishReturnTransitions, n);
		std::set<automaton::State> states = existsDirtyState(d, rubbishStates, rubbishCallTransitions, rubbishLocalTransitions, n);

		if(stateSymbols.empty() && states.empty()) break;

		for(const auto& stateSymbol : stateSymbols) {
			if ( common::GlobalData::verbose )
				std::clog << "Dirty state symbol: " << stateSymbol << std::endl;

			std::tuple<std::set<std::variant<string::Epsilon, alphabet::Symbol>>, std::set<std::variant<string::Epsilon, alphabet::Symbol>>, std::set<std::variant<string::Epsilon, alphabet::Symbol>>> partitioning = getLocalCallRetPartitioning(n, stateSymbol.first);

			std::set<std::variant<string::Epsilon, alphabet::Symbol>>& retPart = std::get<2>(partitioning);

			for(const auto& symbol : retPart) {
				if(stateSymbol.second == d.getBottomOfTheStackSymbol()) {
					retInitial(stateSymbol.first, stateSymbol.second, symbol, n, d, rubbishReturnTransitions);
				} else {
					ret(stateSymbol.first, stateSymbol.second, symbol, n, d, rubbishReturnTransitions);
				}
			}
		}

		for(const auto& state : states) {
			if ( common::GlobalData::verbose )
				std::clog << "Dirty state: " << state << std::endl;
			std::tuple<std::set<std::variant<string::Epsilon, alphabet::Symbol>>, std::set<std::variant<string::Epsilon, alphabet::Symbol>>, std::set<std::variant<string::Epsilon, alphabet::Symbol>>> partitioning = getLocalCallRetPartitioning(n, state);

			std::set<std::variant<string::Epsilon, alphabet::Symbol>>& localPart = std::get<0>(partitioning);
			std::set<std::variant<string::Epsilon, alphabet::Symbol>>& callPart = std::get<1>(partitioning);

			for(const auto& symbol : localPart) {
				local(state, symbol, n, d, rubbishLocalTransitions);
			}
			for(const auto& symbol : callPart) {
				call(state, symbol, n, d, rubbishCallTransitions);
			}
		}
	}

	std::set<label::Label> finalLabels = retrieveLabels(n.getFinalStates());
	for(const automaton::State& state : d.getStates()) {
		const std::set<label::Label> labels = retrieveDSubSet(unpackFromStateLabel(state.getName()));

		if(!std::excludes(finalLabels.begin(), finalLabels.end(), labels.begin(), labels.end())) {
			d.addFinalState(state);
		}
	}

	if ( common::GlobalData::verbose ) {
		std::clog << "Rubbish states: " << rubbishStates << std::endl;
		std::clog << "Rubbish return transitions: " << rubbishReturnTransitions << std::endl;
		std::clog << "Rubbish call transitions: " << rubbishCallTransitions << std::endl;
		std::clog << "Rubbish local transitions:" << rubbishLocalTransitions << std::endl;
	}

	return d;
}

auto DeterminizeRealTimeHeightDeterministicNPDA = Determinize::RegistratorWrapper<automaton::RealTimeHeightDeterministicDPDA, automaton::RealTimeHeightDeterministicNPDA>( Determinize::determinize );

} /* namespace determinize */

} /* namespace automaton */
