/*
 * RHPDAToPDA.cpp
 *
 *  Created on: 23. 3. 2014
 *	  Author: Jan Travnicek
 */

#include "RHPDAToPDA.h"

#include <exception/CommonException.h>
#include <automaton/PDA/RealTimeHeightDeterministicDPDA.h>
#include <automaton/PDA/RealTimeHeightDeterministicNPDA.h>
#include <automaton/PDA/DPDA.h>
#include <automaton/PDA/NPDA.h>

#include <alphabet/BottomOfTheStackSymbol.h>

#include <set>
#include <map>
#include <queue>
#include <iostream>

namespace automaton {

template < class T >
void constructTransitions ( const std::tuple < automaton::State, alphabet::Symbol, std::vector < alphabet::Symbol > > & stFirst, const std::map < automaton::State, std::set < std::tuple < std::vector < alphabet::Symbol >, automaton::State, std::vector < alphabet::Symbol > > > > & epsilonTransitions, const automaton::State & toState, std::vector < alphabet::Symbol > pops, std::vector < alphabet::Symbol > pushes, T & res ) {
	auto epsIter = epsilonTransitions.find ( toState );

	if ( ( epsIter != epsilonTransitions.end ( ) ) && !epsIter->second.empty ( ) ) {
		if ( epsIter->second.size ( ) == 1 ) {
			const auto & epsilonT = * epsIter->second.begin ( );

			pops.insert ( pops.end ( ), std::get < 0 > ( epsilonT ).begin ( ), std::get < 0 > ( epsilonT ).end ( ) );
			pushes.insert ( pushes.begin ( ), std::get < 2 > ( epsilonT ).rbegin ( ), std::get < 2 > ( epsilonT ).rend ( ) );

			constructTransitions ( stFirst, epsilonTransitions, std::get < 1 > ( epsilonT ), std::move ( pops ), std::move ( pushes ), res );
		} else {
			for ( const auto & epsilonT : epsIter->second ) {
				std::vector < alphabet::Symbol > popsCopy = pops;
				std::vector < alphabet::Symbol > pushesCopy = pushes;

				popsCopy.insert ( popsCopy.end ( ), std::get < 0 > ( epsilonT ).begin ( ), std::get < 0 > ( epsilonT ).end ( ) );
				pushesCopy.insert ( pushesCopy.begin ( ), std::get < 2 > ( epsilonT ).rbegin ( ), std::get < 2 > ( epsilonT ).rend ( ) );

				constructTransitions ( stFirst, epsilonTransitions, std::get < 1 > ( epsilonT ), std::move ( popsCopy ), std::move ( pushesCopy ), res );
			}
		}
	} else {
		res.addState ( std::get < 0 > ( stFirst ) );
		res.addState ( toState );

		res.addTransition ( std::get < 0 > ( stFirst ), std::get < 1 > ( stFirst ), std::move ( pops ), toState, std::move ( pushes ) );
	}
}

automaton::DPDA RHPDAToPDA::convert ( const automaton::RealTimeHeightDeterministicDPDA & pda ) {
	std::map < std::tuple < automaton::State, alphabet::Symbol, std::vector < alphabet::Symbol > >, std::set < std::pair < automaton::State, std::vector < alphabet::Symbol > > > > readingTransitions;
	std::map < automaton::State, std::set < std::tuple < std::vector < alphabet::Symbol >, automaton::State, std::vector < alphabet::Symbol > > > > epsilonTransitions;

	for ( const auto & transition : pda.getCallTransitions ( ) ) {
		if ( std::get < 1 > ( transition.first ).is < string::Epsilon > ( ) ) {
			auto & epsT = epsilonTransitions[std::get < 0 > ( transition.first )];
			const auto & to = transition.second;
			epsT.insert ( std::make_tuple ( std::vector < alphabet::Symbol > { }, to.first, std::vector < alphabet::Symbol > { to.second } ) );
		} else {
			auto & readT = readingTransitions[std::make_tuple ( std::get < 0 > ( transition.first ), std::get < 1 > ( transition.first ).get < alphabet::Symbol > ( ), std::vector < alphabet::Symbol > { } )];
			const auto & to = transition.second;
			readT.insert ( std::make_pair ( to.first, std::vector < alphabet::Symbol > { to.second } ) );
		}
	}

	for ( const auto & transition : pda.getLocalTransitions ( ) ) {
		if ( std::get < 1 > ( transition.first ).is < string::Epsilon > ( ) ) {
			auto & epsT = epsilonTransitions[std::get < 0 > ( transition.first )];
			const auto & to = transition.second;
			epsT.insert ( std::make_tuple ( std::vector < alphabet::Symbol > { }, to, std::vector < alphabet::Symbol > { } ) );
		} else {
			auto & readT = readingTransitions[std::make_tuple ( std::get < 0 > ( transition.first ), std::get < 1 > ( transition.first ).get < alphabet::Symbol > ( ), std::vector < alphabet::Symbol > { } )];
			const auto & to = transition.second;
			readT.insert ( std::make_pair ( to, std::vector < alphabet::Symbol > { } ) );
		}
	}

	for ( const auto & transition : pda.getReturnTransitions ( ) ) {
		if ( std::get < 2 > ( transition.first ).getData ( ) == alphabet::BottomOfTheStackSymbol::BOTTOM_OF_THE_STACK ) continue;

		if ( std::get < 1 > ( transition.first ).is < string::Epsilon > ( ) ) {
			auto & epsT = epsilonTransitions[std::get < 0 > ( transition.first )];
			const auto & to = transition.second;
			epsT.insert ( std::make_tuple ( std::vector < alphabet::Symbol > { std::get < 2 > ( transition.first ) }, to, std::vector < alphabet::Symbol > { } ) );
		} else {
			auto & readT = readingTransitions[std::make_tuple ( std::get < 0 > ( transition.first ), std::get < 1 > ( transition.first ).get < alphabet::Symbol > ( ), std::vector < alphabet::Symbol > { std::get < 2 > ( transition.first ) } )];
			const auto & to = transition.second;
			readT.insert ( std::make_pair ( to, std::vector < alphabet::Symbol > { } ) );
		}
	}

	for ( const auto & st : epsilonTransitions ) {
		bool allPops = true;

		for ( const auto & elems : st.second )
			allPops &= std::get < 2 > ( elems ).empty ( ); // if some pushes it will clear allPops variable

		if ( st.second.size ( ) == 1 ) continue;

		if ( ( st.second.size ( ) > 1 ) && allPops ) continue;

		throw exception::CommonException ( "Temporary states has more than one leaving transition" );
	}

	if ( epsilonTransitions[pda.getInitialState ( )].empty ( ) )
		throw exception::CommonException ( "Cannot determine initial pushdown store symbol" );

	// -------------------------------------------------------------------- initial state and initial stack symbol

	const auto & st = * epsilonTransitions[pda.getInitialState ( )].begin ( );

	std::vector < alphabet::Symbol > pops ( std::get < 0 > ( st ).begin ( ), std::get < 0 > ( st ).end ( ) );
	std::vector < alphabet::Symbol > pushes ( std::get < 2 > ( st ).rbegin ( ), std::get < 2 > ( st ).rend ( ) );

	automaton::State toState = std::get < 1 > ( st );

	while ( !epsilonTransitions[toState].empty ( ) ) {
		const auto & epsilonT = * epsilonTransitions[toState].begin ( );

		pops.insert ( pops.end ( ), std::get < 0 > ( epsilonT ).begin ( ), std::get < 0 > ( epsilonT ).end ( ) );
		pushes.insert ( pushes.begin ( ), std::get < 2 > ( epsilonT ).rbegin ( ), std::get < 2 > ( epsilonT ).rend ( ) );

		toState = std::get < 1 > ( epsilonT );
	}

	if ( ( pops.size ( ) != 0 ) && ( pushes.size ( ) != 1 ) )
		throw exception::CommonException ( "Cannot convert" );

	// -------------------------------------------------------------------- initial state and initial stack symbol

	automaton::DPDA res ( toState, pushes[0] );

	res.setInputAlphabet ( pda.getInputAlphabet ( ) );
	res.setStackAlphabet ( pda.getStackAlphabet ( ) );
	res.removeStackSymbol ( pda.getBottomOfTheStackSymbol ( ) );

	for ( const auto & st : readingTransitions )
		for ( const auto & to : st.second ) {
			std::vector < alphabet::Symbol > pops ( std::get < 2 > ( st.first ).begin ( ), std::get < 2 > ( st.first ).end ( ) );
			std::vector < alphabet::Symbol > pushes ( to.second.rbegin ( ), to.second.rend ( ) );

			constructTransitions ( st.first, epsilonTransitions, to.first, std::move ( pops ), std::move ( pushes ), res );
		}

	res.setFinalStates ( pda.getFinalStates ( ) );

	return res;
}

auto RHPDAToPDARealTimeHeightDeterministicDPDA = RHPDAToPDA::RegistratorWrapper < automaton::DPDA, automaton::RealTimeHeightDeterministicDPDA > ( RHPDAToPDA::convert );

 // This may not work correctly -- generation of initial state and initial symbol
automaton::NPDA RHPDAToPDA::convert ( const automaton::RealTimeHeightDeterministicNPDA & pda ) {
	std::map < std::tuple < automaton::State, alphabet::Symbol, std::vector < alphabet::Symbol > >, std::set < std::pair < automaton::State, std::vector < alphabet::Symbol > > > > readingTransitions;
	std::map < automaton::State, std::set < std::tuple < std::vector < alphabet::Symbol >, automaton::State, std::vector < alphabet::Symbol > > > > epsilonTransitions;

	for ( const auto & transition : pda.getCallTransitions ( ) ) {
		if ( std::get < 1 > ( transition.first ).is < string::Epsilon > ( ) ) {
			auto & epsT = epsilonTransitions[std::get < 0 > ( transition.first )];

			for ( const auto & to : transition.second ) {
				epsT.insert ( std::make_tuple ( std::vector < alphabet::Symbol > { }, to.first, std::vector < alphabet::Symbol > { to.second } ) );
			}
		} else {
			auto & readT = readingTransitions[std::make_tuple ( std::get < 0 > ( transition.first ), std::get < 1 > ( transition.first ).get < alphabet::Symbol > ( ), std::vector < alphabet::Symbol > { } )];

			for ( const auto & to : transition.second ) {
				readT.insert ( std::make_pair ( to.first, std::vector < alphabet::Symbol > { to.second } ) );
			}
		}
	}

	for ( const auto & transition : pda.getLocalTransitions ( ) ) {
		if ( std::get < 1 > ( transition.first ).is < string::Epsilon > ( ) ) {
			auto & epsT = epsilonTransitions[std::get < 0 > ( transition.first )];

			for ( const auto & to : transition.second ) {
				epsT.insert ( std::make_tuple ( std::vector < alphabet::Symbol > { }, to, std::vector < alphabet::Symbol > { } ) );
			}
		} else {
			auto & readT = readingTransitions[std::make_tuple ( std::get < 0 > ( transition.first ), std::get < 1 > ( transition.first ).get < alphabet::Symbol > ( ), std::vector < alphabet::Symbol > { } )];

			for ( const auto & to : transition.second ) {
				readT.insert ( std::make_pair ( to, std::vector < alphabet::Symbol > { } ) );
			}
		}
	}

	for ( const auto & transition : pda.getReturnTransitions ( ) ) {
		if ( std::get < 2 > ( transition.first ).getData ( ) == alphabet::BottomOfTheStackSymbol::BOTTOM_OF_THE_STACK ) continue;

		if ( std::get < 1 > ( transition.first ).is < string::Epsilon > ( ) ) {
			auto & epsT = epsilonTransitions[std::get < 0 > ( transition.first )];

			for ( const auto & to : transition.second ) {
				epsT.insert ( std::make_tuple ( std::vector < alphabet::Symbol > { std::get < 2 > ( transition.first ) }, to, std::vector < alphabet::Symbol > { } ) );
			}
		} else {
			auto & readT = readingTransitions[std::make_tuple ( std::get < 0 > ( transition.first ), std::get < 1 > ( transition.first ).get < alphabet::Symbol > ( ), std::vector < alphabet::Symbol > { std::get < 2 > ( transition.first ) } )];

			for ( const auto & to : transition.second ) {
				readT.insert ( std::make_pair ( to, std::vector < alphabet::Symbol > { } ) );
			}
		}
	}

	for ( const auto & st : epsilonTransitions ) {
		bool allPops = true;

		for ( const auto & elems : st.second )
			allPops &= std::get < 2 > ( elems ).empty ( ); // if some pushes it will clear allPops variable

		if ( ( st.second.size ( ) != 1 ) && !allPops ) throw exception::CommonException ( "Temporary states has more than one leaving transition" );
	}

	if ( pda.getInitialStates ( ).size ( ) != 1 )
		throw exception::CommonException ( "Cannot convert" );

	const automaton::State & initialState = * pda.getInitialStates ( ).begin ( );

	if ( epsilonTransitions[initialState].empty ( ) )
		throw exception::CommonException ( "Cannot convert" );

	// -------------------------------------------------------------------- initial state and initial stack symbol

	const auto & st = * epsilonTransitions[initialState].begin ( );

	std::vector < alphabet::Symbol > pops ( std::get < 0 > ( st ).begin ( ), std::get < 0 > ( st ).end ( ) );
	std::vector < alphabet::Symbol > pushes ( std::get < 2 > ( st ).rbegin ( ), std::get < 2 > ( st ).rend ( ) );

	automaton::State toState = std::get < 1 > ( st );

	while ( !epsilonTransitions[toState].empty ( ) ) {
		const auto & epsilonT = * epsilonTransitions[toState].begin ( );

		pops.insert ( pops.end ( ), std::get < 0 > ( epsilonT ).begin ( ), std::get < 0 > ( epsilonT ).end ( ) );
		pushes.insert ( pushes.begin ( ), std::get < 2 > ( epsilonT ).rbegin ( ), std::get < 2 > ( epsilonT ).rend ( ) );

		toState = std::get < 1 > ( epsilonT );
	}

	if ( ( pops.size ( ) != 0 ) && ( pushes.size ( ) != 1 ) )
		throw exception::CommonException ( "Cannot convert" );

	// -------------------------------------------------------------------- initial state and initial stack symbol

	automaton::NPDA res ( toState, pushes[0] );

	res.setInputAlphabet ( pda.getInputAlphabet ( ) );
	res.setStackAlphabet ( pda.getStackAlphabet ( ) );
	res.removeStackSymbol ( pda.getBottomOfTheStackSymbol ( ) );

	for ( const auto & st : readingTransitions )
		for ( const auto & to : st.second ) {
			std::vector < alphabet::Symbol > pops ( std::get < 2 > ( st.first ).begin ( ), std::get < 2 > ( st.first ).end ( ) );
			std::vector < alphabet::Symbol > pushes ( to.second.rbegin ( ), to.second.rend ( ) );

			constructTransitions ( st.first, epsilonTransitions, to.first, std::move ( pops ), std::move ( pushes ), res );
		}

	res.setFinalStates ( pda.getFinalStates ( ) );

	return res;
}

auto RHPDAToPDARealTimeHeightDeterministicNPDA = RHPDAToPDA::RegistratorWrapper < automaton::NPDA, automaton::RealTimeHeightDeterministicNPDA > ( RHPDAToPDA::convert );

automaton::Automaton RHPDAToPDA::convert ( const Automaton & automaton ) {
	return dispatch ( automaton.getData ( ) );
}

}

namespace alib {

auto DPDAFromRealTimeHeightDeterministicNPDA = castApi::CastRegister < automaton::DPDA, automaton::RealTimeHeightDeterministicDPDA > ( automaton::RHPDAToPDA::convert );
auto NPDAFromRealTimeHeightDeterministicNPDA = castApi::CastRegister < automaton::NPDA, automaton::RealTimeHeightDeterministicNPDA > ( automaton::RHPDAToPDA::convert );
}
