#include "HopcroftGraph.h"
#include "HopcroftDebug.h"

#include <algorithm>

#include <exception/CommonException.h>

#include <iostream>

namespace graph
{

namespace isomorphism
{

namespace hopcroft
{

Graph::Graph(const UndirectedGraph &graph)
{
	init(graph);
}

Graph::~Graph()
{
	for (Face *f : faces)
		delete f;

	for (Vertex *v : vertices)
		delete v;

	for (EdgeEnd *e : edge_ends)
		delete e;
}

// In: e1 <cw> [newedge] <cw> e2->otheredge <cw> e1
// Out: newedge
EdgeEnd *Graph::addInnerEdge(EdgeEnd *e1, EdgeEnd *e2)
{
	H_ASSERT(e1->vertex != e2->vertex);
	H_ASSERT(e1->otheredge->ccwedge == e2);
	H_ASSERT(std::find(edge_ends.begin(), edge_ends.end(), e1) != edge_ends.end());
	H_ASSERT(std::find(edge_ends.begin(), edge_ends.end(), e2) != edge_ends.end());

	//H_DEBUG("add inner edge" << e1->dump() << e2->dump());

	EdgeEnd *enew = new EdgeEnd;
	enew->vertex = e2->vertex;

	EdgeEnd *enewother = new EdgeEnd;
	enewother->vertex = e1->vertex;

	enew->otheredge = enewother;
	enewother->otheredge = enew;

	enew->cwedge = e2->otheredge;
	enew->ccwedge = e2->ccwedge;
	e2->ccwedge->otheredge->cwedge = enewother;
	e2->ccwedge = enewother;
	if (e2->cwedge == e2->otheredge)
		e2->cwedge = enewother;

	enewother->cwedge = e1->cwedge;
	enewother->ccwedge = e1->otheredge;
	e1->cwedge->otheredge->ccwedge = enew;
	e1->cwedge = enew;
	if (e1->ccwedge == e1->otheredge)
		e1->ccwedge = enew;

	if (!enew->isMultiEdge()) {
		enew->vertex->degree++;
		enewother->vertex->degree++;
	}

	Face *facenew = new Face;
	facenew->edges = 3;
	facenew->fedge = e2;

	e1->otheredge->ccwface = facenew;
	enewother->ccwface = facenew;
	e2->ccwface = facenew;

	Face *faceold = enew->ccwedge->ccwface;
	faceold->edges--;
	faceold->fedge = enew;
	enew->ccwface = faceold;

	faces.push_back(facenew);
	edge_ends.push_back(enew);
	edge_ends.push_back(enewother);

	return enew;
}

// In: e distinguished edge of 3-face
// Out: face edges are reconnected to the new vertex
void Graph::addAndReconnectVertexIntoTriangularFace(EdgeEnd *e)
{
	Vertex *v = new Vertex;
	EdgeEnd *f = e->ccwedge;
	EdgeEnd *g = f->ccwedge;
	Face *oldface = e->ccwface;

	H_ASSERT(e == g->ccwedge);

	e->cwedge->otheredge->ccwedge = f;
	f->cwedge->otheredge->ccwedge = g;
	g->cwedge->otheredge->ccwedge = e;

	e->cwedge = f->otheredge;
	e->ccwedge = g->otheredge;
	f->cwedge = g->otheredge;
	f->ccwedge = e->otheredge;
	g->cwedge = e->otheredge;
	g->ccwedge = f->otheredge;

	v->degree = 3;
	v->vedge = e->otheredge;

	e->otheredge->vertex->updateDegree();
	f->otheredge->vertex->updateDegree();
	g->otheredge->vertex->updateDegree();

	e->ccwface = g->otheredge->ccwface;
	f->ccwface = e->otheredge->ccwface;
	g->ccwface = f->otheredge->ccwface;

	e->ccwface->edges++;
	f->ccwface->edges++;
	g->ccwface->edges++;

	e->ccwface->fedge = e;
	f->ccwface->fedge = f;
	g->ccwface->fedge = g;

	removeFace(oldface);
	vertices.push_back(v);
}

// In: v -> w edge-end (v = the vertex)
// Out: w remains
void Graph::removeDistinguishedEdgeAndVertex(EdgeEnd *e)
{
	H_ASSERT(std::find(edge_ends.begin(), edge_ends.end(), e) != edge_ends.end());

	//H_DEBUG("remove distinguished edge and vertex" << e->dump());

	EdgeEnd *o = e->otheredge;

	Vertex *w = e->vertex;
	Vertex *v = o->vertex;

	EdgeEnd *wl = e->cwedge;
	EdgeEnd *wk = e->ccwedge;
	EdgeEnd *vl = o->cwedge;
	EdgeEnd *vk = o->ccwedge;

	if (wk == o)
		wk = vk;

	if (wl == o)
		wl = vl;

	v->forEachEdge([=](EdgeEnd *e) {
		e->otheredge->vertex = w;
		return true;
	});

	wl->otheredge->ccwedge = vk;
	wk->otheredge->cwedge = vl;
	vl->otheredge->ccwedge = wk;
	vk->otheredge->cwedge = wl;

	w->vedge = wk;
	w->updateDegree();

	w->forEachEdge([=](EdgeEnd *e) {
		e->vertex->updateDegree();
		return true;
	});

	wk->ccwface->edges--;
	wk->ccwface->fedge = wk;
	vk->ccwface->edges--;
	vk->ccwface->fedge = vk;

	removeVertex(v);
	removeEdgeEnd(e);
	removeEdgeEnd(o);
}

// In: v - exception vertex, edge e points to 4 degree vertex
void Graph::removeFourDegreeExceptionVertex(Vertex *v, EdgeEnd *e)
{
	H_ASSERT(v->degree == 4);
	H_ASSERT(e->vertex->degree == 4);
	H_ASSERT(!e->isMultiEdge());

	//H_DEBUG("remove four degree exception vertex" << v->dump() << e->dump());

	// Out: e <ccw> dest
	auto reconnectEdge = [=](EdgeEnd *e, EdgeEnd *dest) {
		e->vertex = dest->otheredge->vertex;
		e->ccwedge = dest;
		e->cwedge = dest->otheredge->cwedge;
		dest->otheredge->cwedge = e->otheredge;
		if (dest->otheredge->ccwedge == dest->otheredge)
			dest->otheredge->ccwedge = e->otheredge;
		e->cwedge->otheredge->ccwedge = e->otheredge;
	};

	EdgeEnd *er = e;
	EdgeEnd *es = er->otheredge->cwedge;
	EdgeEnd *et = es->otheredge->cwedge;
	EdgeEnd *eu = et->otheredge->cwedge;

	EdgeEnd *destroyed1 = es->otheredge;
	EdgeEnd *destroyed2 = eu->otheredge;

	reconnectEdge(et->otheredge, eu->otheredge);
	reconnectEdge(er->otheredge, es->otheredge);

	// Merge es + eu
	es->otheredge = eu;
	eu->otheredge = es;
	es->vertex->vedge = eu;
	eu->vertex->vedge = es;
	es->cwedge->otheredge->ccwedge = eu;
	es->ccwedge->otheredge->cwedge = eu;
	eu->cwedge->otheredge->ccwedge = es;
	eu->ccwedge->otheredge->cwedge = es;

	// Update faces
	Face *face1 = er->otheredge->ccwface;
	Face *face2 = et->otheredge->ccwface;
	Face *face3 = er->ccwface;
	Face *face4 = et->ccwface;

	eu->ccwface = face1;
	es->ccwface = face2;
	face1->fedge = eu;
	face2->fedge = es;
	face3->fedge = er;
	face4->fedge = et;
	face3->edges--;
	face4->edges--;

	// Update degrees
	er->vertex->updateDegree();
	es->vertex->updateDegree();
	et->vertex->updateDegree();
	eu->vertex->updateDegree();

	removeVertex(v);
	removeEdgeEnd(destroyed1);
	removeEdgeEnd(destroyed2);
}

void Graph::removeEdge(EdgeEnd *e, bool removeVertices)
{
	H_ASSERT(std::find(edge_ends.begin(), edge_ends.end(), e) != edge_ends.end());

	//H_DEBUG("remove edge" << e->dump() << e->otheredge->dump());

	EdgeEnd *o = e->otheredge;

	if (!e->isMultiEdge()) {
		e->vertex->degree--;
		// In case of loops
		if (e->vertex != o->vertex)
			o->vertex->degree--;
	}

	// Update edges
	e->cwedge->otheredge->ccwedge = e->ccwedge;
	if (e->cwedge->otheredge->cwedge == o)
		e->cwedge->otheredge->cwedge = e->cwedge->otheredge;

	e->ccwedge->otheredge->cwedge = e->cwedge;
	if (e->ccwedge->otheredge->ccwedge == o)
		e->ccwedge->otheredge->ccwedge = e->ccwedge->otheredge;

	o->cwedge->otheredge->ccwedge = o->ccwedge;
	if (o->cwedge->otheredge->cwedge == e)
		o->cwedge->otheredge->cwedge = o->cwedge->otheredge;

	o->ccwedge->otheredge->cwedge = o->cwedge;
	if (o->ccwedge->otheredge->ccwedge == e)
		o->ccwedge->otheredge->ccwedge = o->ccwedge->otheredge;

	// Update vedges
	if (e->vertex->vedge == o)
		e->vertex->vedge = e->cwedge;

	if (o->vertex->vedge == e)
		o->vertex->vedge = o->cwedge;

	// Handle zero degree vertices
	if (removeVertices) {
		if (e->vertex->degree == 0)
			removeVertex(e->vertex);

		if (o->vertex->degree == 0)
			removeVertex(o->vertex);
	} else {
		if (e->vertex->degree == 0)
			e->vertex->vedge = nullptr;

		if (o->vertex->degree == 0)
			o->vertex->vedge = nullptr;
	}

	// Update faces
	Face *face1 = e->ccwface;
	Face *face2 = o->ccwface;

	if (face1 != face2) {
		face1->edges = 0;
		face1->fedge = (e->ccwedge == e || e->ccwedge == o) ? o->ccwedge : e->ccwedge;
		EdgeEnd *edge = face1->fedge;
		do {
			face1->edges++;
			edge->ccwface = face1;
			edge = edge->ccwedge;
		} while (edge != face1->fedge);
		removeFace(face2);
	} else {
		face1->edges -= 2;
		face1->fedge = o->ccwedge == o ? e->ccwedge : o->ccwedge;
		if (face1->edges == 0)
			removeFace(face1);
	}

	// Finally remove edge
	removeEdgeEnd(e);
	removeEdgeEnd(o);
}

void Graph::removeFace(Face *f)
{
	H_ASSERT(std::find(faces.begin(), faces.end(), f) != faces.end());

	faces.erase(std::find(faces.begin(), faces.end(), f));
	delete f;
}

void Graph::removeVertex(Vertex *v)
{
	H_ASSERT(std::find(vertices.begin(), vertices.end(), v) != vertices.end());

	vertices.erase(std::find(vertices.begin(), vertices.end(), v));
	delete v;
}

void Graph::removeEdgeEnd(EdgeEnd *e)
{
	H_ASSERT(std::find(edge_ends.begin(), edge_ends.end(), e) != edge_ends.end());

	edge_ends.erase(std::find(edge_ends.begin(), edge_ends.end(), e));
	delete e;
}

std::vector<std::vector<EdgeEnd*>> Graph::findClumps(Vertex *v, Vertex *w) const
{
	std::vector<EdgeEnd*> clump;
	std::vector<std::vector<EdgeEnd*>> clumps;

	v->forEachEdge([&](EdgeEnd *e) {
		if (e->vertex == w && e->otheredge->vertex == v && (clump.empty() || clump.back()->cwedge == e->otheredge)) {
			clump.push_back(e);
		} else if (!clump.empty()) {
			clumps.push_back(clump);
			clump.clear();
			clump.push_back(e);
		}
		return true;
	});

	if (!clump.empty())
		clumps.push_back(clump);

	if (clumps.empty())
		return clumps;

	// Merge potential <-clump1) .. (clump2) .. (clump1->
	if (clumps.size() > 1 && clumps.back().back()->cwedge == clumps.front().front()->otheredge) {
		clumps[0].insert(clumps[0].end(), clumps.back().begin(), clumps.back().end());
		clumps.pop_back();
	}

	// Remove size-one clumps
	for (size_t i = 0; i < clumps.size(); ++i)
		if (clumps[i].size() < 2)
			clumps.erase(clumps.begin() + i--);

	return clumps;
}

void Graph::init(const UndirectedGraph &graph)
{
	std::unordered_map<Node, Vertex*> v_map;
	std::unordered_map<UndirectedEdge, EdgeEnd*> e_map;

	for (const Node &node : graph.getNodes()) {
		Vertex *v = new Vertex;
		v->out = std::string(node.getName());
		vertices.push_back(v);
		v_map[node] = v;
	}

	for (const UndirectedEdge &edge : graph.getEdges()) {
		EdgeEnd *e1 = new EdgeEnd;
		e1->vertex = v_map[edge.getSecondNode()];
		edge_ends.push_back(e1);

		EdgeEnd *e2 = new EdgeEnd;
		e2->vertex = v_map[edge.getFirstNode()];
		edge_ends.push_back(e2);

		e1->otheredge = e2;
		e2->otheredge = e1;
		e1->vertex->vedge = e2;
		e2->vertex->vedge = e1;

		e_map[edge] = e1;
	}

	// Embedding
	std::unordered_map<Node, std::vector<EdgeEnd*>> sortedNodeEdges;

	for (const auto &p : graph.getEmbedding()) {
		Vertex *vertex = v_map[p.first];
		std::vector<EdgeEnd*> sortedEdges;
		std::set<UndirectedEdge> neighborEdges = graph.neighborEdges(p.first);

		// Multi-edges must not cross the vector boundary in embedding
		// eg. (n1, n2, n3, n4, n1) must be changed to (n1, n1, n2, n3, n4)
		auto reorder = [&](const std::vector<Node> &v) -> std::vector<Node> {
			if (v_map.size() == 1 || v.front() != v.back())
				return v;

			std::vector<Node> out = v;
			while (out.front() == out.back()) {
				out.insert(out.begin(), out.back());
				out.pop_back();
			}
			return out;
		};

		for (const Node &node : reorder(p.second)) {
			Vertex *v = v_map[node];
			EdgeEnd *edge = nullptr;

			for (const UndirectedEdge &edg : neighborEdges) {
				EdgeEnd *e = e_map[edg];
				if (e->vertex == vertex)
					e = e->otheredge;
				if (e->vertex == v) {
					edge = e;
					const Node &othernode = edg.getFirstNode() == p.first ? edg.getSecondNode() : edg.getFirstNode();
					auto &vec = sortedNodeEdges[othernode];
					for (size_t i = vec.size(); i-- > 0; ) {
						bool isMultiEdge = false;
						EdgeEnd *e = vec[i]->otheredge;
						if (e->ccwedge && e->ccwedge->vertex == e->otheredge->vertex)
							isMultiEdge = true;
						if (e->cwedge && e->cwedge->vertex == e->otheredge->vertex)
							isMultiEdge = true;
						if (isMultiEdge && e->vertex == v && e->otheredge->vertex == vertex) {
							vec.erase(vec.begin() + i);
							edge = e;
							break;
						}
					}

					neighborEdges.erase(edg);
					break;
				}
			}

			H_ASSERT(edge);
			sortedEdges.push_back(edge);
		}

		sortedNodeEdges[p.first] = sortedEdges;

		for (size_t i = 0; i < sortedEdges.size(); ++i) {
			EdgeEnd *e1 = sortedEdges[i];
			EdgeEnd *e2 = i == sortedEdges.size() - 1 ? sortedEdges[0] : sortedEdges[i + 1];

			H_ASSERT(e1 != e2);
			H_ASSERT(!e1->isLoopEdge());

			// CW                            // CCW
			e1->otheredge->ccwedge = e2;     // e1->otheredge->cwedge = e2;
			e2->otheredge->cwedge = e1;      // e2->otheredge->ccwedge = e1;
		}
	}

	if (sortedNodeEdges.empty() && !e_map.empty())
		throw exception::CommonException("HopcroftWong: Graph is not embedded");

	for (Vertex *vertex : vertices) {
		vertex->degree = 1;
		vertex->updateDegree();
	}

	// Init faces
	for (EdgeEnd *e : edge_ends) {
		if (e->ccwface)
			continue;
		Face *f = new Face;
		faces.push_back(f);
		f->fedge = e;
		EdgeEnd *next = e;
		do {
			f->edges++;
			next->ccwface = f;
			next = next->ccwedge;
		} while (next != e);
	}
}

void Graph::dump()
{
	std::cout << std::endl;
	std::cout << "Vertices: " << vertices.size();
	std::cout << " Edges: " << edge_ends.size() / 2;
	std::cout << std::endl;
	for (Vertex *v : vertices) {
		std::cout << v->dump() << "(" << v->degree << "):";
		v->forEachEdge([this](EdgeEnd *e) {
			std::cout << e->dump() << " (" << std::distance(edge_ends.begin(), std::find(edge_ends.begin(), edge_ends.end(), e)) << ")";
			return true;
		});
		std::cout << std::endl;
	}
	std::cout << std::endl;
}

void Graph::dumpLabels()
{
	std::cout << std::endl;
	std::cout << "Vertices:" << std::endl;
	for (Vertex *v : vertices)
		std::cout << " " << v->vlabel;
	std::cout << std::endl;

	std::cout << "Edges:" << std::endl;
	for (Vertex *v : vertices) {
		std::cout << v->dump() << ":";
		v->forEachEdge([this](EdgeEnd *e) {
			std::cout << " " << e->elabel << "-" << e->otheredge->elabel;
			return true;
		});
		std::cout << std::endl;
	}
	std::cout << std::endl;
}

Vertex::Vertex()
	: vlabel(1)
	, degree(0)
	, vedge(nullptr)
{
}

bool Vertex::isKnot() const
{
	if (degree != 1)
		return false;

	int loops = 0;
	bool res = true;

	forEachEdge([&](EdgeEnd *e) {
		if (!e->isLoopEdge()) {
			res = false;
			return false;
		}
		loops++;
		return true;
	});

	return res && loops == 1;
}

bool Vertex::isCorolla() const
{
	if (degree != 1)
		return false;

	int loops = 0;
	bool res = true;

	forEachEdge([&](EdgeEnd *e) {
		if (!e->isLoopEdge()) {
			res = false;
			return false;
		}
		loops++;
		return true;
	});

	return res && loops > 1;
}

bool Vertex::isStar() const
{
	int edges = 0;
	bool res = true;

	forEachEdge([&](EdgeEnd *e) {
		if (!e->isSpoke())
			res = false;
		edges++;
		return res;
	});

	return res && edges > 1;
}

bool Vertex::isDumbell() const
{
	int edges = 0;
	bool res = true;

	forEachEdge([&](EdgeEnd *e) {
		if (!e->isSpoke())
			res = false;
		edges++;
		return res;
	});

	return res && edges == 1;
}

bool Vertex::isLoopVertex() const
{
	bool res = false;

	forEachEdge([&](EdgeEnd *e) {
		res = e->isLoopEdge();
		return !res;
	});

	return res;
}

bool Vertex::isSpokeCenter() const
{
	bool res = false;

	forEachEdge([&](EdgeEnd *e) {
		res = e->isSpoke();
		return !res;
	});

	return res;
}

bool Vertex::hasNonLoopEdge() const
{
	bool res = false;

	forEachEdge([&](EdgeEnd *e) {
		res = !e->isLoopEdge();
		return !res;
	});

	return res;
}

bool Vertex::hasNonSpokeEdge() const
{
	bool res = false;

	forEachEdge([&](EdgeEnd *e) {
		res = !e->isSpoke();
		return !res;
	});

	return res;
}

bool Vertex::isDegreeIsolated() const
{
	bool res = true;

	forEachEdge([&](EdgeEnd *e) {
		if (e->vertex->degree == degree)
			res = false;
		return res;
	});

	return res;
}

std::vector<Face*> Vertex::neighborFaces(Direction direction) const
{
	std::vector<Face*> out;

	forEachEdge([&](EdgeEnd *e) {
		out.push_back(e->ccwface);
		return true;
	}, direction);

	return out;
}

std::vector<EdgeEnd*> Vertex::neighborEdges(Direction direction) const
{
	std::vector<EdgeEnd*> out;

	forEachEdge([&](EdgeEnd *e) {
		out.push_back(e);
		return true;
	}, direction);

	return out;
}

void Vertex::forEachEdge(std::function<bool(EdgeEnd*)> callback, Direction direction) const
{
	// Current edge can be removed from within the callback (without removing vertices)

	if (!vedge)
		return;

	EdgeEnd *s = vedge;
	EdgeEnd *e = vedge;
	do {
		EdgeEnd *n = direction == ClockWise ? e->otheredge->ccwedge : e->otheredge->cwedge;
		if (!callback(e))
			return;
		e = n;
	} while (e != s && degree > 0);
}

void Vertex::updateDegree()
{
	std::set<Vertex*> nghbrs;

	forEachEdge([&](EdgeEnd *e) {
		nghbrs.insert(e->vertex);
		return true;
	});

	degree = nghbrs.size();
}

std::string Vertex::dump() const
{
	return " " + out;
}

EdgeEnd::EdgeEnd()
	: elabel(2)
	, vertex(nullptr)
	, ccwface(nullptr)
	, ccwedge(nullptr)
	, cwedge(nullptr)
	, otheredge(nullptr)
{
}

bool EdgeEnd::isSpoke() const
{
	return !isMultiEdge() && (vertex->degree == 1 || otheredge->vertex->degree == 1);
}

bool EdgeEnd::isLoopEdge() const
{
	return vertex == otheredge->vertex;
}

bool EdgeEnd::isMultiEdge() const
{
	bool res = false;

	otheredge->vertex->forEachEdge([&](EdgeEnd *e) {
		if (e == this || e == otheredge)
			return true;
		res = e->vertex == vertex && e->otheredge->vertex == otheredge->vertex;
		return !res;
	});

	return res;
}

std::string EdgeEnd::dump() const
{
	return " " + otheredge->vertex->out + "->" + vertex->out;
}

Face::Face()
	: edges(0)
	, fedge(nullptr)
{
}

} // namespace hopcroft

} // namespace isomorphism

} // namespace graph
