#include "HopcroftImpl.h"
#include "HopcroftDebug.h"

#include <tuple>
#include <iostream>
#include <algorithm>

#include <exception/CommonException.h>

namespace graph
{

namespace isomorphism
{

namespace hopcroft
{

// Helpers

#define ARG_REF(x) decltype(x##1) &x
#define ARG_CREF(x) const decltype(x##1) &x

#define REDUCTION_INIT \
	int __flab = 0; \
	(void)__flab;

#define PREPARE_FUNCTION REDUCTION_INIT auto findReductions = [&]
#define APPLY_FUNCTION auto applyReductions = [&]

#define GET_LABELS(labels) \
	( __flab++ == 0 \
		? std::vector<int>(labels.begin(), labels.begin() + labels.size() / 2) \
		: std::vector<int>(labels.begin() + labels.size() / 2 , labels.end()) \
	)

#define VECTOR_PARTITIONING(numbers, labels) \
	labels = vectorPartitioning(numbers);

#define CIRCLE_PARTITIONING(numbers, labels) \
	labels = circlePartitioning(numbers);

#define CHECK_REDUCTIONS(x) \
	if (x##1 .size() != x##2 .size() || __ig1 != __ig2) { \
		m_fail = true; \
		return true; \
	} \
	if (x##1 .empty()) { \
		return false; \
	}

#define CHECK_REDUCTIONS2(x1, x2) \
	if (x1##1 .size() != x1##2 .size() || x2##1 .size() != x2##2 .size() || __ig1 != __ig2) { \
		m_fail = true; \
		return true; \
	} \
	if (x1##1 .empty() && x2##1 .empty()) { \
		return false; \
	}

#define PREPARE_REDUCTIONS_X(x1, x2) \
	saveInteger(); \
	x1; \
	int __ig1 = m_integer; \
	restoreInteger(); \
	x2; \
	int __ig2 = m_integer; \

#define PREPARE_REDUCTIONS2(x1, x2) \
	PREPARE_REDUCTIONS_X( \
		findReductions(x1##1, x2##1), \
		findReductions(x1##2, x2##2))

#define PREPARE_REDUCTIONS3(x1, x2, x3) \
	PREPARE_REDUCTIONS_X( \
		findReductions(x1##1, x2##1, x3##1), \
		findReductions(x1##2, x2##2, x3##2))

#define APPLY_REDUCTIONS_X(x1, x2) \
	saveInteger(); \
	x1; \
	__ig1 = m_integer; \
	restoreInteger(); \
	x2; \
	__ig2 = m_integer; \
	m_fail = __ig1 != __ig2;

#define APPLY_REDUCTIONS2(x1, x2) \
	APPLY_REDUCTIONS_X( \
		applyReductions(x1##1, x2##1), \
		applyReductions(x1##2, x2##2))

#define APPLY_REDUCTIONS3(x1, x2, x3) \
	APPLY_REDUCTIONS_X( \
		applyReductions(x1##1, x2##1, x3##1), \
		applyReductions(x1##2, x2##2, x3##2))

template<typename T, typename S>
inline bool contains(const T &container, const S &value)
{
	return std::find(container.begin(), container.end(), value) != container.end();
}

template<typename T, typename S>
inline int indexOf(const T &container, const S &value)
{
	auto it = std::find(container.begin(), container.end(), value);
	if (it == container.end())
		return -1;
	return std::distance(container.begin(), it);
}

template<typename T>
inline void deleteAll(const T &container)
{
	for (auto v : container)
		delete v;
}

template<typename T>
inline void deleteAllValues(const T &container)
{
	for (const auto &v : container)
		delete v.second;
}

static const int SPECIAL_X = 12345;

// HopcroftImpl

HopcroftImpl::HopcroftImpl(Graph *g1, Graph *g2)
	: m_g1(g1)
	, m_g2(g2)
	, m_fail(false)
	, m_integer(3) // 1 - new vertices, 2 - new edges
	, m_savedInteger(0)
{
}

bool HopcroftImpl::testIsomorphism()
{
	while (doReductions()) {
		if (m_g1->vertices.size() != m_g2->vertices.size())
			m_fail = true;
		if (m_g1->edge_ends.size() != m_g2->edge_ends.size())
			m_fail = true;
		if (m_g1->faces.size() != m_g2->faces.size())
			m_fail = true;
		if (m_fail)
			break;
	}

	return !m_fail && testByDefinition();
}

int HopcroftImpl::newInteger()
{
	return m_integer++;
}

void HopcroftImpl::saveInteger()
{
	m_savedInteger = m_integer;
}

void HopcroftImpl::restoreInteger()
{
	m_integer = m_savedInteger;
}

// 2.A Vector partitioning
std::vector<int> HopcroftImpl::vectorPartitioning(const std::vector<std::vector<int>> &tuples)
{
	std::vector<int> out(tuples.size(), -1);

	for (size_t i = 0; i < out.size(); ++i) {
		if (out[i] != -1)
			continue;
		int c = newInteger();
		out[i] = c;
		for (size_t j = 0; j < out.size(); ++j)
			if (tuples[i] == tuples[j])
				out[j] = c;
	}

	return out;
}

// 2.B Circle partitioning
bool HopcroftImpl::circleIsomorphism(std::vector<int> a, std::vector<int> b)
{
	auto distinctIntegers = [](const std::vector<int> &circle) {
		if (circle.empty())
			return false;

		int first = circle.at(0);
		for (const int &val : circle)
			if (val != first)
				return true;

		return false;
	};

	int savedInteger = m_integer;

	while (distinctIntegers(a) || distinctIntegers(b)) {
		std::vector<std::vector<int>> tuples;

		auto constructTuples = [&tuples](const std::vector<int> &circle) {
			for (size_t i = 0; i < circle.size(); ++i) {
				size_t first = i;
				size_t second = (i == circle.size() - 1) ? 0 : i + 1;
				if (circle.at(first) != circle.at(second))
					tuples.push_back({circle.at(first), circle.at(second)});
			}
		};

		auto replaceComponents = [&tuples](std::vector<int> &circle, int idx, int cls) {
			const std::vector<int> &comp = tuples.at(idx);
			H_ASSERT(comp.size() == 2);

			while (true) {
				size_t first;
				size_t second;
				bool found = false;
				for (size_t i = 0; i < circle.size(); ++i) {
					first = i;
					second = (i == circle.size() - 1) ? 0 : i + 1;
					if (circle.at(first) == comp.at(0) && circle.at(second) == comp.at(1)) {
						found = true;
						break;
					}
				}
				if (!found)
					break;

				circle[first] = cls;
				circle.erase(circle.begin() + second);
			}
		};

		constructTuples(a);
		constructTuples(b);

		const std::vector<int> &part = vectorPartitioning(tuples);
		for (size_t i = 0; i < part.size(); ++i) {
			int cls = part.at(i);
			for (size_t j = i; j < part.size(); ++j) {
				if (part.at(j) != cls)
					continue;

				replaceComponents(a, j, cls);
				replaceComponents(b, j, cls);
			}
		}
	}

	m_integer = savedInteger;

	return a == b;
}

std::vector<int> HopcroftImpl::circlePartitioning(const std::vector<std::vector<int>> &circles)
{
	std::vector<int> out(circles.size(), -1);

	for (size_t i = 0; i < out.size(); ++i) {
		if (out[i] != -1)
			continue;
		int c = newInteger();
		out[i] = c;
		for (size_t j = 0; j < out.size(); ++j)
			if (circleIsomorphism(circles[i], circles[j]))
				out[j] = c;
	}

	return out;
}

bool HopcroftImpl::doReductions()
{
	// Reductions returns true when a reduction is applied or set flag when fail
	// -> loop until there is no available reduction
	if (reductionA() || m_fail)
		return true;
	if (reductionB() || m_fail)
		return true;
	if (reductionC() || m_fail)
		return true;
	return false;
}

bool HopcroftImpl::reductionA()
{
	// Removal of loops and degree one vertices
	if (removeLoops())
		return true;
	if (removeCorollas())
		return true;
	if (removeKnots())
		return true;
	if (removeDegreeOneVertices())
		return true;
	return false;
}

// A(i) Removing loops from loop vertices with non-loop edges
bool HopcroftImpl::removeLoops()
{
	std::vector<int> labels;
	std::vector<std::vector<int>> numbers; // 3-tuple
	std::vector<EdgeEnd*> loops_g1;
	std::vector<EdgeEnd*> loops_g2;

	PREPARE_FUNCTION(Graph *graph, ARG_REF(loops_g)) {
		for (Vertex *vertex : graph->vertices) {
			if (vertex->isLoopVertex() && vertex->hasNonLoopEdge()) {
				vertex->forEachEdge([&](EdgeEnd *edge) {
					if (edge->isLoopEdge() && (edge->cwedge == edge)) {
						EdgeEnd *f = edge->otheredge;
						EdgeEnd *e = f->cwedge;
						loops_g.push_back(f);
						numbers.push_back({e->otheredge->elabel, f->elabel, f->otheredge->elabel});
					}
					return true;
				});
			}
		}
	};

	APPLY_FUNCTION(Graph *graph, ARG_CREF(loops_g)) {
		const std::vector<int> &labels_g = GET_LABELS(labels);

		for (size_t i = 0; i < labels_g.size(); ++i)
			loops_g[i]->cwedge->otheredge->elabel = labels_g[i];

		for (size_t i = 0; i < labels_g.size(); ++i)
			graph->removeEdge(loops_g[i]);
	};

	PREPARE_REDUCTIONS2(m_g, loops_g);
	CHECK_REDUCTIONS(loops_g);
	H_DEBUG("Apply removeLoops");
	VECTOR_PARTITIONING(numbers, labels);
	APPLY_REDUCTIONS2(m_g, loops_g);
	return true;
}

// A(ii) Removing corollas
bool HopcroftImpl::removeCorollas()
{
	std::vector<int> labels;
	std::vector<std::vector<int>> numbers;
	std::vector<Vertex*> corollas_g1;
	std::vector<Vertex*> corollas_g2;

	PREPARE_FUNCTION(Graph *graph, ARG_REF(corollas_g)) {
		int ni = numbers.size();
		for (Vertex *vertex : graph->vertices) {
			if (vertex->isCorolla()) {
				numbers.resize(numbers.size() + 1);
				vertex->forEachEdge([&](EdgeEnd *e) {
					EdgeEnd *f = e->cwedge;
					if (f != e->otheredge) {
						e = e->otheredge;
						f = e->cwedge;
					}
					numbers[ni].push_back(vertex->vlabel);
					numbers[ni].push_back(e->elabel);
					numbers[ni].push_back(f->elabel);
					return true;
				});
				ni++;
				corollas_g.push_back(vertex);
			}
		}
	};

	APPLY_FUNCTION(Graph *graph, ARG_CREF(corollas_g)) {
		const std::vector<int> &labels_g = GET_LABELS(labels);
		for (size_t i = 0; i < labels_g.size(); ++i) {
			Vertex *corolla = corollas_g[i];
			while (corolla->vedge)
				graph->removeEdge(corolla->vedge, /* removeVertices */ false);
			corolla->vlabel = labels_g[i];
		}
	};

	PREPARE_REDUCTIONS2(m_g, corollas_g);
	CHECK_REDUCTIONS(corollas_g);
	H_DEBUG("Apply removeCorollas");
	CIRCLE_PARTITIONING(numbers, labels);
	APPLY_REDUCTIONS2(m_g, corollas_g);
	return true;
}

// A(iii) Removing knots
bool HopcroftImpl::removeKnots()
{
	std::vector<int> labels;
	std::vector<std::vector<int>> numbers;
	std::vector<Vertex*> knots_g1;
	std::vector<Vertex*> knots_g2;

	PREPARE_FUNCTION(Graph *graph, ARG_REF(knots_g)) {
		int ni = numbers.size();
		for (Vertex *vertex : graph->vertices) {
			if (vertex->isKnot()) {
				numbers.resize(numbers.size() + 1);

				numbers[ni].push_back(vertex->vlabel);
				numbers[ni].push_back(vertex->vedge->elabel);
				numbers[ni].push_back(vertex->vedge->otheredge->elabel);
				numbers[ni].push_back(vertex->vlabel);
				numbers[ni].push_back(vertex->vedge->otheredge->elabel);
				numbers[ni].push_back(vertex->vedge->elabel);

				ni++;
				knots_g.push_back(vertex);
			}
		}
	};

	APPLY_FUNCTION(Graph *graph, ARG_CREF(knots_g)) {
		const std::vector<int> &labels_g = GET_LABELS(labels);
		for (size_t i = 0; i < labels_g.size(); ++i) {
			Vertex *knot = knots_g[i];
			graph->removeEdge(knot->vedge, /* removeVertices */ false);
			knot->vlabel = labels_g[i];
		}
	};

	PREPARE_REDUCTIONS2(m_g, knots_g);
	CHECK_REDUCTIONS(knots_g);
	H_DEBUG("Apply removeKnots");
	CIRCLE_PARTITIONING(numbers, labels);
	APPLY_REDUCTIONS2(m_g, knots_g);
	return true;
}

// A(iv) Removing degree one vertices
bool HopcroftImpl::removeDegreeOneVertices()
{
	if (removeSpokeCenters() || m_fail)
		return true;
	if (removeStars() || m_fail)
		return true;
	if (removeDumbells() || m_fail)
		return true;
	return false;
}

bool HopcroftImpl::removeSpokeCenters()
{
	std::vector<int> labels;
	std::vector<std::vector<int>> numbers; // 3-tuple
	std::vector<EdgeEnd*> spokes_g1;
	std::vector<EdgeEnd*> spokes_g2;

	PREPARE_FUNCTION(Graph *graph, ARG_REF(spokes_g)) {
		for (Vertex *vertex : graph->vertices) {
			if (vertex->isSpokeCenter() && vertex->hasNonSpokeEdge()) {
				vertex->forEachEdge([&](EdgeEnd *edge) {
					if (edge->isSpoke()) {
						EdgeEnd *f = edge->otheredge;
						EdgeEnd *e = f->cwedge;
						spokes_g.push_back(f);
						numbers.push_back({e->otheredge->elabel, f->elabel, f->otheredge->elabel});
						H_ASSERT(!contains(spokes_g, f->otheredge));
					}
					return true;
				});
			}
		}
	};

	APPLY_FUNCTION(Graph *graph, ARG_CREF(spokes_g)) {
		const std::vector<int> &labels_g = GET_LABELS(labels);
		for (size_t i = 0; i < labels_g.size(); ++i)
			spokes_g[i]->cwedge->otheredge->elabel = labels_g[i];

		for (size_t i = 0; i < labels_g.size(); ++i)
			graph->removeEdge(spokes_g[i]);
	};

	PREPARE_REDUCTIONS2(m_g, spokes_g);
	CHECK_REDUCTIONS(spokes_g);
	H_DEBUG("Apply removeSpokeCenters");
	VECTOR_PARTITIONING(numbers, labels);
	APPLY_REDUCTIONS2(m_g, spokes_g);
	return true;
}

bool HopcroftImpl::removeStars()
{
	std::vector<int> labels;
	std::vector<std::vector<int>> numbers;
	std::vector<Vertex*> stars_g1;
	std::vector<Vertex*> stars_g2;

	PREPARE_FUNCTION(Graph *graph, ARG_REF(stars_g)) {
		int ni = numbers.size();
		for (Vertex *vertex : graph->vertices) {
			if (vertex->isStar()) {
				numbers.resize(numbers.size() + 1);
				vertex->forEachEdge([&](EdgeEnd *e) {
					EdgeEnd *f = e->cwedge;
					numbers[ni].push_back(vertex->vlabel);
					numbers[ni].push_back(e->elabel);
					numbers[ni].push_back(f->elabel);
					return true;
				});
				ni++;
				stars_g.push_back(vertex);
			}
		}
	};

	APPLY_FUNCTION(Graph *graph, ARG_CREF(stars_g)) {
		const std::vector<int> &labels_g = GET_LABELS(labels);
		for (size_t i = 0; i < labels_g.size(); ++i) {
			Vertex *star = stars_g[i];
			star->forEachEdge([=](EdgeEnd *e) {
				Vertex *v = e->isMultiEdge() ? nullptr : e->vertex;
				graph->removeEdge(e, /* removeVertices */ false);
				if (v)
					graph->removeVertex(v);
				return true;
			});
			star->vlabel = labels_g[i];
		}
	};

	PREPARE_REDUCTIONS2(m_g, stars_g);
	CHECK_REDUCTIONS(stars_g);
	H_DEBUG("Apply removeStars");
	CIRCLE_PARTITIONING(numbers, labels);
	APPLY_REDUCTIONS2(m_g, stars_g);
	return true;
}
bool HopcroftImpl::removeDumbells()
{
	std::vector<int> labels;
	std::vector<std::vector<int>> numbers;
	std::vector<Vertex*> dumbells_g1;
	std::vector<Vertex*> dumbells_g2;

	PREPARE_FUNCTION(Graph *graph, ARG_REF(dumbells_g)) {
		int ni = numbers.size();
		for (Vertex *vertex : graph->vertices) {
			if (vertex->isDumbell()) {
				numbers.resize(numbers.size() + 1);

				numbers[ni].push_back(vertex->vlabel);
				numbers[ni].push_back(vertex->vedge->elabel);
				numbers[ni].push_back(vertex->vedge->otheredge->elabel);
				numbers[ni].push_back(vertex->vlabel);
				numbers[ni].push_back(vertex->vedge->otheredge->elabel);
				numbers[ni].push_back(vertex->vedge->elabel);

				ni++;
				dumbells_g.push_back(vertex);
			}
		}
	};

	APPLY_FUNCTION(Graph *graph, ARG_CREF(dumbells_g)) {
		const std::vector<int> &labels_g = GET_LABELS(labels);
		for (size_t i = 0; i < labels_g.size(); ++i) {
			Vertex *dumbell = dumbells_g[i];
			if (dumbell->vedge) // eh ...
				graph->removeEdge(dumbell->vedge, /* removeVertices */ false);
			dumbell->vlabel = labels_g[i];
			if (graph->vertices.size() > 1)
				graph->removeVertex(dumbell);
		}
	};

	PREPARE_REDUCTIONS2(m_g, dumbells_g);
	CHECK_REDUCTIONS(dumbells_g);
	H_DEBUG("Apply removeDumbells");
	CIRCLE_PARTITIONING(numbers, labels);
	APPLY_REDUCTIONS2(m_g, dumbells_g);
	return true;
}

bool HopcroftImpl::reductionB()
{
	// Bond associated reductions
	if (replaceClumps())
		return true;
	if (replaceSkeins())
		return true;
	return false;
}

// B(i) Replacing clumps
bool HopcroftImpl::replaceClumps()
{
	std::vector<int> labels;
	std::vector<std::vector<int>> numbers;
	std::vector<std::vector<EdgeEnd*>> clumps_g1;
	std::vector<std::vector<EdgeEnd*>> clumps_g2;
	std::vector<std::pair<Vertex*, Vertex*>> clumpvertices_g1;
	std::vector<std::pair<Vertex*, Vertex*>> clumpvertices_g2;

	PREPARE_FUNCTION(Graph *graph, ARG_REF(clumps_g), ARG_REF(clumpvertices_g)) {
		int ni = numbers.size();
		for (EdgeEnd *edge : graph->edge_ends) {
			Vertex *v = edge->vertex;
			Vertex *w = edge->otheredge->vertex;

			if (v->degree < 2 && w->degree < 2)
				continue;

			if (contains(clumpvertices_g, std::make_pair(v, w)) || contains(clumpvertices_g, std::make_pair(w, v)))
				continue;

			const std::vector<std::vector<EdgeEnd*>> &clumps = graph->findClumps(v, w);
			if (clumps.empty())
				continue;

			clumps_g.insert(clumps_g.end(), clumps.begin(), clumps.end());
			clumpvertices_g.push_back(std::make_pair(v, w));

			for (const auto &clump : clumps) {
				numbers.resize(numbers.size() + 2);

				for (size_t i = 0; i < clump.size(); ++i)
					numbers[ni].push_back(clump[i]->otheredge->elabel);

				for (size_t i = clump.size(); i-- > 0; )
					numbers[ni + 1].push_back(clump[i]->elabel);

				ni += 2;
			}
		}
	};

	APPLY_FUNCTION(Graph *graph, ARG_CREF(clumps_g)) {
		const std::vector<int> &labels_g = GET_LABELS(labels);
		for (size_t i = 0; i < labels_g.size(); i += 2) {
			const auto &clump = clumps_g[i / 2];
			for (size_t j = 0; j < clump.size() - 1; ++j)
				graph->removeEdge(clump[j]);

			EdgeEnd *last = clump.back();
			last->elabel = labels_g[i];
			last->otheredge->elabel = labels_g[i + 1];
		}
	};

	PREPARE_REDUCTIONS3(m_g, clumps_g, clumpvertices_g);
	CHECK_REDUCTIONS(clumps_g);
	H_DEBUG("Apply replaceClumps");
	CIRCLE_PARTITIONING(numbers, labels); // FIXME: Vector in paper!
	APPLY_REDUCTIONS2(m_g, clumps_g);
	return true;
}

// B(ii) Replacing skeins
bool HopcroftImpl::replaceSkeins()
{
	std::vector<int> labels;
	std::vector<std::vector<int>> numbers;
	std::vector<std::pair<Vertex*, Vertex*>> skeinvertices_g1;
	std::vector<std::pair<Vertex*, Vertex*>> skeinvertices_g2;

	PREPARE_FUNCTION(Graph *graph, ARG_REF(skeinvertices_g)) {
		int ni = numbers.size();
		for (Vertex *vertex : graph->vertices) {
			if (!vertex->vedge)
				continue;

			Vertex *v = vertex;
			Vertex *w = vertex->vedge->vertex;

			if (v == w || (v->degree != 1 && w->degree != 1))
				continue;

			if (contains(skeinvertices_g, std::make_pair(v, w)) || contains(skeinvertices_g, std::make_pair(w, v)))
				continue;

			const std::vector<EdgeEnd*> &edges = v->neighborEdges();
			if (edges.size() < 2)
				continue;

			skeinvertices_g.push_back(std::make_pair(v, w));
			numbers.resize(numbers.size() + 2);

			for (size_t i = 0; i < edges.size(); ++i) {
				numbers[ni].push_back(v->vlabel);
				numbers[ni].push_back(edges[i]->otheredge->elabel);
				numbers[ni].push_back(edges[i]->elabel);
				numbers[ni].push_back(w->vlabel);
			}

			for (size_t i = edges.size(); i-- > 0; ) {
				numbers[ni + 1].push_back(w->vlabel);
				numbers[ni + 1].push_back(edges[i]->elabel);
				numbers[ni + 1].push_back(edges[i]->otheredge->elabel);
				numbers[ni + 1].push_back(v->vlabel);
			}

			ni += 2;
		}
	};

	APPLY_FUNCTION(Graph *graph, ARG_CREF(skeinvertices_g)) {
		const std::vector<int> &labels_g = GET_LABELS(labels);
		for (size_t i = 0; i < labels_g.size(); i += 2) {
			const auto &skein = skeinvertices_g[i / 2];
			skein.first->forEachEdge([=](EdgeEnd *e) {
				graph->removeEdge(e, /* removeVertices */ false);
				return true;
			});
			graph->removeVertex(skein.second);
			skein.first->vlabel = std::min(labels_g[i], labels_g[i + 1]);
		}
	};

	PREPARE_REDUCTIONS2(m_g, skeinvertices_g);
	CHECK_REDUCTIONS(skeinvertices_g);
	H_DEBUG("Apply replaceSkeins");
	CIRCLE_PARTITIONING(numbers, labels);
	APPLY_REDUCTIONS2(m_g, skeinvertices_g);
	return true;
}

bool HopcroftImpl::reductionC()
{
	// Four general classes of reductions and two special cases
	if (removeIsolatedLowDegreeVertices())
		return true;
	if (removeNonregularGraphsWithNoIsolatedLowDegreeVertices())
		return true;
	if (removeLowDegreeVerticesWithSolitaryFaces())
		return true;
	if (removeTheLastGeneralCase())
		return true;
	return false;
}

// C (i) Removing isolated low degre vertices
bool HopcroftImpl::removeIsolatedLowDegreeVertices()
{
	if (removeIsolatedLowDegree(2) || m_fail)
		return true;
	if (removeIsolatedLowDegree(3) || m_fail)
		return true;
	if (removeIsolatedLowDegree(4) || m_fail)
		return true;
	if (removeIsolatedLowDegree(5) || m_fail)
		return true;
	return false;
}

bool HopcroftImpl::removeIsolatedLowDegree(int degree)
{
	std::vector<int> labels;
	std::vector<std::vector<int>> numbers; // 3-tuple
	std::vector<std::vector<std::pair<EdgeEnd*, EdgeEnd*>>> addededges_g1;
	std::vector<std::vector<std::pair<EdgeEnd*, EdgeEnd*>>> addededges_g2;
	std::vector<Vertex*> isolatedvertices_g1;
	std::vector<Vertex*> isolatedvertices_g2;

	PREPARE_FUNCTION(Graph *graph, ARG_REF(addededges_g), ARG_REF(isolatedvertices_g)) {
		int ni = numbers.size();
		for (Vertex *vertex : graph->vertices) {
			if (vertex->degree == degree && vertex->isDegreeIsolated()) {
				std::vector<std::pair<EdgeEnd*, EdgeEnd*>> addededges;
				const std::vector<EdgeEnd*> &edges = vertex->neighborEdges();
				for (size_t i = 0; i < edges.size(); ++i) {
					EdgeEnd *e1 = edges[i];
					EdgeEnd *e2 = (i == edges.size() - 1) ? edges[0] : edges[i + 1];
					if (e1->vertex == e2->vertex) // eh ...
						continue;

					numbers.resize(numbers.size() + 2);

					numbers[ni] = {vertex->vlabel, e1->otheredge->elabel, e1->elabel};
					numbers[ni + 1] = {e2->elabel, e2->otheredge->elabel, vertex->vlabel};
					addededges.push_back(std::make_pair(e1, e2));

					ni += 2;
				}
				addededges_g.push_back(addededges);
				isolatedvertices_g.push_back(vertex);
			}
		}
	};

	APPLY_FUNCTION(Graph *graph, ARG_CREF(addededges_g), ARG_CREF(isolatedvertices_g)) {
		const std::vector<int> &labels_g = GET_LABELS(labels);
		int ni = 0;
		for (size_t i = 0; i < isolatedvertices_g.size(); ++i) {
			for (size_t j = 0; j < addededges_g[i].size(); ++j) {
				EdgeEnd *e1 = addededges_g[i][j].first;
				EdgeEnd *e2 = addededges_g[i][j].second;
				EdgeEnd *enew = graph->addInnerEdge(e1, e2);
				enew->elabel = labels_g[ni++];
				enew->otheredge->elabel = labels_g[ni++];
			}
		}

		for (Vertex *vertex : isolatedvertices_g) {
			vertex->forEachEdge([=](EdgeEnd *e) {
				graph->removeEdge(e, /* removeVertices */ false);
				return true;
			});
			graph->removeVertex(vertex);
		}
	};

	PREPARE_REDUCTIONS3(m_g, addededges_g, isolatedvertices_g);
	CHECK_REDUCTIONS(addededges_g);
	H_DEBUG("Apply removeIsolatedLowDegree " << degree);
	VECTOR_PARTITIONING(numbers, labels);
	APPLY_REDUCTIONS3(m_g, addededges_g, isolatedvertices_g);
	return true;
}

static bool isFourDegreeException(const std::vector<EdgeEnd*> &nghbrs)
{
	if (nghbrs.size() != 4)
		return false;

	// 4, x, 4, y
	if (nghbrs[0]->vertex->degree == 4)
		return nghbrs[1]->vertex->degree != 4
			&& nghbrs[2]->vertex->degree == 4
			&& nghbrs[3]->vertex->degree != 4;

	// x, 4, y, 4
	return nghbrs[1]->vertex->degree == 4
		&& nghbrs[2]->vertex->degree != 4
		&& nghbrs[3]->vertex->degree == 4;
}

// C (ii) Handling nonregular graphs with no isolated vertices of low degree
bool HopcroftImpl::removeNonregularGraphsWithNoIsolatedLowDegreeVertices()
{
	if (removeIsolatedDegree(2) || m_fail)
		return true;
	if (removeIsolatedDegree(3) || m_fail)
		return true;
	if (removeIsolatedDegree(4) || m_fail)
		return true;
	if (removeIsolatedDegree(5) || m_fail)
		return true;
	return false;
}

bool HopcroftImpl::removeIsolatedDegree(int degree)
{
	std::vector<int> labels;
	std::vector<std::vector<int>> numbers; // 4-tuple
	std::vector<EdgeEnd*> distinguishededges_g1;
	std::vector<EdgeEnd*> distinguishededges_g2;

	// We need to make sure to choose the same distinguished edge in both graphs
	// This is just too complex and doesn't even work in case the G1 is CW-ordered
	// and G2 is CCW-ordered (probably also in some other).
	//
	// There really is not much in paper as how to find the distinguished edge...
#define H_REMOVE_ISOLATED_DEGREE_HEURISTIC 1

#if H_REMOVE_ISOLATED_DEGREE_HEURISTIC
	struct CacheItem {
		int label;
		int rlabel;
		int degree;
		std::vector<int> circle;
	};
	std::vector<CacheItem> distinguishedCache;

	auto makeCircle = [](const std::vector<EdgeEnd*> &edges) {
		std::vector<int> circle;
		for (EdgeEnd *e : edges) {
			circle.push_back(e->vertex->degree);
			circle.push_back(e->otheredge->elabel);
			circle.push_back(e->elabel);
			circle.push_back(e->vertex->degree);
		}
		return circle;
	};

	auto degreeFromCache = [&](const std::vector<EdgeEnd*> &edges) {
		const std::vector<int> &circle = makeCircle(edges);
		std::vector<int> rcircle(circle.begin(), circle.end());
		std::reverse(rcircle.begin(), rcircle.end());
		for (const auto &c : distinguishedCache) {
			if (circleIsomorphism(c.circle, circle))
				return std::make_pair(c.label, c.degree);
			if (circleIsomorphism(c.circle, rcircle))
				return std::make_pair(c.rlabel, c.degree);
		}
		return std::make_pair(-1, -1);
	};
#endif

	PREPARE_FUNCTION(Graph *graph, ARG_REF(distinguishededges_g)) {
		for (Vertex *v : graph->vertices) {
			if (v->degree != degree)
				continue;

			const std::vector<EdgeEnd*> &nghbrs = v->neighborEdges(CounterClockWise);
			if (isFourDegreeException(nghbrs))
				continue;

#if H_REMOVE_ISOLATED_DEGREE_HEURISTIC
			EdgeEnd *distig = nullptr;
			const auto &cached = degreeFromCache(nghbrs);
			if (cached.first != -1) {
				bool afterSame = false;
				for (EdgeEnd *e : nghbrs) {
					if (!distig && e->vertex->degree == cached.second && e->elabel == cached.first)
						distig = e;
					if (e->vertex->degree == degree)
						afterSame = true;
					else if (afterSame && e->vertex->degree == cached.second && e->elabel == cached.first) {
						distig = e;
						break;
					}
				}
			}

			if (!distig)  {
				bool sameDegree = false;
				for (EdgeEnd *e : nghbrs) {
					if (e->vertex->degree == degree)
						sameDegree = true;
					else if (e->vertex->degree != degree)
						distig = e;
				}

				if (!sameDegree || !distig)
					distig = nullptr;
				else
					distinguishedCache.push_back({distig->elabel, distig->otheredge->elabel, distig->vertex->degree, makeCircle(nghbrs)});
			}

			if (!distig)
				continue;
#else

			bool sameDegree = false;
			EdgeEnd *distig = nullptr;
			for (EdgeEnd *e : nghbrs) {
				if (e->vertex->degree == degree) {
					sameDegree = true;
				} else if (e->vertex->degree != degree) {
					distig = e;
					break;
				}
			}

			if (!sameDegree || !distig)
				continue;
#endif

			EdgeEnd *inc = distig->otheredge->cwedge;
			numbers.push_back({distig->elabel, distig->otheredge->elabel, v->vlabel, inc->otheredge->elabel});
			distinguishededges_g.push_back(distig);
			H_ASSERT(!contains(distinguishededges_g, distig->otheredge));
		}
	};

	APPLY_FUNCTION(Graph *graph, ARG_CREF(distinguishededges_g)) {
		const std::vector<int> &labels_g = GET_LABELS(labels);
		for (size_t i = 0; i < labels_g.size(); ++i) {
			EdgeEnd *distig = distinguishededges_g[i];
			distig->otheredge->cwedge->otheredge->elabel = labels_g[i];
			graph->removeDistinguishedEdgeAndVertex(distig);
		}
	};

	PREPARE_REDUCTIONS2(m_g, distinguishededges_g);
	CHECK_REDUCTIONS(distinguishededges_g);
	H_DEBUG("Apply removeIsolatedDegree " << degree);
	VECTOR_PARTITIONING(numbers, labels);
	APPLY_REDUCTIONS2(m_g, distinguishededges_g);
	return true;
}

// C (iii) The first exception
bool HopcroftImpl::removeFourDegreeException()
{
	std::vector<int> labels;
	std::vector<std::vector<int>> numbers; // 3-tuple
	std::vector<std::pair<Vertex*, EdgeEnd*>> circles_g1;
	std::vector<std::pair<Vertex*, EdgeEnd*>> circles_g2;

	PREPARE_FUNCTION(Graph *graph, ARG_REF(circles_g)) {
		for (Vertex *v : graph->vertices) {
			if (v->degree != 4)
				continue;

			const std::vector<EdgeEnd*> &nghbrs = v->neighborEdges(CounterClockWise);
			if (!isFourDegreeException(nghbrs))
				continue;

			int s = nghbrs[0]->vertex->degree == 4 ? 0 : 1;
			circles_g.push_back(std::make_pair(v, nghbrs[s]));
			H_ASSERT(!contains(circles_g, std::make_pair(nghbrs[s]->vertex, nghbrs[s]->otheredge)));

			EdgeEnd *er = nghbrs[s];
			EdgeEnd *es = nghbrs[s + 1];
			EdgeEnd *et = nghbrs[s + 2];
			EdgeEnd *eu = nghbrs[s + 3];

			numbers.push_back({v->vlabel, es->otheredge->elabel, er->otheredge->elabel});
			numbers.push_back({v->vlabel, eu->otheredge->elabel, et->otheredge->elabel});
		}
	};

	APPLY_FUNCTION(Graph *graph, ARG_CREF(circles_g)) {
		const std::vector<int> &labels_g = GET_LABELS(labels);
		for (size_t i = 0; i < labels_g.size(); i += 2) {
			const auto &circle = circles_g[i / 2];
			graph->removeFourDegreeExceptionVertex(circle.first, circle.second);
			circle.second->otheredge->elabel = labels_g[i];
			circle.second->otheredge->cwedge->otheredge->cwedge->otheredge->elabel = labels_g[i + 1];
		}
	};

	PREPARE_REDUCTIONS2(m_g, circles_g);
	CHECK_REDUCTIONS(circles_g);
	H_DEBUG("Apply fourDegreeException");
	VECTOR_PARTITIONING(numbers, labels);
	APPLY_REDUCTIONS2(m_g, circles_g);
	return true;
}

// C (iv) Handling regular graphs with vertices with solitary faces
bool HopcroftImpl::removeLowDegreeVerticesWithSolitaryFaces()
{
	// d = 2  - special case (handled in testByDefinition)
	if (removeVertexWithSolitaryFace(3) || m_fail)
		return true;
	if (removeVertexWithSolitaryFace(4) || m_fail)
		return true;
	if (removeVertexWithSolitaryFace(5) || m_fail)
		return true;
	// more than five face sides
	if (removeVertexWithSolitaryFace(SPECIAL_X) || m_fail)
		return true;
	return false;
}

static int specialFaceLabel(Face *f)
{
	H_ASSERT(f->edges > 2);
	return f->edges > 5 ? SPECIAL_X : f->edges;
}

// C (iv) Handling regular graphs with vertices with solitary faces
bool HopcroftImpl::removeVertexWithSolitaryFace(int sides)
{
	std::vector<int> labels;
	std::vector<std::vector<int>> numbers; // 4-tuple
	std::vector<std::vector<EdgeEnd*>> trees_g1;
	std::vector<std::vector<EdgeEnd*>> trees_g2;
	std::vector<std::vector<EdgeEnd*>> circles_g1;
	std::vector<std::vector<EdgeEnd*>> circles_g2;

	PREPARE_FUNCTION(Graph *graph, ARG_REF(trees_g), ARG_REF(circles_g)) {
		// Find distinguished edges + vertices
		std::unordered_map<EdgeEnd*, EdgeEnd*> omap;
		std::unordered_map<Vertex*, std::vector<EdgeEnd*>> rmap;
		for (Vertex *v : graph->vertices) {
			if (v->degree < 3)
				continue;

			// 4 possible face labels (3, 4, 5, x)
			std::vector<int> counts(4, 0);
			std::vector<EdgeEnd*> edges(4, nullptr);

			v->forEachEdge([&](EdgeEnd *e) {
				int idx = specialFaceLabel(e->ccwface);
				idx = idx == SPECIAL_X ? 3 : idx - 3;
				counts[idx]++;
				edges[idx] = e->otheredge;
				return true;
			});

			EdgeEnd *distig = nullptr;
			for (size_t i = 0; i < 4; ++i)
				if (counts[i] == 1 && specialFaceLabel(edges[i]->otheredge->ccwface) == sides)
					distig = edges[i];

			if (!distig)
				continue;

			rmap[v].push_back(distig);
			rmap[distig->otheredge->vertex].push_back(distig->otheredge);

			omap[distig] = distig;
			omap[distig->otheredge] = distig;
		}

		if (rmap.empty())
			return;

		auto nextEdge = [](const std::vector<EdgeEnd*> &edges, Vertex *v, Vertex *w) {
			EdgeEnd *next;
			if (edges.front()->vertex == v || edges.front()->otheredge->vertex == v)
				next = edges.back();
			else
				next = edges.front();
			if (next->vertex == w)
				next = next->otheredge;
			return next;
		};

		// Build trees
		bool found;
		do {
			found = false;
			for (const auto &p : rmap) {
				if (p.second.size() != 1)
					continue;

				std::vector<EdgeEnd*> tree;
				if (p.second.front()->vertex == p.first)
					tree.push_back(p.second.front()->otheredge);
				else
					tree.push_back(p.second.front());
				rmap.erase(p.first);

				Vertex *v = tree.back()->vertex;
				while (rmap.find(v) != rmap.end()) {
					const std::vector<EdgeEnd*> &edges = rmap[v];
					if (edges.size() == 1) {
						rmap.erase(v);
						break;
					}
					tree.push_back(nextEdge(edges, tree.back()->otheredge->vertex, v));
					rmap.erase(v);
					v = tree.back()->vertex;
				}
				found = true;

				// Use the original distinguished edge-end
				for (size_t i = 0; i < tree.size(); ++i)
					tree[i] = omap[tree[i]];

				trees_g.push_back(tree);
				break;
			}
		} while (found);

		// Build tree vectors
		for (std::vector<EdgeEnd*> &edges : trees_g) {
			for (size_t i = 0; i < edges.size(); i++) {
				EdgeEnd *distig = edges[i];
				EdgeEnd *distig2 = distig->otheredge;
				EdgeEnd *inc = distig2->otheredge->cwedge;
				Vertex *v = distig2->otheredge->vertex;
				numbers.push_back({distig2->elabel, distig2->otheredge->elabel, v->vlabel, inc->otheredge->elabel});
			}
		}

		// Build circles
		while (!rmap.empty()) {
			for (const auto &p : rmap) {
				std::vector<EdgeEnd*> circle;
				circle.push_back(p.second.front()->otheredge);
				rmap.erase(p.first);

				Vertex *v = circle.back()->vertex;
				while (v != circle.front()->otheredge->vertex) {
					const std::vector<EdgeEnd*> &edges = rmap[v];
					circle.push_back(nextEdge(edges, circle.back()->otheredge->vertex, v));
					rmap.erase(v);
					v = circle.back()->vertex;
				}

				circles_g.push_back(circle);
				break;
			}
		};
	};

	APPLY_FUNCTION(Graph *graph, ARG_CREF(trees_g), ARG_CREF(circles_g)) {
		const std::vector<int> &labels_g = GET_LABELS(labels);
		// Trees
		std::vector<EdgeEnd*> distigs;
		for (const std::vector<EdgeEnd*> &edges : trees_g)
			for (EdgeEnd *distig : edges)
				distigs.push_back(distig);

		for (size_t i = 0; i < labels_g.size(); ++i)
			distigs[i]->otheredge->cwedge->otheredge->elabel = labels_g[i];

		for (size_t i = 0; i < labels_g.size(); ++i)
			graph->removeDistinguishedEdgeAndVertex(distigs[i]->otheredge);

		// Circles
		int ulabel = newInteger();
		for (const std::vector<EdgeEnd*> edges : circles_g) {
			for (size_t i = 0; i < edges.size() - 1; ++i) {
				EdgeEnd *distig = edges[i];
				// Need to remove multi-edge with 2 last vertices of circle
				if (distig->ccwedge->vertex == distig->otheredge->vertex) {
					distig->vertex->vlabel = ulabel;
					graph->removeEdge(distig->ccwedge);
				} else if (distig->cwedge->vertex == distig->otheredge->vertex) {
					distig->vertex->vlabel = ulabel;
					graph->removeEdge(distig->cwedge);
				}
				graph->removeDistinguishedEdgeAndVertex(distig);
			}
		}
	};

	PREPARE_REDUCTIONS3(m_g, trees_g, circles_g);
	CHECK_REDUCTIONS2(circles_g, trees_g);
	H_DEBUG("Apply removeVertexWithSolitaryFace " << sides);
	VECTOR_PARTITIONING(numbers, labels);
	APPLY_REDUCTIONS3(m_g, trees_g, circles_g);
	return true;
}

static int anotherSpecialFaceLabel(Face *f)
{
	H_ASSERT(f->edges > 2);
	return f->edges > 3 ? SPECIAL_X : f->edges;
}

static bool isAnotherFourDegreeException(const std::vector<EdgeEnd*> &nghbrs)
{
	if (nghbrs.size() != 4)
		return false;

	int l1 = anotherSpecialFaceLabel(nghbrs[0]->ccwface);
	int l2 = anotherSpecialFaceLabel(nghbrs[1]->ccwface);
	int l3 = anotherSpecialFaceLabel(nghbrs[2]->ccwface);
	int l4 = anotherSpecialFaceLabel(nghbrs[3]->ccwface);

	// 3, x, 3, x
	if (l1 == 3)
		return l2 == SPECIAL_X && l3 == 3 && l4 == SPECIAL_X;

	// x, 3, x, 3
	return l2 == 3 && l3 == SPECIAL_X && l4 == 3;
}

// C (v) The last general case
bool HopcroftImpl::removeTheLastGeneralCase()
{
	std::vector<int> labels;
	std::vector<std::vector<int>> numbers; // 4-tuple
	std::vector<EdgeEnd*> distinguishededges_g1;
	std::vector<EdgeEnd*> distinguishededges_g2;

	PREPARE_FUNCTION(Graph *graph, ARG_REF(distinguishededges_g)) {
		for (Vertex *v : graph->vertices) {
			if (v->degree < 4)
				continue;

			const std::vector<EdgeEnd*> &nghbrs = v->neighborEdges();
			if (isAnotherFourDegreeException(nghbrs))
				continue;

			// 2 possible face labels (3, x)
			std::vector<int> counts(2, 0);
			std::vector<EdgeEnd*> edges(2, nullptr);

			for (EdgeEnd *e : nghbrs) {
				int idx = anotherSpecialFaceLabel(e->ccwface);
				idx = idx == SPECIAL_X ? 1 : 0;
				counts[idx]++;
				edges[idx] = e->otheredge;
			}

			EdgeEnd *distig = nullptr;
			if (counts[0] == 2 && counts[1] != 2)
				distig = edges[0];
			else if (counts[0] > 0 && counts[0] != 2 && counts[1] == 2)
				distig = edges[1];

			if (!distig)
				continue;

			EdgeEnd *inci = distig->otheredge->cwedge;
			numbers.push_back({distig->elabel, distig->otheredge->elabel, v->vlabel, inci->otheredge->elabel});
			distinguishededges_g.push_back(distig);
		}
	};

	APPLY_FUNCTION(Graph *graph, ARG_CREF(distinguishededges_g)) {
		const std::vector<int> &labels_g = GET_LABELS(labels);
		for (size_t i = 0; i < labels_g.size(); ++i) {
			int label = labels_g[i];
			EdgeEnd *distig = distinguishededges_g[i];
			EdgeEnd *distigother = distig->otheredge;
			int otherpos = indexOf(distinguishededges_g, distigother);
			if (otherpos == -1) {
				distig->otheredge->cwedge->otheredge->elabel = label;
				graph->removeDistinguishedEdgeAndVertex(distig);
			} else if (otherpos < int(i)) {
				// Special case
				int labelother = labels_g[otherpos];
				distig->otheredge->cwedge->otheredge->elabel = label;
				distigother->otheredge->cwedge->otheredge->elabel = labelother;
				distig->vertex->vlabel = newInteger();
				graph->removeDistinguishedEdgeAndVertex(distig);
			}
		}
	};

	PREPARE_REDUCTIONS2(m_g, distinguishededges_g);
	CHECK_REDUCTIONS(distinguishededges_g);
	H_DEBUG("Apply removeTheLastGeneralCase");
	VECTOR_PARTITIONING(numbers, labels);
	APPLY_REDUCTIONS2(m_g, distinguishededges_g);
	return true;
}

// C (vi) Another special degree 4 case
bool HopcroftImpl::removeAnotherSpecialDegreeFourCase()
{
	std::vector<EdgeEnd*> distinguishededges_g1;
	std::vector<EdgeEnd*> distinguishededges_g2;

	PREPARE_FUNCTION(Graph *graph, ARG_REF(distinguishededges_g)) {
		for (Vertex *v : graph->vertices) {
			if (v->degree != 4)
				continue;

			const std::vector<EdgeEnd*> &nghbrs = v->neighborEdges();
			if (!isAnotherFourDegreeException(nghbrs))
				continue;

			for (EdgeEnd *e : nghbrs) {
				if (e->ccwface->edges == 4) {
					distinguishededges_g.push_back(e);
				}
			}
		}
	};

	APPLY_FUNCTION(Graph *graph, ARG_CREF(distinguishededges_g)) {
		for (EdgeEnd *e : distinguishededges_g) {
			graph->addAndReconnectVertexIntoTriangularFace(e);
			// Should the edges be relabelled?
		}
	};

	PREPARE_REDUCTIONS2(m_g, distinguishededges_g);
	CHECK_REDUCTIONS(distinguishededges_g);
	H_DEBUG("Apply removeAnotherSpecialDegreeFourCase");
	APPLY_REDUCTIONS2(m_g, distinguishededges_g);
	return true;
}

bool HopcroftImpl::testByDefinition()
{
	H_ASSERT(!m_g1->vertices.empty());
	H_ASSERT(!m_g2->vertices.empty());

	if (isRegularDegreeTwoGraph())
		return regularDegreeTwoGraphIsomorphism();

	if (m_g1->vertices.size() == 1)
		return m_g1->vertices.front()->vlabel == m_g2->vertices.front()->vlabel;

	std::unordered_map<Vertex*, Vertex*> mapping;
	return isomorphism_r(m_g1->vertices, m_g2->vertices, mapping);
}

bool HopcroftImpl::isRegularDegreeTwoGraph()
{
	for (Vertex *v : m_g1->vertices)
		if (v->degree != 2)
			return false;

	return true;
}

// [C(iv)] Regular degree two graphs are n-gons which are handled by applying
// the circle isomorphism directly
//
//  * Note that two circles (one clockwise, one counterclockwise) have to be
//    formed for each n-gon in an analogous manner to skeins (its dual case).
bool HopcroftImpl::regularDegreeTwoGraphIsomorphism()
{
	auto createCircle = [](Graph *graph, Direction d) {
		std::vector<int> circle;
		EdgeEnd *s = graph->edge_ends.front();
		if (d == CounterClockWise)
			s = s->otheredge;
		EdgeEnd *e = s;
		do {
			circle.push_back(e->vertex->vlabel);
			circle.push_back(e->elabel);
			e = e->cwedge;
		} while (e != s);
		return circle;
	};

	const std::vector<int> &cw1 = createCircle(m_g1, ClockWise);
	const std::vector<int> &ccw1 = createCircle(m_g1, CounterClockWise);

	const std::vector<int> &cw2 = createCircle(m_g2, ClockWise);
	const std::vector<int> &ccw2 = createCircle(m_g2, CounterClockWise);

	if (circleIsomorphism(cw1, cw2) && circleIsomorphism(ccw1, ccw2))
		return true;

	if (circleIsomorphism(cw1, ccw2) && circleIsomorphism(ccw1, cw2))
		return true;

	return false;
}

// Recursive exhaustive matching isomorphism test
bool HopcroftImpl::isomorphism_r(std::vector<Vertex*> vertices1, std::vector<Vertex*> vertices2, std::unordered_map<Vertex*, Vertex*> &mapping)
{
	if (vertices1.empty())
		return isomorphism_mapping(mapping);

	std::vector<Vertex*> vrtcs1 = vertices1;
	Vertex *vertex = vertices1.back();
	vrtcs1.pop_back();

	for (size_t i = 0; i < vertices2.size(); ++i) {
		std::vector<Vertex*> vrtcs2 = vertices2;
		mapping.erase(vertex);
		mapping.insert({vertex, vertices2[i]});
		vrtcs2.erase(vrtcs2.begin() + i);
		if (isomorphism_r(vrtcs1, vrtcs2, mapping))
			return true;
	}

	return false;
}

bool HopcroftImpl::isomorphism_mapping(std::unordered_map<Vertex*, Vertex*> &mapping)
{
	auto createCircle = [](Vertex *v, Direction d) {
		std::vector<int> circle;
		v->forEachEdge([&](EdgeEnd *e) {
			circle.push_back(e->elabel);
			return true;
		}, d);
		return circle;
	};

	int cmp = 0; // 1 = cw-cw, 2 = cw-ccw

	for (Vertex *v : m_g1->vertices) {
		const std::vector<int> &cw1 = createCircle(v, ClockWise);
		const std::vector<int> &ccw1 = createCircle(v, CounterClockWise);

		const std::vector<int> &cw2 = createCircle(mapping[v], ClockWise);
		const std::vector<int> &ccw2 = createCircle(mapping[v], CounterClockWise);

		bool cwcw = circleIsomorphism(cw1, cw2) && circleIsomorphism(ccw1, ccw2);
		bool cwccw = circleIsomorphism(cw1, ccw2) && circleIsomorphism(ccw1, cw2);

		if (cmp == 1 && !cwcw)
			return false;
		else if (cmp == 2 && !cwccw)
			return false;
		else if (cmp == 0 && cwcw)
			cmp = 1;
		else if (cmp == 0 && cwccw)
			cmp = 2;
		else if (cmp == 0)
			return false;
	}

	return true;
}

void HopcroftImpl::dumpGraphs()
{
	std::cout << "G1:";
	m_g1->dump();

	std::cout << "G2:";
	m_g2->dump();
}

void HopcroftImpl::dumpLabels()
{
	std::cout << "G1:";
	m_g1->dumpLabels();

	std::cout << "G2:";
	m_g2->dumpLabels();
}

} // namespace hopcroft

} // namespace isomorphism

} // namespace graph
