#include "JarnikPrim.h"

#include <limits>

namespace graph {

namespace spanningtree {

class Data {
public:
	Node start;
	Graph *out;
};

class Value {
public:
	Value(const Node &node, int value)
		: node(node)
		, value(value)
	{
	}

	bool operator<(const Value &other) const
	{
		if (value == other.value) {
			return node < other.node;
		}
		return value < other.value;
	}

	Node node;
	int value;
};

typedef std::unordered_map<Node, std::unordered_map<Node, int>> weights_t;

static weights_t weights(const UndirectedGraph &g) {
	int val;
	weights_t w;

	for (const UndirectedEdge &e : g.getEdges()) {
		auto &map = w[e.getFirstNode()];
		auto search = map.find(e.getSecondNode());

		if (search == map.end()) {
			val = g.getEdgeValue(e);
			map[e.getSecondNode()] = val;
			w[e.getSecondNode()][e.getFirstNode()] = val;
		} else {
			val = std::min(search->second, g.getEdgeValue(e));
			search->second = val;
			w[e.getSecondNode()][e.getFirstNode()] = val;
		}
	}

	return w;
}

static UndirectedGraph jarnikprim_impl(const UndirectedGraph &graph, const Node &start) {
	UndirectedGraph res;              // spanning tree
	std::set<Value> q;                // priority queue
	weights_t w = weights(graph);     // minimum weights
	std::unordered_map<Node, int> d;  // distances
	std::unordered_map<Node, Node> p; // predecessors

	for (const Node &n : graph.getNodes()) {
		int value = n == start ? 0 : std::numeric_limits<int>::max();
		q.insert(Value(n, value));
		d[n] = value;
	}

	while (!q.empty()) {
		Node u = q.begin()->node; q.erase(q.begin());
		auto search = p.find(u);
		if (search != p.end()) {
			res.addEdge(UndirectedEdge(search->second, u), w[search->second][u]);
		}

		for (const Node &v : graph.neighbors(u)) {
			int &dist = d[v];
			if (q.find(Value(v, dist)) != q.end() && w[u][v] < dist) {
				q.erase(q.find(Value(v, dist)));

				if(p.find(v) == p.end()) p.insert(std::make_pair(v, u));
				else p.find(v)->second = u;

				dist = w[u][v];
				q.insert(Value(v, dist));
			}
		}
	}

	return res;
}

Graph JarnikPrim::jarnikprim(const Graph &graph, const Node &start) {
	return dispatch(graph.getData(), start);
}

UndirectedGraph JarnikPrim::jarnikprim(const UndirectedGraph &graph, const Node &start) {
	return jarnikprim_impl(graph, start);
}

auto JarnikPrimUndirectedGraph = JarnikPrim::RegistratorWrapper<graph::UndirectedGraph, graph::UndirectedGraph>(JarnikPrim::jarnikprim);

} // namespace spanningtree

} // namespace graph
