#include "FloydWarshallTest.h"

#include "graph/shortestpath/FloydWarshall.h"

#include <exception/CommonException.h>

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( GraphFloydWarshallTest, "graph" );
CPPUNIT_TEST_SUITE_REGISTRATION( GraphFloydWarshallTest );

void GraphFloydWarshallTest::testSimple()
{
	// Common
	graph::shortestpath::FloydWarshall::Result res;
	graph::Node n1("n1");
	graph::Node n2("n2");
	graph::Node n3("n3");
	graph::Node n4("n4");
	graph::Node n5("n5");
	graph::Node n6("n6");

	// Directed
	graph::DirectedGraph dg;
	dg.addEdge(graph::DirectedEdge(n1, n2), 2);
	dg.addEdge(graph::DirectedEdge(n1, n3), 2);
	dg.addEdge(graph::DirectedEdge(n1, n4), 1);
	dg.addEdge(graph::DirectedEdge(n1, n5), 6);
	dg.addEdge(graph::DirectedEdge(n2, n5), 3);
	dg.addEdge(graph::DirectedEdge(n5, n6), 2);

	CPPUNIT_ASSERT_EQUAL(6, (int)dg.getNodes().size());

	res = graph::shortestpath::FloydWarshall::floydwarshall(dg);

	CPPUNIT_ASSERT_EQUAL(0, res[n1][n1]);
	CPPUNIT_ASSERT_EQUAL(2, res[n1][n2]);
	CPPUNIT_ASSERT_EQUAL(2, res[n1][n3]);
	CPPUNIT_ASSERT_EQUAL(1, res[n1][n4]);
	CPPUNIT_ASSERT_EQUAL(5, res[n1][n5]);
	CPPUNIT_ASSERT_EQUAL(7, res[n1][n6]);

	// Undirected
	graph::UndirectedGraph ug;
	ug.addEdge(graph::UndirectedEdge(n1, n2), 2);
	ug.addEdge(graph::UndirectedEdge(n1, n3), 2);
	ug.addEdge(graph::UndirectedEdge(n1, n4), 1);
	ug.addEdge(graph::UndirectedEdge(n1, n5), 6);
	ug.addEdge(graph::UndirectedEdge(n2, n5), 3);
	ug.addEdge(graph::UndirectedEdge(n5, n6), 2);

	res = graph::shortestpath::FloydWarshall::floydwarshall(ug);

	CPPUNIT_ASSERT_EQUAL(0, res[n1][n1]);
	CPPUNIT_ASSERT_EQUAL(2, res[n1][n2]);
	CPPUNIT_ASSERT_EQUAL(2, res[n1][n3]);
	CPPUNIT_ASSERT_EQUAL(1, res[n1][n4]);
	CPPUNIT_ASSERT_EQUAL(5, res[n1][n5]);
	CPPUNIT_ASSERT_EQUAL(7, res[n1][n6]);
}

void GraphFloydWarshallTest::testCycle()
{
	// Common
	graph::shortestpath::FloydWarshall::Result res;
	graph::Node n1("n1");
	graph::Node n2("n2");
	graph::Node n3("n3");
	graph::Node n4("n4");
	graph::Node n5("n5");
	graph::Node n6("n6");

	// Directed
	graph::DirectedGraph dg;
	dg.addEdge(graph::DirectedEdge(n1, n2), 2);
	dg.addEdge(graph::DirectedEdge(n1, n3), 2);
	dg.addEdge(graph::DirectedEdge(n1, n4), 1);
	dg.addEdge(graph::DirectedEdge(n1, n5), 6);
	dg.addEdge(graph::DirectedEdge(n2, n5), 3);
	dg.addEdge(graph::DirectedEdge(n5, n6), 2);
	dg.addEdge(graph::DirectedEdge(n6, n1), 1);

	res = graph::shortestpath::FloydWarshall::floydwarshall(dg);

	CPPUNIT_ASSERT_EQUAL(0, res[n1][n1]);
	CPPUNIT_ASSERT_EQUAL(2, res[n1][n2]);
	CPPUNIT_ASSERT_EQUAL(2, res[n1][n3]);
	CPPUNIT_ASSERT_EQUAL(1, res[n1][n4]);
	CPPUNIT_ASSERT_EQUAL(5, res[n1][n5]);
	CPPUNIT_ASSERT_EQUAL(7, res[n1][n6]);

	// Undirected
	graph::UndirectedGraph ug;
	ug.addEdge(graph::UndirectedEdge(n1, n2), 2);
	ug.addEdge(graph::UndirectedEdge(n1, n3), 2);
	ug.addEdge(graph::UndirectedEdge(n1, n4), 1);
	ug.addEdge(graph::UndirectedEdge(n1, n5), 6);
	ug.addEdge(graph::UndirectedEdge(n2, n5), 3);
	ug.addEdge(graph::UndirectedEdge(n5, n6), 2);
	ug.addEdge(graph::UndirectedEdge(n6, n1), 1);

	res = graph::shortestpath::FloydWarshall::floydwarshall(ug);

	CPPUNIT_ASSERT_EQUAL(0, res[n1][n1]);
	CPPUNIT_ASSERT_EQUAL(2, res[n1][n2]);
	CPPUNIT_ASSERT_EQUAL(2, res[n1][n3]);
	CPPUNIT_ASSERT_EQUAL(1, res[n1][n4]);
	CPPUNIT_ASSERT_EQUAL(3, res[n1][n5]);
	CPPUNIT_ASSERT_EQUAL(1, res[n1][n6]);
}

void GraphFloydWarshallTest::testMultiEdge()
{
	// Common
	graph::shortestpath::FloydWarshall::Result res;
	graph::Node n1("n1");
	graph::Node n2("n2");
	graph::Node n3("n3");
	graph::Node n4("n4");
	graph::Node n5("n5");
	graph::Node n6("n6");

	// Directed
	graph::DirectedGraph dg;
	dg.addEdge(graph::DirectedEdge(n1, n2), 2);
	dg.addEdge(graph::DirectedEdge(n1, n2, "multi"), 1);
	dg.addEdge(graph::DirectedEdge(n1, n3), 2);
	dg.addEdge(graph::DirectedEdge(n1, n4), 1);
	dg.addEdge(graph::DirectedEdge(n1, n5), 6);
	dg.addEdge(graph::DirectedEdge(n2, n5), 3);
	dg.addEdge(graph::DirectedEdge(n5, n6), 2);
	dg.addEdge(graph::DirectedEdge(n6, n1), 1);

	res = graph::shortestpath::FloydWarshall::floydwarshall(dg);

	CPPUNIT_ASSERT_EQUAL(0, res[n1][n1]);
	CPPUNIT_ASSERT_EQUAL(1, res[n1][n2]);
	CPPUNIT_ASSERT_EQUAL(2, res[n1][n3]);
	CPPUNIT_ASSERT_EQUAL(1, res[n1][n4]);
	CPPUNIT_ASSERT_EQUAL(4, res[n1][n5]);
	CPPUNIT_ASSERT_EQUAL(6, res[n1][n6]);

	// Undirected
	graph::UndirectedGraph ug;
	ug.addEdge(graph::UndirectedEdge(n1, n2), 2);
	ug.addEdge(graph::UndirectedEdge(n1, n2, "multi"), 1);
	ug.addEdge(graph::UndirectedEdge(n1, n3), 2);
	ug.addEdge(graph::UndirectedEdge(n1, n4), 1);
	ug.addEdge(graph::UndirectedEdge(n1, n5), 6);
	ug.addEdge(graph::UndirectedEdge(n2, n5), 3);
	ug.addEdge(graph::UndirectedEdge(n5, n6), 2);

	res = graph::shortestpath::FloydWarshall::floydwarshall(ug);

	CPPUNIT_ASSERT_EQUAL(0, res[n1][n1]);
	CPPUNIT_ASSERT_EQUAL(1, res[n1][n2]);
	CPPUNIT_ASSERT_EQUAL(2, res[n1][n3]);
	CPPUNIT_ASSERT_EQUAL(1, res[n1][n4]);
	CPPUNIT_ASSERT_EQUAL(4, res[n1][n5]);
	CPPUNIT_ASSERT_EQUAL(6, res[n1][n6]);
}

void GraphFloydWarshallTest::testNegativeValue()
{
	// Common
	graph::shortestpath::FloydWarshall::Result res;
	graph::Node n1("n1");
	graph::Node n2("n2");
	graph::Node n3("n3");
	graph::Node n4("n4");
	graph::Node n5("n5");
	graph::Node n6("n6");

	// Directed
	graph::DirectedGraph dg;
	dg.addEdge(graph::DirectedEdge(n1, n2), 5);
	dg.addEdge(graph::DirectedEdge(n1, n3), 1);
	dg.addEdge(graph::DirectedEdge(n1, n4), 2);
	dg.addEdge(graph::DirectedEdge(n1, n5), 5);
	dg.addEdge(graph::DirectedEdge(n2, n5), -1);
	dg.addEdge(graph::DirectedEdge(n5, n6), -2);

	res = graph::shortestpath::FloydWarshall::floydwarshall(dg);

	CPPUNIT_ASSERT_EQUAL(0, res[n1][n1]);
	CPPUNIT_ASSERT_EQUAL(5, res[n1][n2]);
	CPPUNIT_ASSERT_EQUAL(1, res[n1][n3]);
	CPPUNIT_ASSERT_EQUAL(2, res[n1][n4]);
	CPPUNIT_ASSERT_EQUAL(4, res[n1][n5]);
	CPPUNIT_ASSERT_EQUAL(2, res[n1][n6]);
}

void GraphFloydWarshallTest::testNegativeCycle()
{
	// Common
	bool exception;
	graph::Node n1("n1");
	graph::Node n2("n2");
	graph::Node n3("n3");
	graph::Node n4("n4");
	graph::Node n5("n5");
	graph::Node n6("n6");

	// Directed
	graph::DirectedGraph dg;
	dg.addEdge(graph::DirectedEdge(n1, n2), 5);
	dg.addEdge(graph::DirectedEdge(n1, n3), 1);
	dg.addEdge(graph::DirectedEdge(n1, n4), 2);
	dg.addEdge(graph::DirectedEdge(n1, n5), 5);
	dg.addEdge(graph::DirectedEdge(n2, n5), -1);
	dg.addEdge(graph::DirectedEdge(n5, n6), -2);
	dg.addEdge(graph::DirectedEdge(n6, n1), -3);

	exception = false;
	try {
		graph::shortestpath::FloydWarshall::floydwarshall(dg);
	} catch (exception::CommonException) {
		exception = true;
	}

	CPPUNIT_ASSERT_EQUAL(true, exception);

	// Undirected
	graph::UndirectedGraph ug;
	ug.addEdge(graph::UndirectedEdge(n1, n2), 5);
	ug.addEdge(graph::UndirectedEdge(n1, n3), 1);
	ug.addEdge(graph::UndirectedEdge(n1, n4), 2);
	ug.addEdge(graph::UndirectedEdge(n2, n5), -1);

	exception = false;
	try {
		graph::shortestpath::FloydWarshall::floydwarshall(ug);
	} catch (exception::CommonException) {
		exception = true;
	}

	CPPUNIT_ASSERT_EQUAL(true, exception);
}
