/*
 * XmlApi.hpp
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef XML_API_HPP_
#define XML_API_HPP_

#include <functional>
#include <deque>
#include <vector>
#include <map>
#include <string>
#include "sax/Token.h"
#include "sax/FromXMLParserHelper.h"

#include "object/ObjectBase.h"
#include "base/WrapperBase.hpp"
#include "exception/CommonException.h"

namespace alib {

template < typename T, typename Enable = void >
struct xmlApi { };

class xmlApiInputContext : public std::deque < sax::Token >::iterator {
	template < typename T, typename Enable >
	friend struct xmlApi;

	std::map < std::string, std::map < int, void * > > idToInstanceContexts;
	std::map < std::string, int > idToInstanceMaxContexts;

public:
	xmlApiInputContext ( std::deque < sax::Token >::iterator iter ) : std::deque < sax::Token >::iterator ( iter ) {
	}

	~xmlApiInputContext ( ) {
		for ( const std::pair < const std::string, std::map < int, void * > > context : idToInstanceContexts ) {
			std::function < void ( void * ) > & deleteCallback = deleteCallbacks ( ).find ( context.first )->second;

			for ( const std::pair < const int, void * > entry : context.second )
				deleteCallback ( entry.second );
		}
	}

	std::map < int, void * > & idToInstance ( const std::string & name ) {
		return idToInstanceContexts[name];
	}

	int & idToInstanceMax ( const std::string & name ) {
		if ( idToInstanceMaxContexts.find ( name ) == idToInstanceMaxContexts.end ( ) )
			idToInstanceMaxContexts[name] = 0;

		return idToInstanceMaxContexts[name];
	}

	static std::map < std::string, std::function < void ( void * ) > > & deleteCallbacks ( ) {
		static std::map < std::string, std::function < void ( void * ) > > res;

		return res;
	}

};

class xmlApiOutputContext : public std::deque < sax::Token > {
	template < typename T, typename Enable >
	friend struct xmlApi;

	std::map < std::string, std::map < void *, int > > instanceToIdContexts;
	std::map < std::string, int > instanceToIdMaxContexts;

public:
	std::map < void *, int > & instanceToId ( const std::string & name ) {
		return instanceToIdContexts[name];
	}

	int & instanceToIdMax ( const std::string & name ) {
		if ( instanceToIdMaxContexts.find ( name ) == instanceToIdMaxContexts.end ( ) )
			instanceToIdMaxContexts[name] = 0;

		return instanceToIdMaxContexts[name];
	}

};

template < typename Group >
struct xmlApi < Group, typename std::enable_if < std::is_base_of < WrapperBaseBase, Group >::value >::type > {
	class ParserRegisterBase {
	public:
		virtual Group parse ( std::deque < sax::Token >::iterator & input ) = 0;
	};

	static std::map < std::string, typename xmlApi < Group >::ParserRegisterBase * > & parseFunctions ( ) {
		static std::map < std::string, typename xmlApi < Group >::ParserRegisterBase * > res;

		return res;
	}

public:
	class InputContextDeleter {
	public:
		InputContextDeleter ( ) {
			std::function < void ( void * ) > deleteCallback = [] ( void * pointer ) { delete ( Group * ) pointer; };
			xmlApiInputContext::deleteCallbacks ( ).insert ( std::make_pair ( Group::XML_TAG_NAME_REF, deleteCallback ) );
		}

	};

	template < class Type >
	class ParserRegister : public ParserRegisterBase {
		std::function < Type ( std::deque < sax::Token >::iterator & ) > parseFunction;

	public:
		ParserRegister ( ) : parseFunction ( Type::parse ) {
			parseFunctions ( ).insert ( std::make_pair ( Type::XML_TAG_NAME, this ) );
		}

		virtual Group parse ( std::deque < sax::Token >::iterator & input ) {
			return Group ( parseFunction ( input ) );
		}

	};

	static Group parse ( std::deque < sax::Token >::iterator & data ) {
		xmlApiInputContext & input = ( xmlApiInputContext & ) data;
		const std::string & tagName = sax::FromXMLParserHelper::getTokenData ( input, sax::Token::TokenType::START_ELEMENT );

		if ( tagName == Group::XML_TAG_NAME_REF ) {
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, Group::XML_TAG_NAME_REF );
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ATTRIBUTE, "id" );
			int id = std::stoi ( sax::FromXMLParserHelper::popTokenData ( input, sax::Token::TokenType::CHARACTER ) );
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ATTRIBUTE, "id" );
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, Group::XML_TAG_NAME_REF );
			std::map < int, void * >::iterator elem = input.idToInstance ( Group::XML_TAG_NAME_REF ).find ( id );

			if ( elem == input.idToInstance ( Group::XML_TAG_NAME_REF ).end ( ) ) throw exception::CommonException ( "XML Inconsistent" );

			return * ( ( Group * ) elem->second );
		} else {
			typename std::map < std::string, ParserRegisterBase * >::iterator callback = parseFunctions ( ).find ( tagName );

			if ( callback == parseFunctions ( ).end ( ) ) throw exception::CommonException ( "Parse callback for " + tagName + " tag not registered." );

			Group * res = new Group ( callback->second->parse ( input ) );
			input.idToInstance ( Group::XML_TAG_NAME_REF ).insert ( std::make_pair ( input.idToInstanceMax ( Group::XML_TAG_NAME_REF )++, ( void * ) res ) );
			return * res;
		}
	}

	static bool first ( const std::deque < sax::Token >::const_iterator & input ) {
		const std::string & tagName = sax::FromXMLParserHelper::getTokenData ( input, sax::Token::TokenType::START_ELEMENT );

		if ( tagName == Group::XML_TAG_NAME_REF ) return true;

		typename std::map < std::string, ParserRegisterBase * >::iterator callback = parseFunctions ( ).find ( tagName );

		if ( callback == parseFunctions ( ).end ( ) ) return false;

		return true;
	}

	static void compose ( std::deque < sax::Token > & output, const Group & data ) {
		xmlApiOutputContext & context = ( xmlApiOutputContext & ) output;

		typename std::map < void *, int >::iterator elem = context.instanceToId ( Group::XML_TAG_NAME_REF ).find ( ( void * ) & data.getData ( ) );

		if ( elem == context.instanceToId ( Group::XML_TAG_NAME_REF ).end ( ) ) {
			data.getData ( ).compose ( output );
			context.instanceToId ( Group::XML_TAG_NAME_REF ).insert ( std::make_pair ( ( void * ) & data.getData ( ), context.instanceToIdMax ( Group::XML_TAG_NAME_REF )++ ) );
		} else {
			output.emplace_back ( Group::XML_TAG_NAME_REF, sax::Token::TokenType::START_ELEMENT );
			output.emplace_back ( "id", sax::Token::TokenType::START_ATTRIBUTE );
			output.emplace_back ( std::itos ( elem->second ), sax::Token::TokenType::CHARACTER );
			output.emplace_back ( "id", sax::Token::TokenType::END_ATTRIBUTE );
			output.emplace_back ( Group::XML_TAG_NAME_REF, sax::Token::TokenType::END_ELEMENT );
		}
	}

};

template < typename Type >
struct xmlApi < Type, typename std::enable_if < std::is_base_of < ObjectBase, Type >::value >::type > {
	static Type parse ( std::deque < sax::Token >::iterator & input ) {
		return Type::parse ( input );
	}

	static bool first ( const std::deque < sax::Token >::const_iterator & input ) {
		return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, Type::XML_TAG_NAME );
	}

	static void compose ( std::deque < sax::Token > & output, const Type & data ) {
		data.compose ( output );
	}

};

} /* namespace alib */

#endif /* XML_API_HPP_ */
