/*
 * DFA.cpp
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#include "DFA.h"
#include "../AutomatonException.h"
#include <ostream>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace automaton {

DFA::DFA ( std::set < State > states, std::set < alphabet::Symbol > inputAlphabet, State initialState, std::set < State > finalStates ) : std::Components < DFA, alphabet::Symbol, std::tuple < InputAlphabet >, std::tuple < >, automaton::State, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

DFA::DFA ( State initialState ) : DFA ( std::set < State > { initialState }, std::set < alphabet::Symbol > { }, initialState, std::set < State > { } ) {
}

AutomatonBase * DFA::clone ( ) const {
	return new DFA ( * this );
}

AutomatonBase * DFA::plunder ( ) && {
	return new DFA ( std::move ( * this ) );
}

bool DFA::addTransition ( State from, alphabet::Symbol input, State to ) {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ( std::string ) from.getName ( ) + "\" doesn't exist." );

	if ( !getInputAlphabet ( ).count ( input ) )
		throw AutomatonException ( "Input symbol \"" + ( std::string ) input + "\" doesn't exist." );

	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ( std::string ) to.getName ( ) + "\" doesn't exist." );

	std::pair < State, alphabet::Symbol > key = std::make_pair ( std::move ( from ), std::move ( input ) );

	if ( transitions.find ( key ) != transitions.end ( ) ) {
		if ( transitions.find ( key )->second == to )
			return false;
		else
			throw AutomatonException ( "Transition from this state and symbol already exists (\"" + ( std::string ) key.first.getName ( ) + "\", \"" + ( std::string ) key.second + "\") -> \"" + ( std::string ) to.getName ( ) + "\"." );
	}

	transitions.insert ( std::make_pair ( std::move ( key ), std::move ( to ) ) );
	return true;
}

bool DFA::removeTransition ( const State & from, const alphabet::Symbol & input, const State & to ) {
	std::pair < State, alphabet::Symbol > key = std::make_pair ( from, input );

	if ( transitions.find ( key ) == transitions.end ( ) )
		return false;

	if ( transitions.find ( key )->second != to )
		throw AutomatonException ( "Transition (\"" + ( std::string ) from.getName ( ) + "\", \"" + ( std::string ) input + "\") -> \"" + ( std::string ) to.getName ( ) + "\" doesn't exist." );

	transitions.erase ( key );
	return true;
}

const std::map < std::pair < State, alphabet::Symbol >, State > & DFA::getTransitions ( ) const {
	return transitions;
}

std::map < std::pair < State, alphabet::Symbol >, State > DFA::getTransitionsFromState ( const State & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ( std::string ) from.getName ( ) + "\" doesn't exist" );

	std::map < std::pair < State, alphabet::Symbol >, State > transitionsFromState;

	for ( const std::pair < const std::pair < State, alphabet::Symbol >, State > & transition : transitions )
		if ( transition.first.first == from )
			transitionsFromState.insert ( transition );

	return transitionsFromState;
}

std::map < std::pair < State, alphabet::Symbol >, State > DFA::getTransitionsToState ( const State & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ( std::string ) to.getName ( ) + "\" doesn't exist" );

	std::map < std::pair < State, alphabet::Symbol >, State > transitionsToState;

	for ( const std::pair < const std::pair < State, alphabet::Symbol >, State > & transition : transitions )
		if ( transition.second == to )
			transitionsToState.insert ( transition );

	return transitionsToState;
}

bool DFA::isTotal ( ) const {
	return transitions.size ( ) == getInputAlphabet ( ).size ( ) * getStates ( ).size ( );
}

int DFA::compare ( const DFA & other ) const {
	auto first = std::tie ( getStates ( ), getInputAlphabet ( ), getInitialState ( ), getFinalStates ( ), transitions );
	auto second = std::tie ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ), other.transitions );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void DFA::operator >>( std::ostream & out ) const {
	out << "(DFA "
	    << " states = " << getStates ( )
	    << " inputAlphabet = " << getInputAlphabet ( )
	    << " initialState = " << getInitialState ( )
	    << " finalStates = " << getFinalStates ( )
	    << " transitions = " << transitions
	    << ")";
}

DFA::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

const std::string DFA::XML_TAG_NAME = "DFA";

DFA DFA::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, DFA::XML_TAG_NAME );

	std::set < State > states = AutomatonFromXMLParser::parseStates ( input );
	std::set < alphabet::Symbol > inputSymbols = AutomatonFromXMLParser::parseInputAlphabet ( input );
	State initialState = AutomatonFromXMLParser::parseInitialState ( input );
	std::set < State > finalStates = AutomatonFromXMLParser::parseFinalStates ( input );

	DFA automaton ( std::move ( initialState ) );

	automaton.setStates ( std::move ( states ) );
	automaton.setInputAlphabet ( std::move ( inputSymbols ) );
	automaton.setFinalStates ( std::move ( finalStates ) );

	AutomatonFromXMLParser::parseTransitions < DFA > ( input, automaton );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, DFA::XML_TAG_NAME );
	return automaton;
}

void DFA::parseTransition ( std::deque < sax::Token >::iterator & input, DFA & automaton ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "transition" );
	State from = AutomatonFromXMLParser::parseTransitionFrom ( input );
	alphabet::Symbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol ( input );
	State to = AutomatonFromXMLParser::parseTransitionTo ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "transition" );

	automaton.addTransition ( std::move ( from ), std::move ( inputSymbol ), std::move ( to ) );
}

void DFA::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( DFA::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT );

	AutomatonToXMLComposer::composeStates ( out, this->getStates ( ) );
	AutomatonToXMLComposer::composeInputAlphabet ( out, this->getInputAlphabet ( ) );
	AutomatonToXMLComposer::composeInitialState ( out, this->getInitialState ( ) );
	AutomatonToXMLComposer::composeFinalStates ( out, this->getFinalStates ( ) );
	composeTransitions ( out );

	out.emplace_back ( DFA::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT );
}

void DFA::composeTransitions ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "transitions", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & transition : this->getTransitions ( ) ) {
		out.emplace_back ( "transition", sax::Token::TokenType::START_ELEMENT );

		AutomatonToXMLComposer::composeTransitionFrom ( out, transition.first.first );
		AutomatonToXMLComposer::composeTransitionInputSymbol ( out, transition.first.second );
		AutomatonToXMLComposer::composeTransitionTo ( out, transition.second );

		out.emplace_back ( "transition", sax::Token::TokenType::END_ELEMENT );
	}

	out.emplace_back ( "transitions", sax::Token::TokenType::END_ELEMENT );
}

} /* namespace automaton */

namespace std {

template < >
bool automaton::DFA::Component < automaton::DFA, alphabet::Symbol, automaton::InputAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	const automaton::DFA * automaton = static_cast < const automaton::DFA * > ( this );

	for ( const std::pair < const std::pair < automaton::State, alphabet::Symbol >, automaton::State > & transition : automaton->getTransitions ( ) )
		if ( transition.first.second == symbol )
			return true;

	return false;
}

template < >
bool automaton::DFA::Component < automaton::DFA, alphabet::Symbol, automaton::InputAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void automaton::DFA::Component < automaton::DFA, alphabet::Symbol, automaton::InputAlphabet >::valid ( const alphabet::Symbol & ) const {
}

template < >
bool automaton::DFA::Component < automaton::DFA, automaton::State, automaton::States >::used ( const automaton::State & state ) const {
	const automaton::DFA * automaton = static_cast < const automaton::DFA * > ( this );

	if ( automaton->getInitialState ( ) == state )
		return true;

	if ( automaton->getFinalStates ( ).count ( state ) )
		return true;

	for ( const std::pair < const std::pair < automaton::State, alphabet::Symbol >, automaton::State > & t : automaton->getTransitions ( ) )
		if ( ( t.first.first == state ) || ( t.second == state ) )
			return true;

	return false;
}

template < >
bool automaton::DFA::Component < automaton::DFA, automaton::State, automaton::States >::available ( const automaton::State & ) const {
	return true;
}

template < >
void automaton::DFA::Component < automaton::DFA, automaton::State, automaton::States >::valid ( const automaton::State & ) const {
}

template < >
bool automaton::DFA::Component < automaton::DFA, automaton::State, automaton::FinalStates >::used ( const automaton::State & ) const {
	return false;
}

template < >
bool automaton::DFA::Component < automaton::DFA, automaton::State, automaton::FinalStates >::available ( const automaton::State & state ) const {
	const automaton::DFA * automaton = static_cast < const automaton::DFA * > ( this );

	return automaton->accessComponent < automaton::States > ( ).get ( ).count ( state );
}

template < >
void automaton::DFA::Component < automaton::DFA, automaton::State, automaton::FinalStates >::valid ( const automaton::State & ) const {
}

template < >
bool automaton::DFA::Element < automaton::DFA, automaton::State, automaton::InitialState >::available ( const automaton::State & state ) const {
	const automaton::DFA * automaton = static_cast < const automaton::DFA * > ( this );

	return automaton->accessComponent < automaton::States > ( ).get ( ).count ( state );
}

template < >
void automaton::DFA::Element < automaton::DFA, automaton::State, automaton::InitialState >::valid ( const automaton::State & ) const {
}

} /* namespace std */

namespace alib {

auto DFAParserRegister = xmlApi < automaton::Automaton >::ParserRegister < automaton::DFA > ( );
auto DFAParserRegister2 = xmlApi < alib::Object >::ParserRegister < automaton::DFA > ( );

} /* namespace alib */
