/*
 * ExtendedNFA.cpp
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#include "ExtendedNFA.h"
#include "CompactNFA.h"
#include "EpsilonNFA.h"
#include "MultiInitialStateNFA.h"
#include "NFA.h"
#include "DFA.h"
#include "../AutomatonException.h"
#include "../Automaton.h"
#include "../../regexp/RegExp.h"
#include <ostream>
#include <algorithm>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace automaton {

ExtendedNFA::ExtendedNFA ( std::set < State > states, std::set < alphabet::Symbol > inputAlphabet, State initialState, std::set < State > finalStates ) : std::Components < ExtendedNFA, alphabet::Symbol, std::tuple < InputAlphabet >, std::tuple < >, automaton::State, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

ExtendedNFA::ExtendedNFA ( State initialState ) : ExtendedNFA ( std::set < State > { initialState }, std::set < alphabet::Symbol > { }, initialState, std::set < State > { } ) {
}

ExtendedNFA::ExtendedNFA ( const CompactNFA & other ) : ExtendedNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		std::pair < State, regexp::RegExp > key = std::make_pair ( transition.first.first, regexp::RegExp ( regexp::regexpFrom ( transition.first.second ) ) );
		transitions[key] = transition.second;
	}
}

ExtendedNFA::ExtendedNFA ( const EpsilonNFA & other ) : ExtendedNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		if ( transition.first.second.is < string::Epsilon > ( ) ) {
			std::pair < State, regexp::RegExp > key = std::make_pair ( transition.first.first, regexp::RegExp ( regexp::regexpFrom ( string::LinearString ( std::vector < alphabet::Symbol > { } ) ) ) );
			transitions[key] = transition.second;
		} else {
			std::pair < State, regexp::RegExp > key = std::make_pair ( transition.first.first, regexp::RegExp ( regexp::regexpFrom ( transition.first.second.get < alphabet::Symbol > ( ) ) ) );
			transitions[key] = transition.second;
		}
	}
}

ExtendedNFA::ExtendedNFA ( const MultiInitialStateNFA & other ) : ExtendedNFA ( other.getStates ( ) + std::set < State > { automaton::createUniqueState ( automaton::State ( "q0" ), other.getStates ( ) ) }, other.getInputAlphabet ( ), automaton::createUniqueState ( automaton::State ( "q0" ), other.getStates ( ) ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		std::pair < State, regexp::RegExp > key = std::make_pair ( transition.first.first, regexp::RegExp ( regexp::regexpFrom ( transition.first.second ) ) );
		transitions[key] = transition.second;
	}

	std::pair < State, regexp::RegExp > key = std::make_pair ( this->getInitialState ( ), regexp::RegExp ( regexp::regexpFrom ( string::LinearString ( std::vector < alphabet::Symbol > { } ) ) ) );
	transitions[key] = other.getInitialStates ( );
}

ExtendedNFA::ExtendedNFA ( const NFA & other ) : ExtendedNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		std::pair < State, regexp::RegExp > key = std::make_pair ( transition.first.first, regexp::RegExp ( regexp::regexpFrom ( transition.first.second ) ) );
		transitions[key] = transition.second;
	}
}

ExtendedNFA::ExtendedNFA ( const DFA & other ) : ExtendedNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		std::pair < State, regexp::RegExp > key = std::make_pair ( transition.first.first, regexp::RegExp ( regexp::regexpFrom ( transition.first.second ) ) );
		transitions[key].insert ( transition.second );
	}
}

AutomatonBase * ExtendedNFA::clone ( ) const {
	return new ExtendedNFA ( * this );
}

AutomatonBase * ExtendedNFA::plunder ( ) && {
	return new ExtendedNFA ( std::move ( * this ) );
}

bool ExtendedNFA::addTransition ( State from, regexp::RegExp input, State to ) {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ( std::string ) from.getName ( ) + "\" doesn't exist." );

	const std::set < alphabet::Symbol > & inputRegExpAlphabet = input.getAlphabet ( );

	 // Transition regexp's alphabet must be subset of automaton's alphabet
	if ( !std::includes ( getInputAlphabet ( ).begin ( ), getInputAlphabet ( ).end ( ), inputRegExpAlphabet.begin ( ), inputRegExpAlphabet.end ( ) ) )
		throw AutomatonException ( "Input string is over different alphabet than automaton" );

	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ( std::string ) to.getName ( ) + "\" doesn't exist." );

	std::pair < State, regexp::RegExp > key = std::make_pair ( std::move ( from ), std::move ( input ) );

	return transitions[std::move ( key )].insert ( std::move ( to ) ).second;
}

bool ExtendedNFA::removeTransition ( const State & from, const regexp::RegExp & input, const State & to ) {
	std::pair < State, regexp::RegExp > key = std::make_pair ( from, input );

	return transitions[key].erase ( to );
}

const std::map < std::pair < State, regexp::RegExp >, std::set < State > > & ExtendedNFA::getTransitions ( ) const {
	return transitions;
}

std::map < std::pair < State, regexp::RegExp >, std::set < State > > ExtendedNFA::getTransitionsFromState ( const State & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ( std::string ) from.getName ( ) + "\" doesn't exist" );

	std::map < std::pair < State, regexp::RegExp >, std::set < State > > transitionsFromState;

	for ( const std::pair < const std::pair < State, regexp::RegExp >, std::set < State > > & transition : transitions )
		if ( transition.first.first == from )
			transitionsFromState.insert ( make_pair ( transition.first, transition.second ) );

	return transitionsFromState;
}

std::map < std::pair < State, regexp::RegExp >, std::set < State > > ExtendedNFA::getTransitionsToState ( const State & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ( std::string ) to.getName ( ) + "\" doesn't exist" );

	std::map < std::pair < State, regexp::RegExp >, std::set < State > > transitionsToState;

	for ( const std::pair < const std::pair < State, regexp::RegExp >, std::set < State > > & transition : transitions )
		if ( transition.second.find ( to ) != transition.second.end ( ) )
			transitionsToState.insert ( make_pair ( transition.first, transition.second ) );

	return transitionsToState;
}

int ExtendedNFA::compare ( const ExtendedNFA & other ) const {
	auto first = std::tie ( getStates ( ), getInputAlphabet ( ), getInitialState ( ), getFinalStates ( ), transitions );
	auto second = std::tie ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ), other.getTransitions ( ) );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void ExtendedNFA::operator >>( std::ostream & out ) const {
	out << "(ExtendedNFA "
	    << "states = " << getStates ( )
	    << "inputAlphabet = " << getInputAlphabet ( )
	    << "initialState = " << getInitialState ( )
	    << "finalStates = " << getFinalStates ( )
	    << "transitions = " << transitions
	    << ")";
}

ExtendedNFA::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

const std::string ExtendedNFA::XML_TAG_NAME = "ExtendedNFA";

ExtendedNFA ExtendedNFA::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, ExtendedNFA::XML_TAG_NAME );

	std::set < State > states = AutomatonFromXMLParser::parseStates ( input );
	std::set < alphabet::Symbol > inputSymbols = AutomatonFromXMLParser::parseInputAlphabet ( input );
	State initialState = AutomatonFromXMLParser::parseInitialState ( input );
	std::set < State > finalStates = AutomatonFromXMLParser::parseFinalStates ( input );

	ExtendedNFA automaton ( std::move ( initialState ) );

	automaton.setStates ( std::move ( states ) );
	automaton.setInputAlphabet ( std::move ( inputSymbols ) );
	automaton.setFinalStates ( std::move ( finalStates ) );

	AutomatonFromXMLParser::parseTransitions < ExtendedNFA > ( input, automaton );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, ExtendedNFA::XML_TAG_NAME );
	return automaton;
}

void ExtendedNFA::parseTransition ( std::deque < sax::Token >::iterator & input, ExtendedNFA & automaton ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "transition" );
	State from = AutomatonFromXMLParser::parseTransitionFrom ( input );
	regexp::RegExp inputRegexp = AutomatonFromXMLParser::parseTransitionInputRegexp ( input );
	State to = AutomatonFromXMLParser::parseTransitionTo ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "transition" );

	automaton.addTransition ( std::move ( from ), std::move ( inputRegexp ), std::move ( to ) );
}

void ExtendedNFA::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( ExtendedNFA::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT );

	AutomatonToXMLComposer::composeStates ( out, this->getStates ( ) );
	AutomatonToXMLComposer::composeInputAlphabet ( out, this->getInputAlphabet ( ) );
	AutomatonToXMLComposer::composeInitialState ( out, this->getInitialState ( ) );
	AutomatonToXMLComposer::composeFinalStates ( out, this->getFinalStates ( ) );
	composeTransitions ( out );

	out.emplace_back ( ExtendedNFA::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT );
}

void ExtendedNFA::composeTransitions ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "transitions", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & transition : this->getTransitions ( ) )
		for ( const auto & targetState : transition.second ) {
			out.emplace_back ( "transition", sax::Token::TokenType::START_ELEMENT );

			AutomatonToXMLComposer::composeTransitionFrom ( out, transition.first.first );
			AutomatonToXMLComposer::composeTransitionInputRegexp ( out, transition.first.second );
			AutomatonToXMLComposer::composeTransitionTo ( out, targetState );

			out.emplace_back ( "transition", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "transitions", sax::Token::TokenType::END_ELEMENT );
}

} /* namespace automaton */

namespace std {

template < >
bool automaton::ExtendedNFA::Component < automaton::ExtendedNFA, alphabet::Symbol, automaton::InputAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	const automaton::ExtendedNFA * automaton = static_cast < const automaton::ExtendedNFA * > ( this );

	for ( const std::pair < const std::pair < automaton::State, regexp::RegExp >, std::set < automaton::State > > & transition : automaton->getTransitions ( ) )
		if ( transition.first.second.getAlphabet ( ).count ( symbol ) == 1 )
			return true;

	return false;
}

template < >
bool automaton::ExtendedNFA::Component < automaton::ExtendedNFA, alphabet::Symbol, automaton::InputAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void automaton::ExtendedNFA::Component < automaton::ExtendedNFA, alphabet::Symbol, automaton::InputAlphabet >::valid ( const alphabet::Symbol & ) const {
}

template < >
bool automaton::ExtendedNFA::Component < automaton::ExtendedNFA, automaton::State, automaton::States >::used ( const automaton::State & state ) const {
	const automaton::ExtendedNFA * automaton = static_cast < const automaton::ExtendedNFA * > ( this );

	if ( automaton->getInitialState ( ) == state )
		return true;

	if ( automaton->getFinalStates ( ).count ( state ) )
		return true;

	for ( const std::pair < const std::pair < automaton::State, regexp::RegExp >, std::set < automaton::State > > & transition : automaton->getTransitions ( ) )
		if ( ( transition.first.first == state ) || ( transition.second.find ( state ) != transition.second.end ( ) ) )
			return true;

	return false;
}

template < >
bool automaton::ExtendedNFA::Component < automaton::ExtendedNFA, automaton::State, automaton::States >::available ( const automaton::State & ) const {
	return true;
}

template < >
void automaton::ExtendedNFA::Component < automaton::ExtendedNFA, automaton::State, automaton::States >::valid ( const automaton::State & ) const {
}

template < >
bool automaton::ExtendedNFA::Component < automaton::ExtendedNFA, automaton::State, automaton::FinalStates >::used ( const automaton::State & ) const {
	return false;
}

template < >
bool automaton::ExtendedNFA::Component < automaton::ExtendedNFA, automaton::State, automaton::FinalStates >::available ( const automaton::State & state ) const {
	const automaton::ExtendedNFA * automaton = static_cast < const automaton::ExtendedNFA * > ( this );

	return automaton->accessComponent < automaton::States > ( ).get ( ).count ( state );
}

template < >
void automaton::ExtendedNFA::Component < automaton::ExtendedNFA, automaton::State, automaton::FinalStates >::valid ( const automaton::State & ) const {
}

template < >
bool automaton::ExtendedNFA::Element < automaton::ExtendedNFA, automaton::State, automaton::InitialState >::available ( const automaton::State & state ) const {
	const automaton::ExtendedNFA * automaton = static_cast < const automaton::ExtendedNFA * > ( this );

	return automaton->accessComponent < automaton::States > ( ).get ( ).count ( state );
}

template < >
void automaton::ExtendedNFA::Element < automaton::ExtendedNFA, automaton::State, automaton::InitialState >::valid ( const automaton::State & ) const {
}

} /* namespace std */

namespace alib {

auto extendedNFAParserRegister = xmlApi < automaton::Automaton >::ParserRegister < automaton::ExtendedNFA > ( );
auto extendedNFAParserRegister2 = xmlApi < alib::Object >::ParserRegister < automaton::ExtendedNFA > ( );

auto ExtendedNFAFromDFA = castApi::CastRegister < automaton::ExtendedNFA, automaton::DFA > ( );
auto ExtendedNFAFromNFA = castApi::CastRegister < automaton::ExtendedNFA, automaton::NFA > ( );
auto ExtendedNFAFromMultiInitialStateNFA = castApi::CastRegister < automaton::ExtendedNFA, automaton::MultiInitialStateNFA > ( );
auto ExtendedNFAEpsilonNFA = castApi::CastRegister < automaton::ExtendedNFA, automaton::EpsilonNFA > ( );
auto ExtendedNFACompactNFA = castApi::CastRegister < automaton::ExtendedNFA, automaton::CompactNFA > ( );
auto ExtendedNFACastBinder = castApi::CastPoolStringBinder < automaton::ExtendedNFA > ( automaton::ExtendedNFA::XML_TAG_NAME );

} /* namespace alib */
