/*
 * RealTimeHeightDeterministicDPDA.cpp
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#include "RealTimeHeightDeterministicDPDA.h"
#include "../AutomatonException.h"
#include <algorithm>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace automaton {

RealTimeHeightDeterministicDPDA::RealTimeHeightDeterministicDPDA(State initialState, alphabet::Symbol bottomOfTheStackSymbol) : SingleInitialState(std::move(initialState)), BottomOfTheStackSymbolPushdownStoreAlphabet(std::move(bottomOfTheStackSymbol)) {

}

AutomatonBase* RealTimeHeightDeterministicDPDA::clone() const {
	return new RealTimeHeightDeterministicDPDA(*this);
}

AutomatonBase* RealTimeHeightDeterministicDPDA::plunder() && {
	return new RealTimeHeightDeterministicDPDA(std::move(*this));
}

bool RealTimeHeightDeterministicDPDA::removeState(const State& state) {
	if (initialState == state) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (const std::pair<const std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, std::pair<State, alphabet::Symbol> >& callTransition : callTransitions) {
		if (state == callTransition.first.first)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		if(callTransition.second.first == state)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	for (const std::pair<const std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, State>& returnTransition : returnTransitions) {
		if (state == std::get<0>(returnTransition.first))
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		if(returnTransition.second == state)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	for (const std::pair<const std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, State>& localTransition : localTransitions) {
		if (state == localTransition.first.first)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		if(localTransition.second == state)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	return states.erase(state);
}

bool RealTimeHeightDeterministicDPDA::removeInputSymbol(const alphabet::Symbol& symbol) {
	for (const std::pair<const std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, std::pair<State, alphabet::Symbol> >& callTransition : callTransitions) {
		if (callTransition.first.second.is<alphabet::Symbol>() && symbol == callTransition.first.second.get<alphabet::Symbol>())
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	for (const std::pair<const std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, State>& returnTransition : returnTransitions) {
		if (std::get<1>(returnTransition.first).is<alphabet::Symbol>() && symbol == std::get<1>(returnTransition.first).get<alphabet::Symbol>())
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	for (const std::pair<const std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, State>& localTransition : localTransitions) {
		if (localTransition.first.second.is<alphabet::Symbol>() && symbol == localTransition.first.second.get<alphabet::Symbol>())
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	return inputAlphabet.erase(symbol);
}

bool RealTimeHeightDeterministicDPDA::removeStackSymbol(const alphabet::Symbol& symbol) {
	for (const std::pair<const std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, std::pair<State, alphabet::Symbol> >& callTransition : callTransitions) {
		if (symbol == callTransition.second.second)
			throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	for (const std::pair<const std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, State>& returnTransition : returnTransitions) {
		if (symbol == std::get<2>(returnTransition.first))
			throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	if(bottomOfTheStackSymbol == symbol) {
		throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is start symbol.");
	}

	return stackAlphabet.erase(symbol);
}

bool RealTimeHeightDeterministicDPDA::addCallTransition(State from, std::variant<string::Epsilon, alphabet::Symbol> input, State to, alphabet::Symbol push) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (input.is<alphabet::Symbol>() && inputAlphabet.find(input.get<alphabet::Symbol>()) == inputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input.get<alphabet::Symbol>() + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	if (stackAlphabet.find(push) == stackAlphabet.end()) {
		throw AutomatonException("Stack symbol \"" + (std::string) push + "\" doesn't exist.");
	}

	std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>> key(std::move(from), std::move(input));
	std::pair<automaton::State, alphabet::Symbol> value = std::make_pair(std::move(to), std::move(push));

	if(callTransitions.find(key) != callTransitions.end() && callTransitions.find(key)->second == value)
		return false;

	if(key.second.is<string::Epsilon>()) {
		for(const auto& transition : callTransitions)
			if(transition.first.first == key.first )
				throw AutomatonException("Can't add epsilon transition from state \"" + (std::string) key.first + "\" when other transitions are present.");

		for(const auto& transition : returnTransitions)
			if(std::get<0>(transition.first) == key.first)
				throw AutomatonException("Can't add epsilon transition from state \"" + (std::string) key.first + "\" when other transitions are present.");

		for(const auto& transition : localTransitions)
			if(transition.first.first == key.first)
				throw AutomatonException("Can't add epsilon transition from state \"" + (std::string) from + "\" when other transitions are present.");
	} else {
		for(const auto& transition : callTransitions)
			if(transition.first.first == key.first && (transition.first.second.is<string::Epsilon>() || transition.first.second == key.second))
				throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");

		for(const auto& transition : returnTransitions)
			if(std::get<0>(transition.first) == key.first && (std::get<1>(transition.first).is<string::Epsilon>() || std::get<1>(transition.first) == key.second))
				throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");

		for(const auto& transition : localTransitions)
			if(transition.first.first == key.first && (transition.first.second.is<string::Epsilon>() || transition.first.second == key.second))
				throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");
	}

	callTransitions.insert(std::make_pair(std::move(key), std::move(value)));
	return true;
}

bool RealTimeHeightDeterministicDPDA::addCallTransition(State from, State to, alphabet::Symbol push) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return addCallTransition(std::move(from), std::move(inputVariant), std::move(to), std::move(push));
}

bool RealTimeHeightDeterministicDPDA::addCallTransition(State from, alphabet::Symbol input, State to, alphabet::Symbol push) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(std::move(input));
	return addCallTransition(std::move(from), std::move(inputVariant), std::move(to), std::move(push));
}

bool RealTimeHeightDeterministicDPDA::addReturnTransition(State from, std::variant<string::Epsilon, alphabet::Symbol> input, alphabet::Symbol pop, State to) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (input.is<alphabet::Symbol>() && inputAlphabet.find(input.get<alphabet::Symbol>()) == inputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input.get<alphabet::Symbol>() + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	if (stackAlphabet.find(pop) == stackAlphabet.end()) {
		throw AutomatonException("Stack symbol \"" + (std::string) pop + "\" doesn't exist.");
	}

	std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol> key(std::move(from), std::move(input), std::move(pop));

	if(returnTransitions.find(key) != returnTransitions.end() && returnTransitions.find(key)->second == to)
		return false;

	if(std::get<1>(key).is<string::Epsilon>()) {
		for(const auto& transition : callTransitions)
			if(transition.first.first == std::get<0>(key) )
				throw AutomatonException("Can't add epsilon transition from state \"" + (std::string) std::get<0>(key) + "\" when other transitions are present.");

		for(const auto& transition : returnTransitions)
			if(std::get<0>(transition.first) == std::get<0>(key) && std::get<2>(transition.first) == std::get<2>(key))
				throw AutomatonException("Can't add epsilon transition from state \"" + (std::string) std::get<0>(key) + "\" when other transitions are present.");

		for(const auto& transition : localTransitions)
			if(transition.first.first == std::get<0>(key))
				throw AutomatonException("Can't add epsilon transition from state \"" + (std::string) std::get<0>(key) + "\" when other transitions are present.");
	} else {
		for(const auto& transition : callTransitions)
			if(transition.first.first == std::get<0>(key) && (transition.first.second.is<string::Epsilon>() || transition.first.second == std::get<1>(key)))
				throw AutomatonException("Can't add transition from state \"" + (std::string) std::get<0>(key) + "\" when transition reading \"" + (std::string) std::get<1>(key) + "\" is present.");

		for(const auto& transition : returnTransitions)
			if(std::get<0>(transition.first) == std::get<0>(key) && (std::get<1>(transition.first).is<string::Epsilon>() || std::get<1>(transition.first) == std::get<1>(key)) && std::get<2>(transition.first) == std::get<2>(key))
				throw AutomatonException("Can't add transition from state \"" + (std::string) std::get<0>(key) + "\" when transition reading \"" + (std::string) std::get<1>(key) + "\" is present.");

		for(const auto& transition : localTransitions)
			if(transition.first.first == std::get<0>(key) && (transition.first.second.is<string::Epsilon>() || transition.first.second == std::get<1>(key)))
				throw AutomatonException("Can't add transition from state \"" + (std::string) std::get<0>(key) + "\" when transition reading \"" + (std::string) std::get<1>(key) + "\" is present.");
	}

	returnTransitions.insert(std::make_pair(std::move(key), std::move(to)));
	return true;
}

bool RealTimeHeightDeterministicDPDA::addReturnTransition(State from, alphabet::Symbol pop, State to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return addReturnTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to));
}

bool RealTimeHeightDeterministicDPDA::addReturnTransition(State from, alphabet::Symbol input, alphabet::Symbol pop, State to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(std::move(input));
	return addReturnTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to));
}

bool RealTimeHeightDeterministicDPDA::addLocalTransition(State from, std::variant<string::Epsilon, alphabet::Symbol> input, State to) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (input.is<alphabet::Symbol>() && inputAlphabet.find(input.get<alphabet::Symbol>()) == inputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input.get<alphabet::Symbol>() + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>> key(std::move(from), std::move(input));

	if(localTransitions.find(key) != localTransitions.end() && localTransitions.find(key)->second == to)
		return false;

	if(key.second.is<string::Epsilon>()) {
		for(const auto& transition : callTransitions)
			if(transition.first.first == key.first )
				throw AutomatonException("Can't add epsilon transition from state \"" + (std::string) key.first + "\" when other transitions are present.");

		for(const auto& transition : returnTransitions)
			if(std::get<0>(transition.first) == key.first)
				throw AutomatonException("Can't add epsilon transition from state \"" + (std::string) key.first + "\" when other transitions are present.");

		for(const auto& transition : localTransitions)
			if(transition.first.first == key.first)
				throw AutomatonException("Can't add epsilon transition from state \"" + (std::string) key.first + "\" when other transitions are present.");
	} else {
		for(const auto& transition : callTransitions)
			if(transition.first.first == key.first && (transition.first.second.is<string::Epsilon>() || transition.first.second == key.second))
				throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");

		for(const auto& transition : returnTransitions)
			if(std::get<0>(transition.first) == key.first && (std::get<1>(transition.first).is<string::Epsilon>() || std::get<1>(transition.first) == key.second))
				throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");

		for(const auto& transition : localTransitions)
			if(transition.first.first == key.first && (transition.first.second.is<string::Epsilon>() || transition.first.second == key.second))
				throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");
	}

	localTransitions.insert(std::make_pair(std::move(key), std::move(to)));
	return true;
}

bool RealTimeHeightDeterministicDPDA::addLocalTransition(State from, State to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return addLocalTransition(std::move(from), std::move(inputVariant), std::move(to));
}

bool RealTimeHeightDeterministicDPDA::addLocalTransition(State from, alphabet::Symbol input, State to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(std::move(input));
	return addLocalTransition(std::move(from), std::move(inputVariant), std::move(to));
}

bool RealTimeHeightDeterministicDPDA::removeCallTransition(const State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const State& to, const alphabet::Symbol& push) {
	std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>> key(from, input);
	std::pair<automaton::State, alphabet::Symbol> value = std::make_pair(to, push);

	if (callTransitions.find(key) == callTransitions.end())
		return false;

	if(callTransitions.find(key)->second != value)
		throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input
						+ "\") -> \"" + (std::string) to.getName() + "\" doesn't exist.");

	callTransitions.erase(key);
	return true;
}

bool RealTimeHeightDeterministicDPDA::removeCallTransition(const State& from, const State& to, const alphabet::Symbol& push) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return removeCallTransition(from, inputVariant, to, push);
}

bool RealTimeHeightDeterministicDPDA::removeCallTransition(const State& from, const alphabet::Symbol& input, const State& to, const alphabet::Symbol& push) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(input);
	return removeCallTransition(from, inputVariant, to, push);
}

bool RealTimeHeightDeterministicDPDA::removeReturnTransition(const State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const alphabet::Symbol& pop, const State& to) {
	std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol> key(from, input, pop);

	if (returnTransitions.find(key) == returnTransitions.end())
		return false;

	if(returnTransitions.find(key)->second != to)
		throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input
						+ "\") -> \"" + (std::string) to.getName() + "\" doesn't exist.");

	returnTransitions.erase(key);
	return true;
}

bool RealTimeHeightDeterministicDPDA::removeReturnTransition(const State& from, const alphabet::Symbol& pop, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return removeReturnTransition(from, inputVariant, pop, to);
}

bool RealTimeHeightDeterministicDPDA::removeReturnTransition(const State& from, const alphabet::Symbol& input, const alphabet::Symbol& pop, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(input);
	return removeReturnTransition(from, inputVariant, pop, to);
}

bool RealTimeHeightDeterministicDPDA::removeLocalTransition(const State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const State& to) {
	std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>> key(from, input);

	if (localTransitions.find(key) == localTransitions.end())
		return false;

	if(localTransitions.find(key)->second != to)
		throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input
						+ "\") -> \"" + (std::string) to.getName() + "\" doesn't exist.");

	localTransitions.erase(key);
	return true;
}

bool RealTimeHeightDeterministicDPDA::removeLocalTransition(const State& from, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return removeLocalTransition(from, inputVariant, to);
}

bool RealTimeHeightDeterministicDPDA::removeLocalTransition(const State& from, const alphabet::Symbol& input, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(input);
	return removeLocalTransition(from, inputVariant, to);
}

const std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, std::pair<State, alphabet::Symbol> >& RealTimeHeightDeterministicDPDA::getCallTransitions() const {
	return callTransitions;
}

const std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, State>& RealTimeHeightDeterministicDPDA::getReturnTransitions() const {
	return returnTransitions;
}

const std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, State>& RealTimeHeightDeterministicDPDA::getLocalTransitions() const {
	return localTransitions;
}

int RealTimeHeightDeterministicDPDA::compare(const RealTimeHeightDeterministicDPDA& other) const {
	auto first = std::tie(states, inputAlphabet, initialState, finalStates, stackAlphabet, bottomOfTheStackSymbol, callTransitions, returnTransitions, localTransitions);
	auto second = std::tie(other.states, other.inputAlphabet, other.initialState, other.finalStates, other.stackAlphabet, other.bottomOfTheStackSymbol, other.callTransitions, other.returnTransitions, other.localTransitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void RealTimeHeightDeterministicDPDA::operator>>(std::ostream& out) const {
	out << "(RealTimeHeightDeterministicDPDA "
		<< "states = " << states
		<< "inputAlphabet = " << inputAlphabet
		<< "initialState = " << initialState
		<< "finalStates = " << finalStates
		<< "stackAlphabet = " << stackAlphabet
		<< "bottomOfTheStackSymbol = " << bottomOfTheStackSymbol
		<< "callTransitions = " << callTransitions
		<< "returnTransitions = " << returnTransitions
		<< "localTransitions = " << localTransitions
		<< ")";
}

RealTimeHeightDeterministicDPDA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string RealTimeHeightDeterministicDPDA::XML_TAG_NAME = "RealTimeHeightDeterministicDPDA";

RealTimeHeightDeterministicDPDA RealTimeHeightDeterministicDPDA::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, RealTimeHeightDeterministicDPDA::XML_TAG_NAME);

	std::set<State> states = AutomatonFromXMLParser::parseStates(input);
	std::set<alphabet::Symbol> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet(input);
	std::set<alphabet::Symbol> stackSymbols = AutomatonFromXMLParser::parseStackAlphabet(input);
	State initialState = AutomatonFromXMLParser::parseInitialState(input);
	alphabet::Symbol bottomOfTheStackSymbol = AutomatonFromXMLParser::parseBottomOfTheStackSymbol(input);
	std::set<State> finalStates = AutomatonFromXMLParser::parseFinalStates(input);

	RealTimeHeightDeterministicDPDA automaton(std::move(initialState), std::move(bottomOfTheStackSymbol));
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setStackAlphabet(std::move(stackSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<RealTimeHeightDeterministicDPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, RealTimeHeightDeterministicDPDA::XML_TAG_NAME);
	return automaton;
}

void RealTimeHeightDeterministicDPDA::parseTransition(std::deque<sax::Token>::iterator& input, RealTimeHeightDeterministicDPDA& automaton) {
	if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition");
		State from = AutomatonFromXMLParser::parseTransitionFrom(input);
		std::variant<string::Epsilon, alphabet::Symbol> inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol(input);
		State to = AutomatonFromXMLParser::parseTransitionTo(input);
		alphabet::Symbol push = AutomatonFromXMLParser::parseTransitionSinglePush(input);

		automaton.addCallTransition(std::move(from), std::move(inputSymbol), std::move(to), std::move(push));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "callTransition");
	} else if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition");
		State from = AutomatonFromXMLParser::parseTransitionFrom(input);
		std::variant<string::Epsilon, alphabet::Symbol> inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol(input);
		alphabet::Symbol pop = AutomatonFromXMLParser::parseTransitionSinglePop(input);
		State to = AutomatonFromXMLParser::parseTransitionTo(input);

		automaton.addReturnTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "returnTransition");
	} else {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "localTransition");
		State from = AutomatonFromXMLParser::parseTransitionFrom(input);
		std::variant<string::Epsilon, alphabet::Symbol> inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol(input);
		State to = AutomatonFromXMLParser::parseTransitionTo(input);

		automaton.addLocalTransition(std::move(from), std::move(inputSymbol), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "localTransition");
	}
}

void RealTimeHeightDeterministicDPDA::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(RealTimeHeightDeterministicDPDA::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composeStackAlphabet(out, this->getStackAlphabet());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeBottomOfTheStackSymbol(out, this->getBottomOfTheStackSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(RealTimeHeightDeterministicDPDA::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

void RealTimeHeightDeterministicDPDA::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getCallTransitions()) {
		out.emplace_back("callTransition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
		AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, transition.first.second);
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second.first);
		AutomatonToXMLComposer::composeTransitionSinglePush(out, transition.second.second);

		out.emplace_back("callTransition", sax::Token::TokenType::END_ELEMENT);
	}
	for(const auto& transition : this->getReturnTransitions()) {
		out.emplace_back("returnTransition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
		AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, std::get<1>(transition.first));
		AutomatonToXMLComposer::composeTransitionSinglePop(out, std::get<2>(transition.first));
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second);

		out.emplace_back("returnTransition", sax::Token::TokenType::END_ELEMENT);
	}
	for(const auto& transition : this->getLocalTransitions()) {
		out.emplace_back("localTransition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
		AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, transition.first.second);
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second);

		out.emplace_back("localTransition", sax::Token::TokenType::END_ELEMENT);
	}
	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace automaton */

namespace alib {

auto realTimeHeightDeterministicDPDAParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::RealTimeHeightDeterministicDPDA>();
auto realTimeHeightDeterministicDPDAParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::RealTimeHeightDeterministicDPDA>();

auto RealTimeHeightDeterministicDPDACastBinder = castApi::CastPoolStringBinder<automaton::RealTimeHeightDeterministicDPDA>(automaton::RealTimeHeightDeterministicDPDA::XML_TAG_NAME);

} /* namespace alib */
