/*
 * SinglePopDPDA.cpp
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#include "SinglePopDPDA.h"
#include "../AutomatonException.h"
#include <algorithm>
#include <sstream>
#include <ostream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace automaton {

SinglePopDPDA::SinglePopDPDA(State initialState, alphabet::Symbol initialPushdownSymbol) : SingleInitialSymbolPushdownStoreAlphabet(std::move(initialPushdownSymbol)), SingleInitialState(std::move(initialState)) {

}

AutomatonBase* SinglePopDPDA::clone() const {
	return new SinglePopDPDA(*this);
}

AutomatonBase* SinglePopDPDA::plunder() && {
	return new SinglePopDPDA(std::move(*this));
}

bool SinglePopDPDA::removeState(const State& state) {
	if (initialState == state) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (const std::pair<const std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, std::pair<State, std::vector<alphabet::Symbol> > >& transition : transitions) {
		if (state == std::get<0>(transition.first))
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		if(transition.second.first == state)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	return states.erase(state);
}

bool SinglePopDPDA::removeInputSymbol(const alphabet::Symbol& symbol) {
	for (const std::pair<const std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, std::pair<State, std::vector<alphabet::Symbol> > >& transition : transitions) {
		if (std::get<1>(transition.first).is<alphabet::Symbol>() && symbol == std::get<1>(transition.first).get<alphabet::Symbol>())
			throw AutomatonException("Symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	return inputAlphabet.erase(symbol);
}

bool SinglePopDPDA::removeStackSymbol(const alphabet::Symbol& symbol) {
	for (const std::pair<const std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, std::pair<State, std::vector<alphabet::Symbol> > >& transition : transitions) {
		if (symbol == std::get<2>(transition.first))
			throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");

		if (std::find(transition.second.second.begin(), transition.second.second.end(), symbol) != transition.second.second.end())
			throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	if(initialSymbol == symbol) {
		throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is start symbol.");
	}

	return stackAlphabet.erase(symbol);
}

bool SinglePopDPDA::addTransition(State from, std::variant<string::Epsilon, alphabet::Symbol> input, alphabet::Symbol pop, State to, std::vector<alphabet::Symbol> push) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (input.is<alphabet::Symbol>() && inputAlphabet.find(input.get<alphabet::Symbol>()) == inputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input.get<alphabet::Symbol>() + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	if (stackAlphabet.find(pop) == stackAlphabet.end()) {
		throw AutomatonException("Stack symbol \"" + (std::string) pop + "\" doesn't exist.");
	}

	for(const alphabet::Symbol& pushSymbol : push) {
		if (stackAlphabet.find(pushSymbol) == stackAlphabet.end()) {
			throw AutomatonException("Stack symbol \"" + (std::string) pushSymbol + "\" doesn't exist.");
		}
	}

	std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol> key(std::move(from), std::move(input), std::move(pop));
	std::pair<automaton::State, std::vector<alphabet::Symbol> > value = std::make_pair(std::move(to), std::move(push));

	if (transitions.find(key) != transitions.end()) {
		if(transitions.find(key)->second == value)
			return false;
		else if (input.is<alphabet::Symbol>())
			throw AutomatonException(
				"Transition (\"" + (std::string) std::get<0>(key).getName() + "\", \"" + (std::string) std::get<1>(key).get<alphabet::Symbol>() + "\", \"" + (std::string) std::get<2>(key) + "\") -> ?? already exists.");
		else
			throw AutomatonException(
				"Transition (\"" + (std::string) std::get<0>(key).getName() + "\", \"" + (std::string) std::get<1>(key).get<string::Epsilon>() + "\", \"" + (std::string) std::get<2>(key) + "\") -> ?? already exists.");
	}

	if(input.is<string::Epsilon>()) {
		if(std::any_of(transitions.begin(), transitions.end(), [&](const std::pair<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, std::pair<automaton::State, std::vector<alphabet::Symbol> > > & transition) {
				if(std::get<0>(transition.first) == std::get<0>(key) && std::get<2>(transition.first) == std::get<2>(key)) return true;
				return false;
			}))
				throw AutomatonException("Conflicting transition");
	} else {
		std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol> key2(std::get<0>(key), std::variant<string::Epsilon, alphabet::Symbol>{ string::Epsilon::EPSILON }, std::get<2>(key));
		if(transitions.find(key2) != transitions.end())
			throw AutomatonException("Conflicting transition");
	}

	transitions.insert(std::make_pair(std::move(key), std::move(value)) );
	return true;
}

bool SinglePopDPDA::addTransition(State from, alphabet::Symbol input, alphabet::Symbol pop, State to, std::vector<alphabet::Symbol> push) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(std::move(input));
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push));
}

bool SinglePopDPDA::addTransition(State from, alphabet::Symbol pop, State to, std::vector<alphabet::Symbol> push) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push));
}

bool SinglePopDPDA::removeTransition(const State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const alphabet::Symbol& pop, const State& to, const std::vector<alphabet::Symbol>& push) {
	std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol> key(from, input, pop);
	std::pair<automaton::State, std::vector<alphabet::Symbol> > value = std::make_pair(to, push);

	if (transitions.find(key) == transitions.end())
		return false;

	if(transitions.find(key)->second != value) {
		if (input.is<alphabet::Symbol>())
			throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input.get<alphabet::Symbol>() + "\") -> \"" + (std::string) to.getName() + "\" doesn't exist.");
		else
			throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input.get<string::Epsilon>() + "\") -> \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	transitions.erase(key);
	return true;
}

bool SinglePopDPDA::removeTransition(const State& from, const alphabet::Symbol& input, const alphabet::Symbol& pop, const State& to, const std::vector<alphabet::Symbol>& push) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(input);
	return removeTransition(from, inputVariant, pop, to, push);
}

bool SinglePopDPDA::removeTransition(const State& from, const alphabet::Symbol& pop, const State& to, const std::vector<alphabet::Symbol>& push) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return removeTransition(from, inputVariant, pop, to, push);
}

const std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, std::pair<State, std::vector<alphabet::Symbol> > >& SinglePopDPDA::getTransitions() const {
	return transitions;
}

int SinglePopDPDA::compare(const SinglePopDPDA& other) const {
	auto first = std::tie(states, inputAlphabet, initialState, finalStates, stackAlphabet, initialSymbol, transitions);
	auto second = std::tie(other.states, other.inputAlphabet, other.initialState, other.finalStates, other.stackAlphabet, other.initialSymbol, other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void SinglePopDPDA::operator>>(std::ostream& out) const {
	out << "(SinglePopDPDA "
		<< "states = " << states
		<< "inputAlphabet = " << inputAlphabet
		<< "initialState = " << initialState
		<< "finalStates = " << finalStates
		<< "stackAlphabet = " << stackAlphabet
		<< "initialSymbol = " << initialSymbol
		<< "transitions = " << transitions
		<< ")";
}

SinglePopDPDA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string SinglePopDPDA::XML_TAG_NAME = "SinglePopDPDA";

SinglePopDPDA SinglePopDPDA::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, SinglePopDPDA::XML_TAG_NAME);

	std::set<State> states = AutomatonFromXMLParser::parseStates(input);
	std::set<alphabet::Symbol> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet(input);
	std::set<alphabet::Symbol> stackSymbols = AutomatonFromXMLParser::parseStackAlphabet(input);
	State initialState = AutomatonFromXMLParser::parseInitialState(input);
	alphabet::Symbol initialStackSymbol = AutomatonFromXMLParser::parseInitialStackSymbol(input);
	std::set<State> finalStates = AutomatonFromXMLParser::parseFinalStates(input);

	SinglePopDPDA automaton(std::move(initialState), std::move(initialStackSymbol));
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setStackAlphabet(std::move(stackSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<SinglePopDPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, SinglePopDPDA::XML_TAG_NAME);
	return automaton;
}

void SinglePopDPDA::parseTransition(std::deque<sax::Token>::iterator& input, SinglePopDPDA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	State from = AutomatonFromXMLParser::parseTransitionFrom(input);
	std::variant<string::Epsilon, alphabet::Symbol> inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol(input);
	alphabet::Symbol pop = AutomatonFromXMLParser::parseTransitionSinglePop(input);
	State to = AutomatonFromXMLParser::parseTransitionTo(input);
	std::vector<alphabet::Symbol> push = AutomatonFromXMLParser::parseTransitionPush(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to), std::move(push));
}

void SinglePopDPDA::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(SinglePopDPDA::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composeStackAlphabet(out, this->getStackAlphabet());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeInitialStackSymbol(out, this->getInitialSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(SinglePopDPDA::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

void SinglePopDPDA::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getTransitions()) {
		out.emplace_back("transition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
		AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, std::get<1>(transition.first));
		AutomatonToXMLComposer::composeTransitionSinglePop(out, std::get<2>(transition.first));
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second.first);
		AutomatonToXMLComposer::composeTransitionPush(out, transition.second.second);

		out.emplace_back("transition", sax::Token::TokenType::END_ELEMENT);
	}

	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace automaton */

namespace alib {

auto singlePopDPDAParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::SinglePopDPDA>();
auto singlePopDPDAParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::SinglePopDPDA>();

} /* namespace alib */
