/*
 * VisiblyPushdownDPDA.cpp
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#include "VisiblyPushdownDPDA.h"
#include "../AutomatonException.h"
#include <algorithm>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace automaton {

VisiblyPushdownDPDA::VisiblyPushdownDPDA(State initialState, alphabet::Symbol bottomOfTheStackSymbol) : SingleInitialState(std::move(initialState)), BottomOfTheStackSymbolPushdownStoreAlphabet(std::move(bottomOfTheStackSymbol)) {

}

AutomatonBase* VisiblyPushdownDPDA::clone() const {
	return new VisiblyPushdownDPDA(*this);
}

AutomatonBase* VisiblyPushdownDPDA::plunder() && {
	return new VisiblyPushdownDPDA(std::move(*this));
}

bool VisiblyPushdownDPDA::removeState(const State& state) {
	if (initialState == state) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (const std::pair<const std::pair<State, alphabet::Symbol>, std::pair<State, alphabet::Symbol> >& callTransition : callTransitions) {
		if (state == callTransition.first.first)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		if(callTransition.second.first == state)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	for (const std::pair<const std::tuple<State, alphabet::Symbol, alphabet::Symbol>, State>& returnTransition : returnTransitions) {
		if (state == std::get<0>(returnTransition.first))
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		if(returnTransition.second == state)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	for (const std::pair<const std::pair<State, alphabet::Symbol>, State>& localTransition : localTransitions) {
		if (state == localTransition.first.first)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		if(localTransition.second == state)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	return states.erase(state);
}

bool VisiblyPushdownDPDA::removeInputSymbol(const alphabet::Symbol& symbol) {
	for (const std::pair<const std::pair<State, alphabet::Symbol>, std::pair<State, alphabet::Symbol> >& callTransition : callTransitions) {
		if (symbol == callTransition.first.second)
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	for (const std::pair<const std::tuple<State, alphabet::Symbol, alphabet::Symbol>, State>& returnTransition : returnTransitions) {
		if (symbol == std::get<1>(returnTransition.first))
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	for (const std::pair<const std::pair<State, alphabet::Symbol>, State>& localTransition : localTransitions) {
		if (symbol == localTransition.first.second)
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	return callInputAlphabet.erase(symbol) || returnInputAlphabet.erase(symbol) || localInputAlphabet.erase(symbol);
}

bool VisiblyPushdownDPDA::removeStackSymbol(const alphabet::Symbol& symbol) {
	for (const std::pair<const std::pair<State, alphabet::Symbol>, std::pair<State, alphabet::Symbol> >& callTransition : callTransitions) {
		if (symbol == callTransition.second.second)
			throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	for (const std::pair<const std::tuple<State, alphabet::Symbol, alphabet::Symbol>, State>& returnTransition : returnTransitions) {
		if (symbol == std::get<2>(returnTransition.first))
			throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	if(bottomOfTheStackSymbol == symbol) {
		throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is start symbol.");
	}

	return stackAlphabet.erase(symbol);
}

bool VisiblyPushdownDPDA::addCallTransition(State from, alphabet::Symbol input, State to, alphabet::Symbol push) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (callInputAlphabet.find(input) == callInputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	if (stackAlphabet.find(push) == stackAlphabet.end()) {
		throw AutomatonException("Stack symbol \"" + (std::string) push + "\" doesn't exist.");
	}

	std::pair<State, alphabet::Symbol> key(std::move(from), std::move(input));
	std::pair<automaton::State, alphabet::Symbol> value = std::make_pair(std::move(to), std::move(push));

	if(callTransitions.find(key) != callTransitions.end() && callTransitions.find(key)->second == value)
		return false;

	for(const auto& transition : callTransitions)
		if(transition.first.first == key.first && transition.first.second == key.second)
			throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");

	for(const auto& transition : returnTransitions)
		if(std::get<0>(transition.first) == key.first && std::get<1>(transition.first) == key.second)
			throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");

	for(const auto& transition : localTransitions)
		if(transition.first.first == key.first && transition.first.second == key.second)
			throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");

	callTransitions.insert(std::make_pair(std::move(key), std::move(value)));
	return true;
}

bool VisiblyPushdownDPDA::addReturnTransition(State from, alphabet::Symbol input, alphabet::Symbol pop, State to) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (returnInputAlphabet.find(input) == returnInputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	if (stackAlphabet.find(pop) == stackAlphabet.end()) {
		throw AutomatonException("Stack symbol \"" + (std::string) pop + "\" doesn't exist.");
	}

	std::tuple<State, alphabet::Symbol, alphabet::Symbol> key(std::move(from), std::move(input), std::move(pop));

	if(returnTransitions.find(key) != returnTransitions.end() && returnTransitions.find(key)->second == to)
		return false;

	for(const auto& transition : callTransitions)
		if(transition.first.first == std::get<0>(key) && transition.first.second == std::get<1>(key))
			throw AutomatonException("Can't add transition from state \"" + (std::string) std::get<0>(key) + "\" when transition reading \"" + (std::string) std::get<1>(key) + "\" is present.");

	for(const auto& transition : returnTransitions)
		if(std::get<0>(transition.first) == std::get<0>(key) && std::get<1>(transition.first) == std::get<1>(key) && std::get<2>(transition.first) == std::get<2>(key))
			throw AutomatonException("Can't add transition from state \"" + (std::string) from + "\" when transition reading \"" + (std::string) input + "\" is present.");

	for(const auto& transition : localTransitions)
		if(transition.first.first == std::get<0>(key) && transition.first.second == std::get<1>(key))
			throw AutomatonException("Can't add transition from state \"" + (std::string) std::get<0>(key) + "\" when transition reading \"" + (std::string) std::get<1>(key) + "\" is present.");

	returnTransitions.insert(std::make_pair(std::move(key), std::move(to)));
	return true;
}

bool VisiblyPushdownDPDA::addLocalTransition(State from, alphabet::Symbol input, State to) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (localInputAlphabet.find(input) == localInputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	std::pair<State, alphabet::Symbol> key(std::move(from), std::move(input));

	if(localTransitions.find(key) != localTransitions.end() && localTransitions.find(key)->second == to)
		return false;

	for(const auto& transition : callTransitions)
		if(transition.first.first == key.first && transition.first.second == key.second)
			throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");

	for(const auto& transition : returnTransitions)
		if(std::get<0>(transition.first) == key.first && std::get<1>(transition.first) == key.second)
			throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");

	for(const auto& transition : localTransitions)
		if(transition.first.first == key.first && transition.first.second == key.second)
			throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");

	localTransitions.insert(std::make_pair(std::move(key), std::move(to)));
	return true;
}

bool VisiblyPushdownDPDA::removeCallTransition(const State& from, const alphabet::Symbol& input, const State& to, const alphabet::Symbol& push) {
	std::pair<State, alphabet::Symbol> key(from, input);
	std::pair<automaton::State, alphabet::Symbol> value = std::make_pair(to, push);

	if (callTransitions.find(key) == callTransitions.end())
		return false;

	if(callTransitions.find(key)->second != value)
		throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input
						+ "\") -> \"" + (std::string) to.getName() + "\" doesn't exist.");

	callTransitions.erase(key);
	return true;
}

bool VisiblyPushdownDPDA::removeReturnTransition(const State& from, const alphabet::Symbol& input, const alphabet::Symbol& pop, const State& to) {
	std::tuple<State, alphabet::Symbol, alphabet::Symbol> key(from, input, pop);

	if (returnTransitions.find(key) == returnTransitions.end())
		return false;

	if(returnTransitions.find(key)->second != to)
		throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input
						+ "\") -> \"" + (std::string) to.getName() + "\" doesn't exist.");

	returnTransitions.erase(key);
	return true;
}

bool VisiblyPushdownDPDA::removeLocalTransition(const State& from, const alphabet::Symbol& input, const State& to) {
	std::pair<State, alphabet::Symbol> key(from, input);

	if (localTransitions.find(key) == localTransitions.end())
		return false;

	if(localTransitions.find(key)->second != to)
		throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input
						+ "\") -> \"" + (std::string) to.getName() + "\" doesn't exist.");

	localTransitions.erase(key);
	return true;
}

const std::map<std::pair<State, alphabet::Symbol>, std::pair<State, alphabet::Symbol> >& VisiblyPushdownDPDA::getCallTransitions() const {
	return callTransitions;
}

const std::map<std::tuple<State, alphabet::Symbol, alphabet::Symbol>, State>& VisiblyPushdownDPDA::getReturnTransitions() const {
	return returnTransitions;
}

const std::map<std::pair<State, alphabet::Symbol>, State>& VisiblyPushdownDPDA::getLocalTransitions() const {
	return localTransitions;
}

int VisiblyPushdownDPDA::compare(const VisiblyPushdownDPDA& other) const {
	auto first = std::tie(states, callInputAlphabet, returnInputAlphabet, localInputAlphabet, initialState, finalStates, stackAlphabet, bottomOfTheStackSymbol, callTransitions, returnTransitions, localTransitions);
	auto second = std::tie(other.states, other.callInputAlphabet, other.returnInputAlphabet, other.localInputAlphabet, other.initialState, other.finalStates, other.stackAlphabet, other.bottomOfTheStackSymbol, other.callTransitions, other.returnTransitions, other.localTransitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void VisiblyPushdownDPDA::operator>>(std::ostream& out) const {
	out << "(VisiblyPushdownDPDA "
		<< "states = " << states
		<< "callInputAlphabet = " << callInputAlphabet
		<< "returnInputAlphabet = " << returnInputAlphabet
		<< "localInputAlphabet = " << localInputAlphabet
		<< "initialState = " << initialState
		<< "finalStates = " << finalStates
		<< "stackAlphabet = " << stackAlphabet
		<< "bottomOfTheStackSymbol = " << bottomOfTheStackSymbol
		<< "callTransitions = " << callTransitions
		<< "returnTransitions = " << returnTransitions
		<< "localTransitions = " << localTransitions
		<< ")";
}

VisiblyPushdownDPDA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string VisiblyPushdownDPDA::XML_TAG_NAME = "VisiblyPushdownDPDA";

VisiblyPushdownDPDA VisiblyPushdownDPDA::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, VisiblyPushdownDPDA::XML_TAG_NAME);

	std::set<State> states = AutomatonFromXMLParser::parseStates(input);
	std::set<alphabet::Symbol> callInputAlphabet = AutomatonFromXMLParser::parseCallInputAlphabet(input);
	std::set<alphabet::Symbol> returnInputAlphabet = AutomatonFromXMLParser::parseReturnInputAlphabet(input);
	std::set<alphabet::Symbol> localInputAlphabet = AutomatonFromXMLParser::parseLocalInputAlphabet(input);
	std::set<alphabet::Symbol> stackSymbols = AutomatonFromXMLParser::parseStackAlphabet(input);
	State initialState = AutomatonFromXMLParser::parseInitialState(input);
	alphabet::Symbol bottomOfTheStackSymbol = AutomatonFromXMLParser::parseBottomOfTheStackSymbol(input);
	std::set<State> finalStates = AutomatonFromXMLParser::parseFinalStates(input);

	VisiblyPushdownDPDA automaton(std::move(initialState), std::move(bottomOfTheStackSymbol));
	automaton.setStates(std::move(states));
	automaton.setCallInputAlphabet(std::move(callInputAlphabet));
	automaton.setReturnInputAlphabet(std::move(returnInputAlphabet));
	automaton.setLocalInputAlphabet(std::move(localInputAlphabet));
	automaton.setStackAlphabet(std::move(stackSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<VisiblyPushdownDPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, VisiblyPushdownDPDA::XML_TAG_NAME);
	return automaton;
}

void VisiblyPushdownDPDA::parseTransition(std::deque<sax::Token>::iterator& input, VisiblyPushdownDPDA& automaton) {
	if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition");
		State from = AutomatonFromXMLParser::parseTransitionFrom(input);
		alphabet::Symbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol(input);
		State to = AutomatonFromXMLParser::parseTransitionTo(input);
		alphabet::Symbol push = AutomatonFromXMLParser::parseTransitionSinglePush(input);

		automaton.addCallTransition(std::move(from), std::move(inputSymbol), std::move(to), std::move(push));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "callTransition");
	} else if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition");
		State from = AutomatonFromXMLParser::parseTransitionFrom(input);
		alphabet::Symbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol(input);
		alphabet::Symbol pop = AutomatonFromXMLParser::parseTransitionSinglePop(input);
		State to = AutomatonFromXMLParser::parseTransitionTo(input);

		automaton.addReturnTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "returnTransition");
	} else {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "localTransition");
		State from = AutomatonFromXMLParser::parseTransitionFrom(input);
		alphabet::Symbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol(input);
		State to = AutomatonFromXMLParser::parseTransitionTo(input);

		automaton.addLocalTransition(std::move(from), std::move(inputSymbol), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "localTransition");
	}
}

void VisiblyPushdownDPDA::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(VisiblyPushdownDPDA::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeCallInputAlphabet(out, this->getCallInputAlphabet());
	AutomatonToXMLComposer::composeReturnInputAlphabet(out, this->getReturnInputAlphabet());
	AutomatonToXMLComposer::composeLocalInputAlphabet(out, this->getLocalInputAlphabet());
	AutomatonToXMLComposer::composeStackAlphabet(out, this->getStackAlphabet());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeBottomOfTheStackSymbol(out, this->getBottomOfTheStackSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(VisiblyPushdownDPDA::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

void VisiblyPushdownDPDA::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getCallTransitions()) {
		out.emplace_back("callTransition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
		AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.second);
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second.first);
		AutomatonToXMLComposer::composeTransitionSinglePush(out, transition.second.second);

		out.emplace_back("callTransition", sax::Token::TokenType::END_ELEMENT);
	}
	for(const auto& transition : this->getReturnTransitions()) {
		out.emplace_back("returnTransition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
		AutomatonToXMLComposer::composeTransitionInputSymbol(out, std::get<1>(transition.first));
		AutomatonToXMLComposer::composeTransitionSinglePop(out, std::get<2>(transition.first));
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second);

		out.emplace_back("returnTransition", sax::Token::TokenType::END_ELEMENT);
	}
	for(const auto& transition : this->getLocalTransitions()) {
		out.emplace_back("localTransition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
		AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.second);
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second);

		out.emplace_back("localTransition", sax::Token::TokenType::END_ELEMENT);
	}
	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace automaton */

namespace alib {

auto visiblyPushdownDPDAParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::VisiblyPushdownDPDA>();
auto visiblyPushdownDPDAParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::VisiblyPushdownDPDA>();

} /* namespace alib */
