/*
 * VisiblyPushdownNPDA.cpp
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#include "VisiblyPushdownNPDA.h"
#include "../AutomatonException.h"
#include <algorithm>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace automaton {

VisiblyPushdownNPDA::VisiblyPushdownNPDA(alphabet::Symbol bottomOfTheStackSymbol) : BottomOfTheStackSymbolPushdownStoreAlphabet(std::move(bottomOfTheStackSymbol)) {

}

AutomatonBase* VisiblyPushdownNPDA::clone() const {
	return new VisiblyPushdownNPDA(*this);
}

AutomatonBase* VisiblyPushdownNPDA::plunder() && {
	return new VisiblyPushdownNPDA(std::move(*this));
}

bool VisiblyPushdownNPDA::removeState(const State& state) {
	if (initialStates.find(state) != initialStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (const std::pair<const std::pair<State, alphabet::Symbol>, std::set<std::pair<State, alphabet::Symbol> > >& callTransition : callTransitions) {
		if (state == callTransition.first.first)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		for(const std::pair<State, alphabet::Symbol>& target : callTransition.second) {
			if(target.first == state)
				throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		}
	}

	for (const std::pair<const std::tuple<State, alphabet::Symbol, alphabet::Symbol>, std::set<State> >& returnTransition : returnTransitions) {
		if (state == std::get<0>(returnTransition.first))
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		for(const State& target : returnTransition.second) {
			if(target == state)
				throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		}
	}

	for (const std::pair<const std::pair<State, alphabet::Symbol>, std::set<State> >& localTransition : localTransitions) {
		if (state == localTransition.first.first)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		for(const State& target : localTransition.second) {
			if(target == state)
				throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		}
	}

	return states.erase(state);
}

bool VisiblyPushdownNPDA::removeInputSymbol(const alphabet::Symbol& symbol) {
	for (const std::pair<const std::pair<State, alphabet::Symbol>, std::set<std::pair<State, alphabet::Symbol> > >& callTransition : callTransitions) {
		if (symbol == callTransition.first.second)
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	for (const std::pair<const std::tuple<State, alphabet::Symbol, alphabet::Symbol>, std::set<State> >& returnTransition : returnTransitions) {
		if (symbol == std::get<1>(returnTransition.first))
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	for (const std::pair<const std::pair<State, alphabet::Symbol>, std::set<State> >& localTransition : localTransitions) {
		if (symbol == localTransition.first.second)
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	return callInputAlphabet.erase(symbol) || returnInputAlphabet.erase(symbol) || localInputAlphabet.erase(symbol);
}

bool VisiblyPushdownNPDA::removeStackSymbol(const alphabet::Symbol& symbol) {
	for (const std::pair<const std::pair<State, alphabet::Symbol>, std::set<std::pair<State, alphabet::Symbol> > >& callTransition : callTransitions) {
		for(const std::pair<State, alphabet::Symbol>& to : callTransition.second) {
			if (symbol == to.second)
				throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
		}
	}

	for (const std::pair<const std::tuple<State, alphabet::Symbol, alphabet::Symbol>, std::set<State> >& returnTransition : returnTransitions) {
		if (symbol == std::get<2>(returnTransition.first))
			throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	if(bottomOfTheStackSymbol == symbol) {
		throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is start symbol.");
	}

	return stackAlphabet.erase(symbol);
}

bool VisiblyPushdownNPDA::addCallTransition(State from, alphabet::Symbol input, State to, alphabet::Symbol push) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (callInputAlphabet.find(input) == callInputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	if (stackAlphabet.find(push) == stackAlphabet.end()) {
		throw AutomatonException("Stack symbol \"" + (std::string) push + "\" doesn't exist.");
	}

	std::pair<State, alphabet::Symbol> key(std::move(from), std::move(input));
	std::pair<automaton::State, alphabet::Symbol> value = std::make_pair(std::move(to), std::move(push));

	return callTransitions[std::move(key)].insert(std::move(value)).second;
}

bool VisiblyPushdownNPDA::addReturnTransition(State from, alphabet::Symbol input, alphabet::Symbol pop, State to) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (returnInputAlphabet.find(input) == returnInputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	if (stackAlphabet.find(pop) == stackAlphabet.end()) {
		throw AutomatonException("Stack symbol \"" + (std::string) pop + "\" doesn't exist.");
	}

	std::tuple<State, alphabet::Symbol, alphabet::Symbol> key(std::move(from), std::move(input), std::move(pop));

	return returnTransitions[std::move(key)].insert(std::move(to)).second;
}

bool VisiblyPushdownNPDA::addLocalTransition(State from, alphabet::Symbol input, State to) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (localInputAlphabet.find(input) == localInputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	std::pair<State, alphabet::Symbol> key(std::move(from), std::move(input));

	return localTransitions[std::move(key)].insert(std::move(to)).second;
}

bool VisiblyPushdownNPDA::removeCallTransition(const State& from, const alphabet::Symbol& input, const State& to, const alphabet::Symbol& push) {
	std::pair<State, alphabet::Symbol> key(from, input);
	std::pair<automaton::State, alphabet::Symbol> value = std::make_pair(to, push);

	return callTransitions[key].erase(value);
}

bool VisiblyPushdownNPDA::removeReturnTransition(const State& from, const alphabet::Symbol& input, const alphabet::Symbol& pop, const State& to) {
	std::tuple<State, alphabet::Symbol, alphabet::Symbol> key(from, input, pop);

	return returnTransitions[key].erase(to);
}

bool VisiblyPushdownNPDA::removeLocalTransition(const State& from, const alphabet::Symbol& input, const State& to) {
	std::pair<State, alphabet::Symbol> key(from, input);

	return localTransitions[key].erase(to);
}

const std::map<std::pair<State, alphabet::Symbol>, std::set<std::pair<State, alphabet::Symbol> > >& VisiblyPushdownNPDA::getCallTransitions() const {
	return callTransitions;
}

const std::map<std::tuple<State, alphabet::Symbol, alphabet::Symbol>, std::set<State> >& VisiblyPushdownNPDA::getReturnTransitions() const {
	return returnTransitions;
}

const std::map<std::pair<State, alphabet::Symbol>, std::set<State> >& VisiblyPushdownNPDA::getLocalTransitions() const {
	return localTransitions;
}

int VisiblyPushdownNPDA::compare(const VisiblyPushdownNPDA& other) const {
	auto first = std::tie(states, callInputAlphabet, returnInputAlphabet, localInputAlphabet, initialStates, finalStates, stackAlphabet, bottomOfTheStackSymbol, callTransitions, returnTransitions, localTransitions);
	auto second = std::tie(other.states, other.callInputAlphabet, other.returnInputAlphabet, other.localInputAlphabet, other.initialStates, other.finalStates, other.stackAlphabet, other.bottomOfTheStackSymbol, other.callTransitions, other.returnTransitions, other.localTransitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void VisiblyPushdownNPDA::operator>>(std::ostream& out) const {
	out << "(VisiblyPushdownNPDA "
		<< "states = " << states
		<< "callInputAlphabet = " << callInputAlphabet
		<< "returnInputAlphabet = " << returnInputAlphabet
		<< "localInputAlphabet = " << localInputAlphabet
		<< "initialStates = " << initialStates
		<< "finalStates = " << finalStates
		<< "stackAlphabet = " << stackAlphabet
		<< "bottomOfTheStackSymbol = " << bottomOfTheStackSymbol
		<< "callTransitions = " << callTransitions
		<< "returnTransitions = " << returnTransitions
		<< "localTransitions = " << localTransitions
		<< ")";
}

VisiblyPushdownNPDA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string VisiblyPushdownNPDA::XML_TAG_NAME = "VisiblyPushdownNPDA";

VisiblyPushdownNPDA VisiblyPushdownNPDA::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, VisiblyPushdownNPDA::XML_TAG_NAME);

	std::set<State> states = AutomatonFromXMLParser::parseStates(input);
	std::set<alphabet::Symbol> callInputAlphabet = AutomatonFromXMLParser::parseCallInputAlphabet(input);
	std::set<alphabet::Symbol> returnInputAlphabet = AutomatonFromXMLParser::parseReturnInputAlphabet(input);
	std::set<alphabet::Symbol> localInputAlphabet = AutomatonFromXMLParser::parseLocalInputAlphabet(input);
	std::set<alphabet::Symbol> stackSymbols = AutomatonFromXMLParser::parseStackAlphabet(input);
	std::set<State> initialStates = AutomatonFromXMLParser::parseInitialStates(input);
	alphabet::Symbol bottomOfTheStackSymbol = AutomatonFromXMLParser::parseBottomOfTheStackSymbol(input);
	std::set<State> finalStates = AutomatonFromXMLParser::parseFinalStates(input);

	VisiblyPushdownNPDA automaton(std::move(bottomOfTheStackSymbol));
	automaton.setStates(std::move(states));
	automaton.setCallInputAlphabet(std::move(callInputAlphabet));
	automaton.setReturnInputAlphabet(std::move(returnInputAlphabet));
	automaton.setLocalInputAlphabet(std::move(localInputAlphabet));
	automaton.setStackAlphabet(std::move(stackSymbols));
	automaton.setInitialStates(std::move(initialStates));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<VisiblyPushdownNPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, VisiblyPushdownNPDA::XML_TAG_NAME);
	return automaton;
}

void VisiblyPushdownNPDA::parseTransition(std::deque<sax::Token>::iterator& input, VisiblyPushdownNPDA& automaton) {
	if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition");
		State from = AutomatonFromXMLParser::parseTransitionFrom(input);
		alphabet::Symbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol(input);
		State to = AutomatonFromXMLParser::parseTransitionTo(input);
		alphabet::Symbol push = AutomatonFromXMLParser::parseTransitionSinglePush(input);

		automaton.addCallTransition(std::move(from), std::move(inputSymbol), std::move(to), std::move(push));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "callTransition");
	} else if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition");
		State from = AutomatonFromXMLParser::parseTransitionFrom(input);
		alphabet::Symbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol(input);
		alphabet::Symbol pop = AutomatonFromXMLParser::parseTransitionSinglePop(input);
		State to = AutomatonFromXMLParser::parseTransitionTo(input);

		automaton.addReturnTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "returnTransition");
	} else {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "localTransition");
		State from = AutomatonFromXMLParser::parseTransitionFrom(input);
		alphabet::Symbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol(input);
		State to = AutomatonFromXMLParser::parseTransitionTo(input);

		automaton.addLocalTransition(std::move(from), std::move(inputSymbol), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "localTransition");
	}
}

void VisiblyPushdownNPDA::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(VisiblyPushdownNPDA::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeCallInputAlphabet(out, this->getCallInputAlphabet());
	AutomatonToXMLComposer::composeReturnInputAlphabet(out, this->getReturnInputAlphabet());
	AutomatonToXMLComposer::composeLocalInputAlphabet(out, this->getLocalInputAlphabet());
	AutomatonToXMLComposer::composeStackAlphabet(out, this->getStackAlphabet());
	AutomatonToXMLComposer::composeInitialStates(out, this->getInitialStates());
	AutomatonToXMLComposer::composeBottomOfTheStackSymbol(out, this->getBottomOfTheStackSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(VisiblyPushdownNPDA::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

void VisiblyPushdownNPDA::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getCallTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("callTransition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
			AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.second);
			AutomatonToXMLComposer::composeTransitionTo(out, target.first);
			AutomatonToXMLComposer::composeTransitionSinglePush(out, target.second);

			out.emplace_back("callTransition", sax::Token::TokenType::END_ELEMENT);
		}
	}
	for(const auto& transition : this->getReturnTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("returnTransition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
			AutomatonToXMLComposer::composeTransitionInputSymbol(out, std::get<1>(transition.first));
			AutomatonToXMLComposer::composeTransitionSinglePop(out, std::get<2>(transition.first));
			AutomatonToXMLComposer::composeTransitionTo(out, target);

			out.emplace_back("returnTransition", sax::Token::TokenType::END_ELEMENT);
		}
	}
	for(const auto& transition : this->getLocalTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("localTransition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
			AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.second);
			AutomatonToXMLComposer::composeTransitionTo(out, target);

			out.emplace_back("localTransition", sax::Token::TokenType::END_ELEMENT);
		}
	}
	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace automaton */

namespace alib {

auto visiblyPushdownNPDAParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::VisiblyPushdownNPDA>();
auto visiblyPushdownNPDAParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::VisiblyPushdownNPDA>();

} /* namespace alib */
