/*
 * DFTA.cpp
 *
 *  Created on: Apr 14, 2015
 *      Author: Stepan Plachy
 */

#include "DFTA.h"
#include "../AutomatonException.h"
#include <ostream>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace automaton {

DFTA::DFTA() {

}

AutomatonBase* DFTA::clone() const {
	return new DFTA(*this);
}

AutomatonBase* DFTA::plunder() && {
	return new DFTA(std::move(*this));
}

bool DFTA::removeState(const State& state) {
	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	AutomatonException ex ("State \"" + (std::string) state.getName() + "\" is used in transition.");
	for (const std::pair<const std::pair<alphabet::RankedSymbol, std::vector<State> >, State>& t : transitions) {
		const std::vector<State>& states = t . first.second;
		for (const State& it : states) {
			if (it == state) throw ex;
		}
		if (t . second == state) throw ex;
	}

	return states.erase(state);
}

bool DFTA::removeInputSymbol(const alphabet::RankedSymbol& symbol) {
	for (const std::pair<const std::pair<alphabet::RankedSymbol, std::vector<State> >, State>& t : transitions) {
		if (t . first.first == symbol)
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used.");
	}

	return inputAlphabet.erase(symbol);
}

bool DFTA::addTransition(const alphabet::RankedSymbol & symbol, const std::vector<State> & prevStates, const State & next) {
	if (prevStates.size() != symbol.getRank().getData())
		throw AutomatonException("Number of states doesn't match rank of the symbol");

	if (inputAlphabet.find(symbol) == inputAlphabet.end())
		throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" doesn't exist.");

	if (states.find(next) == states.end())
		throw AutomatonException("State \"" + (std::string) next.getName() + "\" doesn't exist.");

	for (const State& it : prevStates) {
		if (states.find(it) == states.end())
			throw AutomatonException("State \"" + (std::string) it . getName() + "\" doesn't exist.");
	}

	std::pair<alphabet::RankedSymbol, std::vector<State> > key = std::make_pair(symbol, prevStates);
	const auto & it = transitions.find(key);
	if (it != transitions.end()) {
		if (it -> second == next) return false;
		else throw AutomatonException("Transition already exists");
	}

	transitions.insert(std::make_pair(key, next));
	return true;
}

bool DFTA::removeTransition(const alphabet::RankedSymbol symbol, const std::vector<State> & states, const State & next) {
	std::pair<alphabet::RankedSymbol, std::vector<State> > key = std::make_pair(symbol, states);
	const auto & it = transitions.find(key);
	if (it == transitions.end()) return false;
	if (it -> second != next) throw AutomatonException("Transition does not exist");
	return transitions.erase(key);
}

int DFTA::compare(const DFTA& other) const {
	auto first = std::tie(states, inputAlphabet, finalStates, transitions);
	auto second = std::tie(other.states, other.inputAlphabet, other.finalStates, other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void DFTA::operator>>(std::ostream& out) const {
	out << "(DFTA "
		<< " states = " << states
		<< " inputAlphabet = " << inputAlphabet
		<< " finalStates = " << finalStates
		<< " transitions = " << transitions
		<< ")";
}

DFTA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string DFTA::XML_TAG_NAME = "DFTA";

DFTA DFTA::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, DFTA::XML_TAG_NAME);

	std::set<State> states = AutomatonFromXMLParser::parseStates(input);
	std::set<alphabet::RankedSymbol> inputSymbols = AutomatonFromXMLParser::parseRankedInputAlphabet(input);
	std::set<State> finalStates = AutomatonFromXMLParser::parseFinalStates(input);

	DFTA automaton;
	automaton.setStates(std::move(states));
	automaton.setInputSymbols(std::move(inputSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<DFTA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, DFTA::XML_TAG_NAME);
	return automaton;
}

void DFTA::parseTransition(std::deque<sax::Token>::iterator& input, DFTA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	alphabet::RankedSymbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputRankedSymbol(input);
	std::vector<State> from = AutomatonFromXMLParser::parseTransitionFromMultiple(input);
	State to = AutomatonFromXMLParser::parseTransitionTo(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(inputSymbol), std::move(from), std::move(to));
}

void DFTA::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(DFTA::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeRankedInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(DFTA::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

void DFTA::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back(sax::Token("transitions", sax::Token::TokenType::START_ELEMENT));
	for(const auto& transition : this->getTransitions()) {
		out.emplace_back(sax::Token("transition", sax::Token::TokenType::START_ELEMENT));

		AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.first);
		AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.second);
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second);

		out.emplace_back(sax::Token("transition", sax::Token::TokenType::END_ELEMENT));
	}

	out.emplace_back(sax::Token("transitions", sax::Token::TokenType::END_ELEMENT));
}

} /* namespace automaton */

namespace alib {

auto DFTAParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::DFTA>();
auto DFTAParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::DFTA>();

} /* namespace alib */
