/*
 * NFTA.h
 *
 *  Created on: Mar 21, 2015
 *      Author: Stepan Plachy
 */

#ifndef NFTA_H_
#define NFTA_H_

#include <map>
#include <vector>
#include "../AutomatonBase.h"
#include "../common/States.h"

// #include "../common/InputAlphabet.h"
#include "../../alphabet/RankedSymbol.h"

#include "DFTA.h"

namespace automaton {

/**
 * Represents Finite Tree Automaton.
 * Can store nondeterministic finite tree automaton without epsilon transitions.
 */
class NFTA : public AutomatonBase, public States {
protected:
	std::set < alphabet::RankedSymbol > inputAlphabet;
	std::map < std::pair < alphabet::RankedSymbol, std::vector < State > >, std::set < State > > transitions;

public:
	explicit NFTA ( );
	explicit NFTA ( const DFTA & other );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	/**
	 * @copydoc Automaton::removeState(const State&)
	 */
	virtual bool removeState ( const State & state );

	/**
	 * @copydoc Automaton::removeInputSymbol(const Symbol&)
	 */
	virtual bool removeInputSymbol ( const alphabet::RankedSymbol & symbol );

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current symbol in node
	 * @param children states of children of current node
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition ( const alphabet::RankedSymbol & current, const std::vector < State > & children, const State & next );

	/**
	 * Removes transition from the automaton.
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition ( const alphabet::RankedSymbol symbol, const std::vector < State > & states, const State & next );

	/**
	 * @return automaton transitions
	 */
	const std::map < std::pair < alphabet::RankedSymbol, std::vector < State > >, std::set < State > > & getTransitions ( ) const { return transitions; }

	std::set < State > getTransitionRightSide ( const alphabet::RankedSymbol symbol, const std::vector < State > & states );

	/**
	 * Determines whether NFTA is deterministic.
	 * FA is deterministic if and only if:
	 * \li \c is epsilon free. Trivial for this class
	 * \li \c size of transition function \delta (from state, input symbol) \leq 1
	 * @return true when automaton is deterministic, false otherwise
	 */
	bool isDeterministic ( ) const;

	unsigned transitionsSize ( ) const;

	const std::set < alphabet::RankedSymbol > & getInputAlphabet ( ) const { return inputAlphabet; }

	void setInputSymbols ( const std::set < alphabet::RankedSymbol > & symbols ) { inputAlphabet = symbols; }

	void addInputSymbol ( const alphabet::RankedSymbol & symbol ) { inputAlphabet.insert ( symbol ); }

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const NFTA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual operator std::string ( ) const;

	const static std::string XML_TAG_NAME;

	static NFTA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, NFTA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;
};

} /* namespace automaton */

#endif /* NFTA_H_ */
