/*
 * OneTapeDTM.cpp
 *
 *  Created on: Apr 24, 2013
 *      Author: Martin Zak
 */

#include "OneTapeDTM.h"
#include "../AutomatonException.h"
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace automaton {

OneTapeDTM::OneTapeDTM(State initialState, alphabet::Symbol blank) : SingleInitialState(std::move(initialState)), BlankSymbolInputTapeAlphabet(std::move(blank)) {

}

AutomatonBase* OneTapeDTM::clone() const {
	return new OneTapeDTM(*this);
}

AutomatonBase* OneTapeDTM::plunder() && {
	return new OneTapeDTM(std::move(*this));
}

bool OneTapeDTM::removeState(const State& state) {
	if (initialState == state) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (const std::pair<const std::pair<State, alphabet::Symbol>, std::tuple<State, alphabet::Symbol, Shift> >& transition : transitions) {
		if (state == transition.first.first || state == std::get<0>(transition.second))
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in a transition.");
	}

	return states.erase(state);
}

bool OneTapeDTM::removeInputSymbol(const alphabet::Symbol& symbol) {
	return inputAlphabet.erase(symbol);
}

bool OneTapeDTM::removeTapeSymbol(const alphabet::Symbol& symbol) {
	if (inputAlphabet.find(symbol) != inputAlphabet.end()) {
		throw AutomatonException("Tape symbol \"" + (std::string) symbol + "\" is in input alphabet.");
	}

	for (const std::pair<const std::pair<State, alphabet::Symbol>, std::tuple<State, alphabet::Symbol, Shift> >& transition : transitions) {
		if (symbol == transition.first.second || symbol == std::get<1>(transition.second))
			throw AutomatonException("Tape symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	return inputAlphabet.erase(symbol);
}

bool OneTapeDTM::addTransition(State from, alphabet::Symbol input, State to, alphabet::Symbol output, Shift shift) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (input != blankSymbol)
		if (tapeAlphabet.find(input) == tapeAlphabet.end()) {
			throw AutomatonException("Tape symbol \"" + (std::string) input + "\" doesn't exist.");
		}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	if (input != blankSymbol)
		if (tapeAlphabet.find(output) == tapeAlphabet.end()) {
			throw AutomatonException("Tape symbol  \"" + (std::string) output + "\" doesn't exist.");
		}

	if(shift == Shift::NOT_SET)
		throw AutomatonException("Invalid shift.");

	std::pair<State, alphabet::Symbol> key = std::make_pair(std::move(from), std::move(input));

	std::tuple<State, alphabet::Symbol, Shift > value(std::move(to), std::move(output), shift);

	if (transitions.find(key) != transitions.end()) {
		if(transitions.find(key)->second == value)
			return false;
		else
			throw AutomatonException(
				"Transition (\"" + (std::string) key.first.getName() + "\", \"" + (std::string) key.second + "\") -> ? already exists.");
	}

	transitions.insert(std::make_pair(std::move(key), std::move(value)));
	return true;
}

bool OneTapeDTM::removeTransition(const State& from, const alphabet::Symbol& input, const State& to, const alphabet::Symbol& output, const Shift& shift) {
	std::pair<State, alphabet::Symbol> key = std::make_pair(from, input);
  
	if (transitions.find(key) == transitions.end())
		return false;
	
	std::tuple<State, alphabet::Symbol, Shift > value(to, output, shift);
	if(transitions.find(key)->second != value) 
		throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input + "\") -> ? doesn't exists.");

	transitions.erase(key);
	return true;
}

const std::map<std::pair<State, alphabet::Symbol>, std::tuple<State, alphabet::Symbol, Shift> >& OneTapeDTM::getTransitions() const {
	return transitions;
}

int OneTapeDTM::compare(const OneTapeDTM& other) const {
	auto first = std::tie(states, inputAlphabet, initialState, finalStates, tapeAlphabet, blankSymbol, transitions);
	auto second = std::tie(other.states, other.inputAlphabet, other.initialState, other.finalStates, other.tapeAlphabet, other.blankSymbol, other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void OneTapeDTM::operator>>(std::ostream& out) const {
	out << "(OneTapeDTM "
			<< "states = " << states
			<< "inputAlphabet = " << inputAlphabet
			<< "initialState = " << initialState
			<< "finalStates = " << finalStates
			<< "tapeAlphabet = " << tapeAlphabet
			<< "blankSymbol = " << blankSymbol
			<< "transitions = " << transitions
			<< ")";
}

OneTapeDTM::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string OneTapeDTM::XML_TAG_NAME = "OneTapeDTM";

OneTapeDTM OneTapeDTM::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, OneTapeDTM::XML_TAG_NAME);

	std::set<State> states = AutomatonFromXMLParser::parseStates(input);
	std::set<alphabet::Symbol> tapeSymbols = AutomatonFromXMLParser::parseTapeAlphabet(input);
	std::set<alphabet::Symbol> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet(input);
	alphabet::Symbol blank = AutomatonFromXMLParser::parseBlankSymbol(input);
	State initialState = AutomatonFromXMLParser::parseInitialState(input);
	std::set<State> finalStates = AutomatonFromXMLParser::parseFinalStates(input);

	OneTapeDTM automaton(std::move(initialState), std::move(blank));
	automaton.setStates(std::move(states));
	automaton.setTapeAlphabet(std::move(tapeSymbols));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<OneTapeDTM>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, OneTapeDTM::XML_TAG_NAME);
	return automaton;
}

void OneTapeDTM::parseTransition(std::deque<sax::Token>::iterator& input, OneTapeDTM& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	State from = AutomatonFromXMLParser::parseTransitionFrom(input);
	alphabet::Symbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol(input);
	State to = AutomatonFromXMLParser::parseTransitionTo(input);
	alphabet::Symbol outputSymbol = AutomatonFromXMLParser::parseTransitionOutputSymbol(input);
	Shift shift = AutomatonFromXMLParser::parseTransitionShift(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(from), std::move(inputSymbol), std::move(to), std::move(outputSymbol), shift);
}

void OneTapeDTM::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(OneTapeDTM::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeTapeAlphabet(out, this->getTapeAlphabet());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composeBlankSymbol(out, this->getBlankSymbol());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(OneTapeDTM::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

void OneTapeDTM::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getTransitions()) {
		out.emplace_back("transition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
		AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.second);
		AutomatonToXMLComposer::composeTransitionTo(out, std::get<0>(transition.second));
		AutomatonToXMLComposer::composeTransitionOutputSymbol(out, std::get<1>(transition.second));
		AutomatonToXMLComposer::composeTransitionShift(out, std::get<2>(transition.second));

		out.emplace_back("transition", sax::Token::TokenType::END_ELEMENT);
	}

	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace automaton */

namespace alib {

auto oneTapeDTMParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::OneTapeDTM>();
auto oneTapeDTMParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::OneTapeDTM>();

} /* namespace alib */
