/*
 * CFG.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "CFG.h"
#include "../GrammarException.h"
#include <algorithm>
#include <sstream>

#include "../../alphabet/Symbol.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../Grammar.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace grammar {

CFG::CFG ( alphabet::Symbol initialSymbol ) : CFG ( std::set < alphabet::Symbol > { initialSymbol }, std::set < alphabet::Symbol > ( ), initialSymbol ) {
}

CFG::CFG ( std::set < alphabet::Symbol > nonterminalAlphabet, std::set < alphabet::Symbol > terminalAlphabet, alphabet::Symbol initialSymbol ) : std::Components < CFG, alphabet::Symbol, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > ( std::make_tuple ( std::move ( terminalAlphabet ), std::move ( nonterminalAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ) ) {
}

CFG::CFG ( const EpsilonFreeCFG & other ) : CFG ( other.getNonterminalAlphabet ( ), other.getTerminalAlphabet ( ), other.getInitialSymbol ( ) ) {
	for ( const std::pair < const alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > & rule : other.getRules ( ) ) {
		const alphabet::Symbol & lhs = rule.first;

		for ( const std::vector < alphabet::Symbol > & rhs : rule.second )
			addRule ( lhs, rhs );
	}
}

GrammarBase * CFG::clone ( ) const {
	return new CFG ( * this );
}

GrammarBase * CFG::plunder ( ) && {
	return new CFG ( std::move ( * this ) );
}

bool CFG::addRule ( alphabet::Symbol leftHandSide, std::vector < alphabet::Symbol > rightHandSide ) {
	if ( !getNonterminalAlphabet ( ).count ( leftHandSide ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	for ( const alphabet::Symbol & symbol : rightHandSide )
		if ( !getTerminalAlphabet ( ).count ( symbol ) && !getNonterminalAlphabet ( ).count ( symbol ) )
			throw GrammarException ( "Symbol \"" + ( std::string ) symbol + "\" is not neither terminal nor nonterminal symbol" );

	return rules[std::move ( leftHandSide )].insert ( std::move ( rightHandSide ) ).second;
}

const std::map < alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > & CFG::getRules ( ) const {
	return rules;
}

bool CFG::addRawRule ( alphabet::Symbol leftHandSide, std::vector < alphabet::Symbol > rightHandSide ) {
	return addRule ( std::move ( leftHandSide ), std::move ( rightHandSide ) );
}

std::map < alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > CFG::getRawRules ( ) const {
	return rules;
}

bool CFG::removeRawRule ( const alphabet::Symbol & leftHandSide, const std::vector < alphabet::Symbol > & rightHandSide ) {
	return removeRule ( leftHandSide, rightHandSide );
}

bool CFG::removeRule ( const alphabet::Symbol & leftHandSide, const std::vector < alphabet::Symbol > & rightHandSide ) {
	return rules[leftHandSide].erase ( rightHandSide );
}

int CFG::compare ( const CFG & other ) const {
	auto first = std::tie ( getTerminalAlphabet ( ), getNonterminalAlphabet ( ), getInitialSymbol ( ), rules );
	auto second = std::tie ( other.getTerminalAlphabet ( ), other.getNonterminalAlphabet ( ), other.getInitialSymbol ( ), other.rules );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void CFG::operator >>( std::ostream & out ) const {
	out << "(CFG "
	    << "nonterminalAlphabet = " << getNonterminalAlphabet ( )
	    << "terminalAlphabet = " << getTerminalAlphabet ( )
	    << "initialSymbol = " << getInitialSymbol ( )
	    << "rules = " << rules << ")";
}

CFG::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

const std::string CFG::XML_TAG_NAME = "CFG";

CFG CFG::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, CFG::XML_TAG_NAME );

	std::set < alphabet::Symbol > nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet ( input );
	std::set < alphabet::Symbol > terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet ( input );
	alphabet::Symbol initialSymbol = GrammarFromXMLParser::parseInitialSymbol ( input );

	CFG grammar ( std::move ( initialSymbol ) );

	grammar.setNonterminalAlphabet ( std::move ( nonterminalAlphabet ) );
	grammar.setTerminalAlphabet ( std::move ( terminalAlphabet ) );

	GrammarFromXMLParser::parseRules ( input, grammar );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, CFG::XML_TAG_NAME );
	return grammar;
}

void CFG::parseRule ( std::deque < sax::Token >::iterator & input, CFG & grammar ) {
	alphabet::Symbol lhs = GrammarFromXMLParser::parseRuleSingleSymbolLHS ( input );
	std::vector < alphabet::Symbol > rhs = GrammarFromXMLParser::parseRuleRHS ( input );

	grammar.addRule ( std::move ( lhs ), std::move ( rhs ) );
}

void CFG::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( CFG::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT );

	GrammarToXMLComposer::composeNonterminalAlphabet ( out, this->getNonterminalAlphabet ( ) );
	GrammarToXMLComposer::composeTerminalAlphabet ( out, this->getTerminalAlphabet ( ) );
	GrammarToXMLComposer::composeInitialSymbol ( out, this->getInitialSymbol ( ) );
	composeRules ( out );

	out.emplace_back ( CFG::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT );
}

void CFG::composeRules ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "rules", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & rule : this->getRules ( ) )

		for ( const auto & rhs : rule.second ) {
			out.emplace_back ( "rule", sax::Token::TokenType::START_ELEMENT );

			GrammarToXMLComposer::composeRuleSingleSymbolLHS ( out, rule.first );
			GrammarToXMLComposer::composeRuleRHS ( out, rhs );

			out.emplace_back ( "rule", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "rules", sax::Token::TokenType::END_ELEMENT );
}

} /* namespace grammar */

namespace std {

template < >
bool grammar::CFG::Component < grammar::CFG, alphabet::Symbol, grammar::TerminalAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	const grammar::CFG * grammar = static_cast < const grammar::CFG * > ( this );

	for ( const std::pair < const alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > & rule : grammar->getRules ( ) )
		for ( const std::vector < alphabet::Symbol > & rhs : rule.second )
			if ( std::find ( rhs.begin ( ), rhs.end ( ), symbol ) != rhs.end ( ) )
				return true;

	return false;
}

template < >
bool grammar::CFG::Component < grammar::CFG, alphabet::Symbol, grammar::TerminalAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void grammar::CFG::Component < grammar::CFG, alphabet::Symbol, grammar::TerminalAlphabet >::valid ( const alphabet::Symbol & symbol ) const {
	const grammar::CFG * grammar = static_cast < const grammar::CFG * > ( this );

	if ( grammar->accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol ) )
		throw grammar::GrammarException ( "Symbol " + ( std::string ) symbol + "cannot be in terminal alphabet since it is already nonterminal alphabet" );
}

template < >
bool grammar::CFG::Component < grammar::CFG, alphabet::Symbol, grammar::NonterminalAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	const grammar::CFG * grammar = static_cast < const grammar::CFG * > ( this );

	for ( const std::pair < const alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > & rule : grammar->getRules ( ) ) {
		if ( rule.first == symbol )
			return true;

		for ( const std::vector < alphabet::Symbol > & rhs : rule.second )
			if ( std::find ( rhs.begin ( ), rhs.end ( ), symbol ) != rhs.end ( ) )
				return true;

	}

	if ( grammar->accessElement < grammar::InitialSymbol > ( ).get ( ) == symbol )
		return true;

	return false;
}

template < >
bool grammar::CFG::Component < grammar::CFG, alphabet::Symbol, grammar::NonterminalAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void grammar::CFG::Component < grammar::CFG, alphabet::Symbol, grammar::NonterminalAlphabet >::valid ( const alphabet::Symbol & symbol ) const {
	const grammar::CFG * grammar = static_cast < const grammar::CFG * > ( this );

	if ( grammar->accessComponent < grammar::TerminalAlphabet > ( ).get ( ).count ( symbol ) )
		throw grammar::GrammarException ( "Symbol " + ( std::string ) symbol + "cannot be in nonterminal alphabet since it is already in terminal alphabet" );
}

template < >
bool grammar::CFG::Element < grammar::CFG, alphabet::Symbol, grammar::InitialSymbol >::available ( const alphabet::Symbol & symbol ) const {
	const grammar::CFG * grammar = static_cast < const grammar::CFG * > ( this );

	return grammar->accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol );
}

template < >
void grammar::CFG::Element < grammar::CFG, alphabet::Symbol, grammar::InitialSymbol >::valid ( const alphabet::Symbol & ) const {
}

} /* namespace std */

namespace alib {

auto CFGParserRegister = xmlApi < grammar::Grammar >::ParserRegister < grammar::CFG > ( );
auto CFGParserRegister2 = xmlApi < alib::Object >::ParserRegister < grammar::CFG > ( );

auto CFGEpsilonFreeCFG = castApi::CastRegister < grammar::CFG, grammar::EpsilonFreeCFG > ( );
auto CFGCastBinder = castApi::CastPoolStringBinder < grammar::CFG > ( grammar::CFG::XML_TAG_NAME );

} /* namespace alib */
