/*
 * EpsilonFreeCFG.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "EpsilonFreeCFG.h"
#include "../GrammarException.h"
#include <algorithm>
#include <sstream>

#include "../../alphabet/Symbol.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../Grammar.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace grammar {

EpsilonFreeCFG::EpsilonFreeCFG ( alphabet::Symbol initialSymbol ) : EpsilonFreeCFG ( std::set < alphabet::Symbol > { initialSymbol }, std::set < alphabet::Symbol > ( ), initialSymbol ) {
}

EpsilonFreeCFG::EpsilonFreeCFG ( std::set < alphabet::Symbol > nonterminalAlphabet, std::set < alphabet::Symbol > terminalAlphabet, alphabet::Symbol initialSymbol ) : std::Components < EpsilonFreeCFG, alphabet::Symbol, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > ( std::make_tuple ( std::move ( terminalAlphabet ), std::move ( nonterminalAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ) ), generatesEpsilon ( false ) {
}

GrammarBase * EpsilonFreeCFG::clone ( ) const {
	return new EpsilonFreeCFG ( * this );
}

GrammarBase * EpsilonFreeCFG::plunder ( ) && {
	return new EpsilonFreeCFG ( std::move ( * this ) );
}

bool EpsilonFreeCFG::addRule ( alphabet::Symbol leftHandSide, std::vector < alphabet::Symbol > rightHandSide ) {
	int rSize = rightHandSide.size ( );

	if ( rSize == 0 )
		throw GrammarException ( "Epsilon rule is not allowed" );

	if ( !getNonterminalAlphabet ( ).count ( leftHandSide ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	for ( const alphabet::Symbol & symbol : rightHandSide )
		if ( !getTerminalAlphabet ( ).count ( symbol ) && !getNonterminalAlphabet ( ).count ( symbol ) )
			throw GrammarException ( "Symbol \"" + ( std::string ) symbol + "\" is not neither terminal nor nonterminal symbol" );

	return rules[std::move ( leftHandSide )].insert ( std::move ( rightHandSide ) ).second;
}

const std::map < alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > & EpsilonFreeCFG::getRules ( ) const {
	return rules;
}

bool EpsilonFreeCFG::removeRule ( const alphabet::Symbol & leftHandSide, const std::vector < alphabet::Symbol > & rightHandSide ) {
	return rules[leftHandSide].erase ( rightHandSide );
}

bool EpsilonFreeCFG::addRawRule ( alphabet::Symbol leftHandSide, std::vector < alphabet::Symbol > rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 ) {
		if ( leftHandSide != getInitialSymbol ( ) ) throw GrammarException ( "Illegal left hand side of epsilon rule" );

		bool res = getGeneratesEpsilon ( );
		setGeneratesEpsilon ( true );
		return !res;
	} else {
		return addRule ( std::move ( leftHandSide ), std::move ( rightHandSide ) );
	}
}

std::map < alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > EpsilonFreeCFG::getRawRules ( ) const {
	std::map < alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > res ( getRules ( ).begin ( ), getRules ( ).end ( ) );

	if ( getGeneratesEpsilon ( ) )
		res [ getInitialSymbol ( ) ].insert ( std::vector < alphabet::Symbol > { } );

	return rules;
}

bool EpsilonFreeCFG::removeRawRule ( const alphabet::Symbol & leftHandSide, const std::vector < alphabet::Symbol > & rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 ) {
		if ( leftHandSide != getInitialSymbol ( ) ) throw GrammarException ( "Illegal left hand side of epsilon rule" );

		bool res = getGeneratesEpsilon ( );
		setGeneratesEpsilon ( false );
		return res;
	} else {
		return removeRule ( leftHandSide, rightHandSide );
	}
}

void EpsilonFreeCFG::setGeneratesEpsilon ( bool genEps ) {
	generatesEpsilon = genEps;
}

bool EpsilonFreeCFG::getGeneratesEpsilon ( ) const {
	return generatesEpsilon;
}

int EpsilonFreeCFG::compare ( const EpsilonFreeCFG & other ) const {
	auto first = std::tie ( getTerminalAlphabet ( ), getNonterminalAlphabet ( ), getInitialSymbol ( ), rules );
	auto second = std::tie ( other.getTerminalAlphabet ( ), other.getNonterminalAlphabet ( ), other.getInitialSymbol ( ), other.rules );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void EpsilonFreeCFG::operator >>( std::ostream & out ) const {
	out << "(EpsilonFreeCFG "
	    << "nonterminalAlphabet = " << getNonterminalAlphabet ( )
	    << "terminalAlphabet = " << getTerminalAlphabet ( )
	    << "initialSymbol = " << getInitialSymbol ( )
	    << "rules = " << rules
	    << "generatesEpsilon = " << generatesEpsilon << ")";
}

EpsilonFreeCFG::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

const std::string EpsilonFreeCFG::XML_TAG_NAME = "EpsilonFreeCFG";

EpsilonFreeCFG EpsilonFreeCFG::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, EpsilonFreeCFG::XML_TAG_NAME );

	std::set < alphabet::Symbol > nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet ( input );
	std::set < alphabet::Symbol > terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet ( input );
	alphabet::Symbol initialSymbol = GrammarFromXMLParser::parseInitialSymbol ( input );

	EpsilonFreeCFG grammar ( std::move ( initialSymbol ) );

	grammar.setNonterminalAlphabet ( std::move ( nonterminalAlphabet ) );
	grammar.setTerminalAlphabet ( std::move ( terminalAlphabet ) );

	GrammarFromXMLParser::parseRules ( input, grammar );

	bool generatesEpsilon = GrammarFromXMLParser::parseGeneratesEpsilon ( input );
	grammar.setGeneratesEpsilon ( generatesEpsilon );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, EpsilonFreeCFG::XML_TAG_NAME );
	return grammar;
}

void EpsilonFreeCFG::parseRule ( std::deque < sax::Token >::iterator & input, EpsilonFreeCFG & grammar ) {
	alphabet::Symbol lhs = GrammarFromXMLParser::parseRuleSingleSymbolLHS ( input );
	std::vector < alphabet::Symbol > rhs = GrammarFromXMLParser::parseRuleRHS ( input );

	grammar.addRule ( std::move ( lhs ), std::move ( rhs ) );
}

void EpsilonFreeCFG::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( EpsilonFreeCFG::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT );

	GrammarToXMLComposer::composeNonterminalAlphabet ( out, this->getNonterminalAlphabet ( ) );
	GrammarToXMLComposer::composeTerminalAlphabet ( out, this->getTerminalAlphabet ( ) );
	GrammarToXMLComposer::composeInitialSymbol ( out, this->getInitialSymbol ( ) );
	composeRules ( out );
	GrammarToXMLComposer::composeGeneratesEpsilon ( out, this->getGeneratesEpsilon ( ) );

	out.emplace_back ( EpsilonFreeCFG::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT );
}

void EpsilonFreeCFG::composeRules ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "rules", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & rule : this->getRules ( ) )

		for ( const auto & rhs : rule.second ) {
			out.emplace_back ( "rule", sax::Token::TokenType::START_ELEMENT );

			GrammarToXMLComposer::composeRuleSingleSymbolLHS ( out, rule.first );
			GrammarToXMLComposer::composeRuleRHS ( out, rhs );

			out.emplace_back ( "rule", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "rules", sax::Token::TokenType::END_ELEMENT );
}

} /* namespace grammar */

namespace std {

template < >
bool grammar::EpsilonFreeCFG::Component < grammar::EpsilonFreeCFG, alphabet::Symbol, grammar::TerminalAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	const grammar::EpsilonFreeCFG * grammar = static_cast < const grammar::EpsilonFreeCFG * > ( this );

	for ( const std::pair < const alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > & rule : grammar->getRules ( ) )
		for ( const std::vector < alphabet::Symbol > & rhs : rule.second )
			if ( std::find ( rhs.begin ( ), rhs.end ( ), symbol ) != rhs.end ( ) )
				return true;

	return false;
}

template < >
bool grammar::EpsilonFreeCFG::Component < grammar::EpsilonFreeCFG, alphabet::Symbol, grammar::TerminalAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void grammar::EpsilonFreeCFG::Component < grammar::EpsilonFreeCFG, alphabet::Symbol, grammar::TerminalAlphabet >::valid ( const alphabet::Symbol & symbol ) const {
	const grammar::EpsilonFreeCFG * grammar = static_cast < const grammar::EpsilonFreeCFG * > ( this );

	if ( grammar->accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol ) )
		throw grammar::GrammarException ( "Symbol " + ( std::string ) symbol + "cannot be in terminal alphabet since it is already nonterminal alphabet" );
}

template < >
bool grammar::EpsilonFreeCFG::Component < grammar::EpsilonFreeCFG, alphabet::Symbol, grammar::NonterminalAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	const grammar::EpsilonFreeCFG * grammar = static_cast < const grammar::EpsilonFreeCFG * > ( this );

	for ( const std::pair < const alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > & rule : grammar->getRules ( ) ) {
		if ( rule.first == symbol )
			return true;

		for ( const std::vector < alphabet::Symbol > & rhs : rule.second )
			if ( std::find ( rhs.begin ( ), rhs.end ( ), symbol ) != rhs.end ( ) )
				return true;

	}

	if ( grammar->accessElement < grammar::InitialSymbol > ( ).get ( ) == symbol )
		return true;

	return false;
}

template < >
bool grammar::EpsilonFreeCFG::Component < grammar::EpsilonFreeCFG, alphabet::Symbol, grammar::NonterminalAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void grammar::EpsilonFreeCFG::Component < grammar::EpsilonFreeCFG, alphabet::Symbol, grammar::NonterminalAlphabet >::valid ( const alphabet::Symbol & symbol ) const {
	const grammar::EpsilonFreeCFG * grammar = static_cast < const grammar::EpsilonFreeCFG * > ( this );

	if ( grammar->accessComponent < grammar::TerminalAlphabet > ( ).get ( ).count ( symbol ) )
		throw grammar::GrammarException ( "Symbol " + ( std::string ) symbol + "cannot be in nonterminal alphabet since it is already in terminal alphabet" );
}

template < >
bool grammar::EpsilonFreeCFG::Element < grammar::EpsilonFreeCFG, alphabet::Symbol, grammar::InitialSymbol >::available ( const alphabet::Symbol & symbol ) const {
	const grammar::EpsilonFreeCFG * grammar = static_cast < const grammar::EpsilonFreeCFG * > ( this );

	return grammar->accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol );
}

template < >
void grammar::EpsilonFreeCFG::Element < grammar::EpsilonFreeCFG, alphabet::Symbol, grammar::InitialSymbol >::valid ( const alphabet::Symbol & ) const {
}

} /* namespace std */

namespace alib {

auto EpsilonFreeCFGParserRegister = xmlApi < grammar::Grammar >::ParserRegister < grammar::EpsilonFreeCFG > ( );
auto EpsilonFreeCFGParserRegister2 = xmlApi < alib::Object >::ParserRegister < grammar::EpsilonFreeCFG > ( );

} /* namespace alib */
