/*
 * GNF.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "GNF.h"
#include "../GrammarException.h"
#include <algorithm>
#include <sstream>

#include "../../alphabet/Symbol.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../Grammar.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace grammar {

GNF::GNF ( alphabet::Symbol initialSymbol ) : GNF ( std::set < alphabet::Symbol > { initialSymbol }, std::set < alphabet::Symbol > ( ), initialSymbol ) {
}

GNF::GNF ( std::set < alphabet::Symbol > nonterminalAlphabet, std::set < alphabet::Symbol > terminalAlphabet, alphabet::Symbol initialSymbol ) : std::Components < GNF, alphabet::Symbol, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > ( std::make_tuple ( std::move ( terminalAlphabet ), std::move ( nonterminalAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ) ), generatesEpsilon ( false ) {
}

GrammarBase * GNF::clone ( ) const {
	return new GNF ( * this );
}

GrammarBase * GNF::plunder ( ) && {
	return new GNF ( std::move ( * this ) );
}

bool GNF::addRule ( alphabet::Symbol leftHandSide, std::pair < alphabet::Symbol, std::vector < alphabet::Symbol > > rightHandSide ) {
	if ( !getNonterminalAlphabet ( ).count ( leftHandSide ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	if ( !getTerminalAlphabet ( ).count ( rightHandSide.first ) )
		throw GrammarException ( "Rule must rewrite to terminal symbol" );

	for ( const alphabet::Symbol & rhsNTs : rightHandSide.second )
		if ( !getNonterminalAlphabet ( ).count ( rhsNTs ) )
			throw GrammarException ( "Symbol \"" + ( std::string ) rhsNTs + "\" is not a nonterminal symbol" );

	return rules[std::move ( leftHandSide )].insert ( std::move ( rightHandSide ) ).second;
}

const std::map < alphabet::Symbol, std::set < std::pair < alphabet::Symbol, std::vector < alphabet::Symbol > > > > & GNF::getRules ( ) const {
	return rules;
}

bool GNF::removeRule ( const alphabet::Symbol & leftHandSide, const std::pair < alphabet::Symbol, std::vector < alphabet::Symbol > > & rightHandSide ) {
	return rules[leftHandSide].erase ( rightHandSide );
}

bool GNF::addRawRule ( alphabet::Symbol leftHandSide, std::vector < alphabet::Symbol > rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 ) {
		if ( leftHandSide != getInitialSymbol ( ) ) throw GrammarException ( "Illegal left hand side of epsilon rule" );

		bool res = getGeneratesEpsilon ( );
		setGeneratesEpsilon ( true );
		return !res;
	} else {
		alphabet::Symbol first = std::move ( rightHandSide[0] );
		std::vector < alphabet::Symbol > rest ( std::make_move_iterator ( rightHandSide.begin ( ) ) + 1, std::make_move_iterator ( rightHandSide.end ( ) ) );
		return addRule ( std::move ( leftHandSide ), std::make_pair ( std::move ( first ), std::move ( rest ) ) );
	}
}

std::map < alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > GNF::getRawRules ( ) const {
	std::map < alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > res;

	for ( const auto & rule : getRules ( ) )
		for ( const auto & rhs : rule.second ) {
			std::vector < alphabet::Symbol > tmp { rhs.first };
			tmp.insert ( tmp.end ( ), rhs.second.begin ( ), rhs.second.end ( ) );
			res[rule.first].insert ( std::move ( tmp ) );
		}

	if ( getGeneratesEpsilon ( ) )
		res[getInitialSymbol ( )].insert ( std::vector < alphabet::Symbol > { } );

	return res;
}

bool GNF::removeRawRule ( const alphabet::Symbol & leftHandSide, const std::vector < alphabet::Symbol > & rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 ) {
		if ( leftHandSide != getInitialSymbol ( ) ) throw GrammarException ( "Illegal left hand side of epsilon rule" );

		bool res = getGeneratesEpsilon ( );
		setGeneratesEpsilon ( false );
		return res;
	} else {
		alphabet::Symbol first = rightHandSide[0];
		std::vector < alphabet::Symbol > rest ( rightHandSide.begin ( ) + 1, rightHandSide.end ( ) );
		return removeRule ( leftHandSide, std::make_pair ( first, rest ) );
	}
}

void GNF::setGeneratesEpsilon ( bool genEps ) {
	generatesEpsilon = genEps;
}

bool GNF::getGeneratesEpsilon ( ) const {
	return generatesEpsilon;
}

int GNF::compare ( const GNF & other ) const {
	auto first = std::tie ( getTerminalAlphabet ( ), getNonterminalAlphabet ( ), getInitialSymbol ( ), rules );
	auto second = std::tie ( other.getTerminalAlphabet ( ), other.getNonterminalAlphabet ( ), other.getInitialSymbol ( ), other.rules );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void GNF::operator >>( std::ostream & out ) const {
	out << "(GNF "
	    << "nonterminalAlphabet = " << getNonterminalAlphabet ( )
	    << "terminalAlphabet = " << getTerminalAlphabet ( )
	    << "initialSymbol = " << getInitialSymbol ( )
	    << "rules = " << rules << ")";
}

GNF::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

const std::string GNF::XML_TAG_NAME = "GNF";

GNF GNF::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, GNF::XML_TAG_NAME );

	std::set < alphabet::Symbol > nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet ( input );
	std::set < alphabet::Symbol > terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet ( input );
	alphabet::Symbol initialSymbol = GrammarFromXMLParser::parseInitialSymbol ( input );

	GNF grammar ( std::move ( initialSymbol ) );

	grammar.setNonterminalAlphabet ( std::move ( nonterminalAlphabet ) );
	grammar.setTerminalAlphabet ( std::move ( terminalAlphabet ) );

	GrammarFromXMLParser::parseRules ( input, grammar );

	bool generatesEpsilon = GrammarFromXMLParser::parseGeneratesEpsilon ( input );
	grammar.setGeneratesEpsilon ( generatesEpsilon );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, GNF::XML_TAG_NAME );
	return grammar;
}

void GNF::parseRule ( std::deque < sax::Token >::iterator & input, GNF & grammar ) {
	alphabet::Symbol lhs = GrammarFromXMLParser::parseRuleSingleSymbolLHS ( input );
	std::pair < alphabet::Symbol, std::vector < alphabet::Symbol > > rhs = GrammarFromXMLParser::parseRuleGNFRHS ( input );

	grammar.addRule ( std::move ( lhs ), std::move ( rhs ) );
}

void GNF::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( GNF::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT );

	GrammarToXMLComposer::composeNonterminalAlphabet ( out, this->getNonterminalAlphabet ( ) );
	GrammarToXMLComposer::composeTerminalAlphabet ( out, this->getTerminalAlphabet ( ) );
	GrammarToXMLComposer::composeInitialSymbol ( out, this->getInitialSymbol ( ) );
	composeRules ( out );
	GrammarToXMLComposer::composeGeneratesEpsilon ( out, this->getGeneratesEpsilon ( ) );

	out.emplace_back ( GNF::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT );
}

void GNF::composeRules ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "rules", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & rule : this->getRules ( ) )

		for ( const auto & rhs : rule.second ) {
			out.emplace_back ( "rule", sax::Token::TokenType::START_ELEMENT );

			GrammarToXMLComposer::composeRuleSingleSymbolLHS ( out, rule.first );
			GrammarToXMLComposer::composeRuleGNFRHS ( out, rhs );

			out.emplace_back ( "rule", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "rules", sax::Token::TokenType::END_ELEMENT );
}

} /* namespace grammar */

namespace std {

template < >
bool grammar::GNF::Component < grammar::GNF, alphabet::Symbol, grammar::TerminalAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	const grammar::GNF * grammar = static_cast < const grammar::GNF * > ( this );

	for ( const std::pair < const alphabet::Symbol, std::set < std::pair < alphabet::Symbol, std::vector < alphabet::Symbol > > > > & rule : grammar->getRules ( ) )
		for ( const std::pair < alphabet::Symbol, std::vector < alphabet::Symbol > > & rhs : rule.second )
			if ( rhs.first == symbol )
				return true;

	return false;
}

template < >
bool grammar::GNF::Component < grammar::GNF, alphabet::Symbol, grammar::TerminalAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void grammar::GNF::Component < grammar::GNF, alphabet::Symbol, grammar::TerminalAlphabet >::valid ( const alphabet::Symbol & symbol ) const {
	const grammar::GNF * grammar = static_cast < const grammar::GNF * > ( this );

	if ( grammar->accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol ) )
		throw grammar::GrammarException ( "Symbol " + ( std::string ) symbol + "cannot be in terminal alphabet since it is already nonterminal alphabet" );
}

template < >
bool grammar::GNF::Component < grammar::GNF, alphabet::Symbol, grammar::NonterminalAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	const grammar::GNF * grammar = static_cast < const grammar::GNF * > ( this );

	for ( const std::pair < const alphabet::Symbol, std::set < std::pair < alphabet::Symbol, std::vector < alphabet::Symbol > > > > & rule : grammar->getRules ( ) ) {
		if ( rule.first == symbol )
			return true;

		for ( const std::pair < alphabet::Symbol, std::vector < alphabet::Symbol > > & rhs : rule.second )
			if ( std::find ( rhs.second.begin ( ), rhs.second.end ( ), symbol ) != rhs.second.end ( ) )
				return true;

	}

	if ( grammar->accessElement < grammar::InitialSymbol > ( ).get ( ) == symbol )
		return true;

	return false;
}

template < >
bool grammar::GNF::Component < grammar::GNF, alphabet::Symbol, grammar::NonterminalAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void grammar::GNF::Component < grammar::GNF, alphabet::Symbol, grammar::NonterminalAlphabet >::valid ( const alphabet::Symbol & symbol ) const {
	const grammar::GNF * grammar = static_cast < const grammar::GNF * > ( this );

	if ( grammar->accessComponent < grammar::TerminalAlphabet > ( ).get ( ).count ( symbol ) )
		throw grammar::GrammarException ( "Symbol " + ( std::string ) symbol + "cannot be in nonterminal alphabet since it is already in terminal alphabet" );
}

template < >
bool grammar::GNF::Element < grammar::GNF, alphabet::Symbol, grammar::InitialSymbol >::available ( const alphabet::Symbol & symbol ) const {
	const grammar::GNF * grammar = static_cast < const grammar::GNF * > ( this );

	return grammar->accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol );
}

template < >
void grammar::GNF::Element < grammar::GNF, alphabet::Symbol, grammar::InitialSymbol >::valid ( const alphabet::Symbol & ) const {
}

} /* namespace std */

namespace alib {

auto GNFParserRegister = xmlApi < grammar::Grammar >::ParserRegister < grammar::GNF > ( );
auto GNFParserRegister2 = xmlApi < alib::Object >::ParserRegister < grammar::GNF > ( );

} /* namespace alib */
