/*
 * LG.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "LG.h"
#include "../GrammarException.h"
#include <algorithm>
#include <sstream>

#include "../../alphabet/Symbol.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../Grammar.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace grammar {

LG::LG ( alphabet::Symbol initialSymbol ) : LG ( std::set < alphabet::Symbol > { initialSymbol }, std::set < alphabet::Symbol > ( ), initialSymbol ) {
}

LG::LG ( std::set < alphabet::Symbol > nonterminalAlphabet, std::set < alphabet::Symbol > terminalAlphabet, alphabet::Symbol initialSymbol ) : std::Components < LG, alphabet::Symbol, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > ( std::make_tuple ( std::move ( terminalAlphabet ), std::move ( nonterminalAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ) ) {
}

GrammarBase * LG::clone ( ) const {
	return new LG ( * this );
}

GrammarBase * LG::plunder ( ) && {
	return new LG ( std::move ( * this ) );
}

bool LG::addRule ( alphabet::Symbol leftHandSide, std::variant < std::vector < alphabet::Symbol >, std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > rightHandSide ) {
	if ( !getNonterminalAlphabet ( ).count ( leftHandSide ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	if ( rightHandSide.is < std::vector < alphabet::Symbol > > ( ) ) {
		const std::vector < alphabet::Symbol > & rhs = rightHandSide.get < std::vector < alphabet::Symbol > > ( );

		for ( const auto & symbol : rhs )
			if ( !getTerminalAlphabet ( ).count ( symbol ) )
				throw GrammarException ( "Symbol " + ( std::string ) symbol + " is not a terminal symbol" );

		return rules[std::move ( leftHandSide )].insert ( std::move ( rightHandSide ) ).second;
	} else {
		const std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > & rhs = rightHandSide.get < std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > ( );

		for ( const auto & symbol : std::get < 0 > ( rhs ) )
			if ( !getTerminalAlphabet ( ).count ( symbol ) )
				throw GrammarException ( "Symbol " + ( std::string ) symbol + " is not a terminal symbol" );

		if ( !getNonterminalAlphabet ( ).count ( std::get < 1 > ( rhs ) ) )
			throw GrammarException ( "Symbol " + ( std::string ) std::get < 1 > ( rhs ) + " is not a nonterminal symbol" );

		for ( const auto & symbol : std::get < 2 > ( rhs ) )
			if ( !getTerminalAlphabet ( ).count ( symbol ) )
				throw GrammarException ( "Symbol " + ( std::string ) symbol + " is not a terminal symbol" );

		return rules[std::move ( leftHandSide )].insert ( std::move ( rightHandSide ) ).second;
	}
}

bool LG::addRule ( alphabet::Symbol leftHandSide, std::vector < alphabet::Symbol > rightHandSide ) {
	std::variant < std::vector < alphabet::Symbol >, std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > rhs ( std::move ( rightHandSide ) );

	return addRule ( std::move ( leftHandSide ), std::move ( rhs ) );
}

bool LG::addRule ( alphabet::Symbol leftHandSide, std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > rightHandSide ) {
	std::variant < std::vector < alphabet::Symbol >, std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > rhs ( std::move ( rightHandSide ) );

	return addRule ( std::move ( leftHandSide ), std::move ( rhs ) );
}

const std::map < alphabet::Symbol, std::set < std::variant < std::vector < alphabet::Symbol >, std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > > > & LG::getRules ( ) const {
	return rules;
}

bool LG::removeRule ( const alphabet::Symbol & leftHandSide, const std::variant < std::vector < alphabet::Symbol >, std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > & rightHandSide ) {
	return rules[leftHandSide].erase ( rightHandSide );
}

bool LG::removeRule ( const alphabet::Symbol & leftHandSide, const std::vector < alphabet::Symbol > & rightHandSide ) {
	std::variant < std::vector < alphabet::Symbol >, std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > rhs ( rightHandSide );

	return removeRule ( leftHandSide, rhs );
}

bool LG::removeRule ( const alphabet::Symbol & leftHandSide, const std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > & rightHandSide ) {
	std::variant < std::vector < alphabet::Symbol >, std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > rhs ( rightHandSide );

	return removeRule ( leftHandSide, rhs );
}

bool LG::addRawRule ( alphabet::Symbol leftHandSide, std::vector < alphabet::Symbol > rightHandSide ) {
	std::vector < alphabet::Symbol >::iterator nonterminalPosition = rightHandSide.begin ( );

	for ( ; nonterminalPosition != rightHandSide.end ( ); ++nonterminalPosition )
		if ( getNonterminalAlphabet ( ).count ( * nonterminalPosition ) ) break;

	if ( nonterminalPosition == rightHandSide.end ( ) )
		return addRule ( leftHandSide, rightHandSide );
	else
		return addRule ( leftHandSide, std::make_tuple ( std::vector < alphabet::Symbol > ( std::make_move_iterator ( rightHandSide.begin ( ) ), std::make_move_iterator ( nonterminalPosition ) ), std::move ( * nonterminalPosition ), std::vector < alphabet::Symbol > ( std::make_move_iterator ( nonterminalPosition ) + 1, std::make_move_iterator ( rightHandSide.end ( ) ) ) ) );
}

std::map < alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > LG::getRawRules ( ) const {
	std::map < alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > res;

	for ( const auto & rule : getRules ( ) )
		for ( const auto & rhs : rule.second ) {
			if ( rhs.is < std::vector < alphabet::Symbol > > ( ) ) {
				res[rule.first].insert ( rhs.get < std::vector < alphabet::Symbol > > ( ) );
			} else {
				const auto & rhsTuple = rhs.get < std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > ( );
				std::vector < alphabet::Symbol > tmp { std::get < 0 > ( rhsTuple ).begin ( ), std::get < 0 > ( rhsTuple ).end ( ) };
				tmp.push_back ( std::get < 1 > ( rhsTuple ) );
				tmp.insert ( tmp.end ( ), std::get < 2 > ( rhsTuple ).begin ( ), std::get < 2 > ( rhsTuple ).end ( ) );
				res[rule.first].insert ( std::move ( tmp ) );
			}
		}

	return res;
}

bool LG::removeRawRule ( const alphabet::Symbol & leftHandSide, const std::vector < alphabet::Symbol > & rightHandSide ) {
	std::vector < alphabet::Symbol >::const_iterator nonterminalPosition = rightHandSide.begin ( );

	for ( ; nonterminalPosition != rightHandSide.end ( ); ++nonterminalPosition )
		if ( getNonterminalAlphabet ( ).count ( * nonterminalPosition ) ) break;

	if ( nonterminalPosition == rightHandSide.end ( ) )
		return removeRule ( leftHandSide, rightHandSide );
	else
		return removeRule ( leftHandSide, std::make_tuple ( std::vector < alphabet::Symbol > ( rightHandSide.begin ( ), nonterminalPosition ), * nonterminalPosition, std::vector < alphabet::Symbol > ( nonterminalPosition + 1, rightHandSide.end ( ) ) ) );
}

int LG::compare ( const LG & other ) const {
	auto first = std::tie ( getTerminalAlphabet ( ), getNonterminalAlphabet ( ), getInitialSymbol ( ), rules );
	auto second = std::tie ( other.getTerminalAlphabet ( ), other.getNonterminalAlphabet ( ), other.getInitialSymbol ( ), other.rules );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void LG::operator >>( std::ostream & out ) const {
	out << "(LG "
	    << "nonterminalAlphabet = " << getNonterminalAlphabet ( )
	    << "terminalAlphabet = " << getTerminalAlphabet ( )
	    << "initialSymbol = " << getInitialSymbol ( )
	    << "rules = " << rules << ")";
}

LG::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

const std::string LG::XML_TAG_NAME = "LG";

LG LG::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, LG::XML_TAG_NAME );

	std::set < alphabet::Symbol > nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet ( input );
	std::set < alphabet::Symbol > terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet ( input );
	alphabet::Symbol initialSymbol = GrammarFromXMLParser::parseInitialSymbol ( input );

	LG grammar ( std::move ( initialSymbol ) );

	grammar.setNonterminalAlphabet ( std::move ( nonterminalAlphabet ) );
	grammar.setTerminalAlphabet ( std::move ( terminalAlphabet ) );

	GrammarFromXMLParser::parseRules ( input, grammar );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, LG::XML_TAG_NAME );
	return grammar;
}

void LG::parseRule ( std::deque < sax::Token >::iterator & input, LG & grammar ) {
	alphabet::Symbol lhs = GrammarFromXMLParser::parseRuleSingleSymbolLHS ( input );
	std::vector < alphabet::Symbol > rhs = GrammarFromXMLParser::parseRuleRHS ( input );

	grammar.addRawRule ( std::move ( lhs ), std::move ( rhs ) );
}

void LG::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( LG::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT );

	GrammarToXMLComposer::composeNonterminalAlphabet ( out, this->getNonterminalAlphabet ( ) );
	GrammarToXMLComposer::composeTerminalAlphabet ( out, this->getTerminalAlphabet ( ) );
	GrammarToXMLComposer::composeInitialSymbol ( out, this->getInitialSymbol ( ) );
	composeRules ( out );

	out.emplace_back ( LG::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT );
}

void LG::composeRules ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "rules", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & rule : this->getRules ( ) )

		for ( const auto & rhs : rule.second ) {
			out.emplace_back ( "rule", sax::Token::TokenType::START_ELEMENT );

			GrammarToXMLComposer::composeRuleSingleSymbolLHS ( out, rule.first );
			GrammarToXMLComposer::composeRuleLGRHS ( out, rhs );

			out.emplace_back ( "rule", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "rules", sax::Token::TokenType::END_ELEMENT );
}

} /* namespace grammar */

namespace std {

template < >
bool grammar::LG::Component < grammar::LG, alphabet::Symbol, grammar::TerminalAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	const grammar::LG * grammar = static_cast < const grammar::LG * > ( this );

	for ( const std::pair < const alphabet::Symbol, std::set < std::variant < std::vector < alphabet::Symbol >, std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > > > & rule : grammar->getRules ( ) ) {
		for ( const std::variant < std::vector < alphabet::Symbol >, std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > & rhsTmp : rule.second )
			if ( rhsTmp.is < std::vector < alphabet::Symbol > > ( ) ) {
				const std::vector < alphabet::Symbol > & rhs = rhsTmp.get < std::vector < alphabet::Symbol > > ( );

				if ( std::find ( rhs.begin ( ), rhs.end ( ), symbol ) != rhs.end ( ) )
					return true;
			} else {
				const std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > & rhs = rhsTmp.get < std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > ( );

				const std::vector < alphabet::Symbol > & lPart = std::get < 0 > ( rhs );

				if ( std::find ( lPart.begin ( ), lPart.end ( ), symbol ) != lPart.end ( ) )
					return true;

				const std::vector < alphabet::Symbol > & rPart = std::get < 2 > ( rhs );

				if ( std::find ( rPart.begin ( ), rPart.end ( ), symbol ) != rPart.end ( ) )
					return true;
			}

	}

	return false;
}

template < >
bool grammar::LG::Component < grammar::LG, alphabet::Symbol, grammar::TerminalAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void grammar::LG::Component < grammar::LG, alphabet::Symbol, grammar::TerminalAlphabet >::valid ( const alphabet::Symbol & symbol ) const {
	const grammar::LG * grammar = static_cast < const grammar::LG * > ( this );

	if ( grammar->accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol ) )
		throw grammar::GrammarException ( "Symbol " + ( std::string ) symbol + "cannot be in terminal alphabet since it is already nonterminal alphabet" );
}

template < >
bool grammar::LG::Component < grammar::LG, alphabet::Symbol, grammar::NonterminalAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	const grammar::LG * grammar = static_cast < const grammar::LG * > ( this );

	for ( const std::pair < const alphabet::Symbol, std::set < std::variant < std::vector < alphabet::Symbol >, std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > > > & rule : grammar->getRules ( ) ) {
		if ( rule.first == symbol )
			return true;

		for ( const std::variant < std::vector < alphabet::Symbol >, std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > & rhsTmp : rule.second )
			if ( rhsTmp.is < std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > ( ) ) {
				const std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > & rhs = rhsTmp.get < std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > ( );

				if ( std::get < 1 > ( rhs ) == symbol )
					return true;
			}

	}

	if ( grammar->accessElement < grammar::InitialSymbol > ( ).get ( ) == symbol )
		return true;

	return false;
}

template < >
bool grammar::LG::Component < grammar::LG, alphabet::Symbol, grammar::NonterminalAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void grammar::LG::Component < grammar::LG, alphabet::Symbol, grammar::NonterminalAlphabet >::valid ( const alphabet::Symbol & symbol ) const {
	const grammar::LG * grammar = static_cast < const grammar::LG * > ( this );

	if ( grammar->accessComponent < grammar::TerminalAlphabet > ( ).get ( ).count ( symbol ) )
		throw grammar::GrammarException ( "Symbol " + ( std::string ) symbol + "cannot be in nonterminal alphabet since it is already in terminal alphabet" );
}

template < >
bool grammar::LG::Element < grammar::LG, alphabet::Symbol, grammar::InitialSymbol >::available ( const alphabet::Symbol & symbol ) const {
	const grammar::LG * grammar = static_cast < const grammar::LG * > ( this );

	return grammar->accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol );
}

template < >
void grammar::LG::Element < grammar::LG, alphabet::Symbol, grammar::InitialSymbol >::valid ( const alphabet::Symbol & ) const {
}

} /* namespace std */

namespace alib {

auto LGParserRegister = xmlApi < grammar::Grammar >::ParserRegister < grammar::LG > ( );
auto LGParserRegister2 = xmlApi < alib::Object >::ParserRegister < grammar::LG > ( );

} /* namespace alib */
