/*
 * LG.h
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#ifndef LG_H_
#define LG_H_

#include "../GrammarBase.h"
#include <map>
#include <tuple>
#include <vector>
#include <variant>
#include <core/components.hpp>
#include "../../alphabet/Symbol.h"

namespace grammar {

/**
 * Linear grammar in chomsky normal form. Type 3 in Chomsky hierarchy. Produces regular languages.
 */
class TerminalAlphabet;
class NonterminalAlphabet;
class InitialSymbol;

class LG : public GrammarBase, public std::Components < LG, alphabet::Symbol, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > {
	std::map < alphabet::Symbol, std::set < std::variant < std::vector < alphabet::Symbol >, std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > > > rules;

public:
	explicit LG ( alphabet::Symbol initialSymbol );

	explicit LG ( std::set < alphabet::Symbol > nonTerminalSymbols, std::set < alphabet::Symbol > terminalSymbols, alphabet::Symbol initialSymbol );

	virtual GrammarBase * clone ( ) const;

	virtual GrammarBase * plunder ( ) &&;

	bool addRule ( alphabet::Symbol leftHandSide, std::variant < std::vector < alphabet::Symbol >, std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > rightHandSide );
	bool addRule ( alphabet::Symbol leftHandSide, std::vector < alphabet::Symbol > rightHandSide );
	bool addRule ( alphabet::Symbol leftHandSide, std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > rightHandSide );

	const std::map < alphabet::Symbol, std::set < std::variant < std::vector < alphabet::Symbol >, std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > > > & getRules ( ) const;

	bool removeRule ( const alphabet::Symbol & leftHandSide, const std::variant < std::vector < alphabet::Symbol >, std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > > & rightHandSide );
	bool removeRule ( const alphabet::Symbol & leftHandSide, const std::vector < alphabet::Symbol > & rightHandSide );
	bool removeRule ( const alphabet::Symbol & leftHandSide, const std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > > & rightHandSide );

	bool addRawRule ( alphabet::Symbol leftHandSide, std::vector < alphabet::Symbol > rightHandSide );

	std::map < alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > getRawRules ( ) const;

	bool removeRawRule ( const alphabet::Symbol & leftHandSide, const std::vector < alphabet::Symbol > & rightHandSide );

	const alphabet::Symbol & getInitialSymbol ( ) const {
		return accessElement < InitialSymbol > ( ).get ( );
	}

	bool setInitialSymbol ( alphabet::Symbol symbol ) {
		return accessElement < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < alphabet::Symbol > & getNonterminalAlphabet ( ) const {
		return accessComponent < NonterminalAlphabet > ( ).get ( );
	}

	bool addNonterminalSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < NonterminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setNonterminalAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < NonterminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	const std::set < alphabet::Symbol > & getTerminalAlphabet ( ) const {
		return accessComponent < TerminalAlphabet > ( ).get ( );
	}

	bool addTerminalSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < TerminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setTerminalAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < TerminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const LG & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	const static std::string XML_TAG_NAME;

	static LG parse ( std::deque < sax::Token >::iterator & input );
	static void parseRule ( std::deque < sax::Token >::iterator & input, LG & grammar );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeRules ( std::deque < sax::Token > & out ) const;
};

} /* namespace grammar */

#endif /* LG_H_ */
