/*
 * LeftRG.h
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#ifndef LEFT_RG_H_
#define LEFT_RG_H_

#include "../GrammarBase.h"
#include <map>
#include <vector>
#include <variant>
#include <core/components.hpp>
#include "../../alphabet/Symbol.h"

namespace grammar {

/**
 * Left regular grammar in chomsky normal form. Type 2 in Chomsky hierarchy. Produces regular languages.
 * Definition is similar to all common definitions of regular grammars. Additionaly contains boolean signaling whether the grammar generates empty string or don't.
 * G = (N, T, P, S, E)
 * N = nonempty finite set of nonterminal symbols
 * T = finite set of terminal symbols - having this empty won't let grammar do much though
 * P = set of production rules of the form A -> aB or A -> a, where A, B \in N and a \in T
 * S = initial nonterminal symbol
 * E = boolean signaling wheter grammar generates empty string or don't
 *
 * This definition has simplier handling of empty string generation and it is compatible with common definitions where the transformation from this definition to the common definition and backwards is trivial
 */
class TerminalAlphabet;
class NonterminalAlphabet;
class InitialSymbol;

class LeftRG : public GrammarBase, public std::Components < LeftRG, alphabet::Symbol, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > {
	/**
	 * Transition function as mapping from nonterminal symbol on the left hand side to set of either terminal symbols or doublets of terminal symbol and nonterminal symbol
	 */
	std::map < alphabet::Symbol, std::set < std::variant < alphabet::Symbol, std::pair < alphabet::Symbol, alphabet::Symbol > > > > rules;

	/**
	 * Boolean signaling whether grammar generates empty string or don't.
	 */
	bool generatesEpsilon;

public:
	/**
	 * Creates a new instance of Left regular grammar with concrete initial symbol
	 */
	explicit LeftRG ( alphabet::Symbol initialSymbol );

	/**
	 * Creates a new instance of Left regular grammar with concrete nonterminal, terminal alphabet and initial symbol
	 */
	explicit LeftRG ( std::set < alphabet::Symbol > nonTerminalSymbols, std::set < alphabet::Symbol > terminalSymbols, alphabet::Symbol initialSymbol );

	/**
	 * @copydoc alib::base::clone()
	 */
	virtual GrammarBase * clone ( ) const;

	/**
	 * @copydoc alib::base::plunder()
	 */
	virtual GrammarBase * plunder ( ) &&;

	/**
	 * Add a new rule of a grammar in form of A -> aB or A -> a, where A, B \in N and a \in T
	 */
	bool addRule ( alphabet::Symbol leftHandSide, std::variant < alphabet::Symbol, std::pair < alphabet::Symbol, alphabet::Symbol > > rightHandSide );

	/**
	 * Add a new rule of a grammar in form of A -> a, where A \in N and a \in T
	 */
	bool addRule ( alphabet::Symbol leftHandSide, alphabet::Symbol rightHandSide );

	/**
	 * Add a new rule of a grammar in form of A -> aB, where A, B \in N and a \in T
	 */
	bool addRule ( alphabet::Symbol leftHandSide, std::pair < alphabet::Symbol, alphabet::Symbol > rightHandSide );

	/**
	 * Get rules of the grammar
	 */
	const std::map < alphabet::Symbol, std::set < std::variant < alphabet::Symbol, std::pair < alphabet::Symbol, alphabet::Symbol > > > > & getRules ( ) const;

	/**
	 * Add a new rule of a grammar in form of A -> aB or A -> a, where A, B \in N and a \in T
	 */
	bool removeRule ( const alphabet::Symbol & leftHandSide, const std::variant < alphabet::Symbol, std::pair < alphabet::Symbol, alphabet::Symbol > > & rightHandSide );

	/**
	 * Add a new rule of a grammar in form of A -> a, where A \in N and a \in T
	 */
	bool removeRule ( const alphabet::Symbol & leftHandSide, const alphabet::Symbol & rightHandSide );

	/**
	 * Add a new rule of a grammar in form of A -> aB, where A, B \in N and a \in T
	 */
	bool removeRule ( const alphabet::Symbol & leftHandSide, const std::pair < alphabet::Symbol, alphabet::Symbol > & rightHandSide );

	bool addRawRule ( alphabet::Symbol leftHandSide, std::vector < alphabet::Symbol > rightHandSide );

	std::map < alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > getRawRules ( ) const;

	bool removeRawRule ( const alphabet::Symbol & leftHandSide, const std::vector < alphabet::Symbol > & rightHandSide );

	const alphabet::Symbol & getInitialSymbol ( ) const {
		return accessElement < InitialSymbol > ( ).get ( );
	}

	bool setInitialSymbol ( alphabet::Symbol symbol ) {
		return accessElement < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < alphabet::Symbol > & getNonterminalAlphabet ( ) const {
		return accessComponent < NonterminalAlphabet > ( ).get ( );
	}

	bool addNonterminalSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < NonterminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setNonterminalAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < NonterminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	const std::set < alphabet::Symbol > & getTerminalAlphabet ( ) const {
		return accessComponent < TerminalAlphabet > ( ).get ( );
	}

	bool addTerminalSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < TerminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setTerminalAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < TerminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	/**
	 * Sets sign representing that grammar generates or doesn't generate empty word
	 */
	void setGeneratesEpsilon ( bool genEps );

	/**
	 * Gets sign representing that grammar generates or doesn't generate empty word
	 */
	bool getGeneratesEpsilon ( ) const;

	/**
	 * double dispatch operator helper
	 */
	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	/**
	 * compare
	 * @param other the other instance
	 */
	virtual int compare ( const LeftRG & other ) const;

	/**
	 * print this instance as raw representation to ostream
	 * @param os ostream where to print
	 */
	virtual void operator >>( std::ostream & os ) const;

	/**
	 * casts this instance to as compact as possible string representation
	 */
	virtual explicit operator std::string ( ) const;

	const static std::string XML_TAG_NAME;

	static LeftRG parse ( std::deque < sax::Token >::iterator & input );
	static void parseRule ( std::deque < sax::Token >::iterator & input, LeftRG & grammar );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeRules ( std::deque < sax::Token > & out ) const;
};

} /* namespace grammar */

#endif /* LEFT_RG_H_ */
