/*
 * ContextPreservingUnrestrictedGrammar.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "ContextPreservingUnrestrictedGrammar.h"
#include "../GrammarException.h"
#include <algorithm>
#include <sstream>

#include "../../alphabet/Symbol.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../Grammar.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace grammar {

ContextPreservingUnrestrictedGrammar::ContextPreservingUnrestrictedGrammar ( alphabet::Symbol initialSymbol ) : ContextPreservingUnrestrictedGrammar ( std::set < alphabet::Symbol > { initialSymbol }, std::set < alphabet::Symbol > ( ), initialSymbol ) {
}

ContextPreservingUnrestrictedGrammar::ContextPreservingUnrestrictedGrammar ( std::set < alphabet::Symbol > nonterminalAlphabet, std::set < alphabet::Symbol > terminalAlphabet, alphabet::Symbol initialSymbol ) : std::Components < ContextPreservingUnrestrictedGrammar, alphabet::Symbol, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > ( std::make_tuple ( std::move ( terminalAlphabet ), std::move ( nonterminalAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ) ) {
}

GrammarBase * ContextPreservingUnrestrictedGrammar::clone ( ) const {
	return new ContextPreservingUnrestrictedGrammar ( * this );
}

GrammarBase * ContextPreservingUnrestrictedGrammar::plunder ( ) && {
	return new ContextPreservingUnrestrictedGrammar ( std::move ( * this ) );
}

bool ContextPreservingUnrestrictedGrammar::addRule ( std::vector < alphabet::Symbol > lContext, alphabet::Symbol leftHandSide, std::vector < alphabet::Symbol > rContext, std::vector < alphabet::Symbol > rightHandSide ) {
	for ( const alphabet::Symbol & symbol : lContext )
		if ( !getTerminalAlphabet ( ).count ( symbol ) && !getNonterminalAlphabet ( ).count ( symbol ) )
			throw GrammarException ( "Symbol \"" + ( std::string ) symbol + "\" is not neither terminal nor nonterminal symbol" );

	if ( !getNonterminalAlphabet ( ).count ( leftHandSide ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	for ( const alphabet::Symbol & symbol : rContext )
		if ( !getTerminalAlphabet ( ).count ( symbol ) && !getNonterminalAlphabet ( ).count ( symbol ) )
			throw GrammarException ( "Symbol \"" + ( std::string ) symbol + "\" is not neither terminal nor nonterminal symbol" );

	for ( const alphabet::Symbol & symbol : rightHandSide )
		if ( !getTerminalAlphabet ( ).count ( symbol ) && !getNonterminalAlphabet ( ).count ( symbol ) )
			throw GrammarException ( "Symbol \"" + ( std::string ) symbol + "\" is not neither terminal nor nonterminal symbol" );

	return rules[make_tuple ( std::move ( lContext ), std::move ( leftHandSide ), std::move ( rContext ) )].insert ( std::move ( rightHandSide ) ).second;
}

const std::map < std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > >, std::set < std::vector < alphabet::Symbol > > > & ContextPreservingUnrestrictedGrammar::getRules ( ) const {
	return rules;
}

bool ContextPreservingUnrestrictedGrammar::removeRule ( const std::vector < alphabet::Symbol > & lContext, const alphabet::Symbol & leftHandSide, const std::vector < alphabet::Symbol > & rContext, const std::vector < alphabet::Symbol > & rightHandSide ) {
	return rules[make_tuple ( lContext, leftHandSide, rContext )].erase ( rightHandSide );
}

int ContextPreservingUnrestrictedGrammar::compare ( const ContextPreservingUnrestrictedGrammar & other ) const {
	auto first = std::tie ( getTerminalAlphabet ( ), getNonterminalAlphabet ( ), getInitialSymbol ( ), rules );
	auto second = std::tie ( other.getTerminalAlphabet ( ), other.getNonterminalAlphabet ( ), other.getInitialSymbol ( ), other.rules );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void ContextPreservingUnrestrictedGrammar::operator >>( std::ostream & out ) const {
	out << "(ContextPreservingUnrestrictedGrammar "
	    << "nonterminalAlphabet = " << getNonterminalAlphabet ( )
	    << "terminalAlphabet = " << getTerminalAlphabet ( )
	    << "initialSymbol = " << getInitialSymbol ( )
	    << "rules = " << rules << ")";
}

ContextPreservingUnrestrictedGrammar::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

const std::string ContextPreservingUnrestrictedGrammar::XML_TAG_NAME = "ContextPreservingUnrestrictedGrammar";

ContextPreservingUnrestrictedGrammar ContextPreservingUnrestrictedGrammar::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, ContextPreservingUnrestrictedGrammar::XML_TAG_NAME );

	std::set < alphabet::Symbol > nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet ( input );
	std::set < alphabet::Symbol > terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet ( input );
	alphabet::Symbol initialSymbol = GrammarFromXMLParser::parseInitialSymbol ( input );

	ContextPreservingUnrestrictedGrammar grammar ( std::move ( initialSymbol ) );

	grammar.setNonterminalAlphabet ( std::move ( nonterminalAlphabet ) );
	grammar.setTerminalAlphabet ( std::move ( terminalAlphabet ) );

	GrammarFromXMLParser::parseRules ( input, grammar );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, ContextPreservingUnrestrictedGrammar::XML_TAG_NAME );
	return grammar;
}

void ContextPreservingUnrestrictedGrammar::parseRule ( std::deque < sax::Token >::iterator & input, ContextPreservingUnrestrictedGrammar & grammar ) {
	std::vector < alphabet::Symbol > lContext = GrammarFromXMLParser::parseRuleLContext ( input );
	alphabet::Symbol lhs = GrammarFromXMLParser::parseRuleSingleSymbolLHS ( input );
	std::vector < alphabet::Symbol > rContext = GrammarFromXMLParser::parseRuleRContext ( input );
	std::vector < alphabet::Symbol > rhs = GrammarFromXMLParser::parseRuleRHS ( input );

	grammar.addRule ( std::move ( lContext ), std::move ( lhs ), std::move ( rContext ), std::move ( rhs ) );
}

void ContextPreservingUnrestrictedGrammar::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( ContextPreservingUnrestrictedGrammar::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT );

	GrammarToXMLComposer::composeNonterminalAlphabet ( out, this->getNonterminalAlphabet ( ) );
	GrammarToXMLComposer::composeTerminalAlphabet ( out, this->getTerminalAlphabet ( ) );
	GrammarToXMLComposer::composeInitialSymbol ( out, this->getInitialSymbol ( ) );
	composeRules ( out );

	out.emplace_back ( ContextPreservingUnrestrictedGrammar::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT );
}

void ContextPreservingUnrestrictedGrammar::composeRules ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "rules", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & rule : this->getRules ( ) )

		for ( const auto & rhs : rule.second ) {
			out.emplace_back ( "rule", sax::Token::TokenType::START_ELEMENT );

			GrammarToXMLComposer::composeRuleLContext ( out, std::get < 0 > ( rule.first ) );
			GrammarToXMLComposer::composeRuleSingleSymbolLHS ( out, std::get < 1 > ( rule.first ) );
			GrammarToXMLComposer::composeRuleRContext ( out, std::get < 2 > ( rule.first ) );
			GrammarToXMLComposer::composeRuleRHS ( out, rhs );

			out.emplace_back ( "rule", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "rules", sax::Token::TokenType::END_ELEMENT );
}

} /* namespace grammar */

namespace std {

template < >
bool grammar::ContextPreservingUnrestrictedGrammar::Component < grammar::ContextPreservingUnrestrictedGrammar, alphabet::Symbol, grammar::TerminalAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	const grammar::ContextPreservingUnrestrictedGrammar * grammar = static_cast < const grammar::ContextPreservingUnrestrictedGrammar * > ( this );

	for ( const std::pair < const std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > >, std::set < std::vector < alphabet::Symbol > > > & rule : grammar->getRules ( ) ) {
		for ( const alphabet::Symbol & lCont : std::get < 0 > ( rule.first ) )
			if ( lCont == symbol )
				return true;

		for ( const alphabet::Symbol & rCont : std::get < 2 > ( rule.first ) )
			if ( rCont == symbol )
				return true;

		for ( const std::vector < alphabet::Symbol > & rhs : rule.second )
			if ( std::find ( rhs.begin ( ), rhs.end ( ), symbol ) != rhs.end ( ) )
				return true;

	}

	return false;
}

template < >
bool grammar::ContextPreservingUnrestrictedGrammar::Component < grammar::ContextPreservingUnrestrictedGrammar, alphabet::Symbol, grammar::TerminalAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void grammar::ContextPreservingUnrestrictedGrammar::Component < grammar::ContextPreservingUnrestrictedGrammar, alphabet::Symbol, grammar::TerminalAlphabet >::valid ( const alphabet::Symbol & symbol ) const {
	const grammar::ContextPreservingUnrestrictedGrammar * grammar = static_cast < const grammar::ContextPreservingUnrestrictedGrammar * > ( this );

	if ( grammar->accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol ) )
		throw grammar::GrammarException ( "Symbol " + ( std::string ) symbol + "cannot be in terminal alphabet since it is already nonterminal alphabet" );
}

template < >
bool grammar::ContextPreservingUnrestrictedGrammar::Component < grammar::ContextPreservingUnrestrictedGrammar, alphabet::Symbol, grammar::NonterminalAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	const grammar::ContextPreservingUnrestrictedGrammar * grammar = static_cast < const grammar::ContextPreservingUnrestrictedGrammar * > ( this );

	for ( const std::pair < const std::tuple < std::vector < alphabet::Symbol >, alphabet::Symbol, std::vector < alphabet::Symbol > >, std::set < std::vector < alphabet::Symbol > > > & rule : grammar->getRules ( ) ) {
		for ( const alphabet::Symbol & lCont : std::get < 0 > ( rule.first ) )
			if ( lCont == symbol )
				return true;

		if ( std::get < 1 > ( rule.first ) == symbol )
			return true;

		for ( const alphabet::Symbol & rCont : std::get < 2 > ( rule.first ) )
			if ( rCont == symbol )
				return true;

		for ( const std::vector < alphabet::Symbol > & rhs : rule.second )
			if ( std::find ( rhs.begin ( ), rhs.end ( ), symbol ) != rhs.end ( ) )
				return true;

	}

	if ( grammar->accessElement < grammar::InitialSymbol > ( ).get ( ) == symbol )
		return true;

	return false;
}

template < >
bool grammar::ContextPreservingUnrestrictedGrammar::Component < grammar::ContextPreservingUnrestrictedGrammar, alphabet::Symbol, grammar::NonterminalAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void grammar::ContextPreservingUnrestrictedGrammar::Component < grammar::ContextPreservingUnrestrictedGrammar, alphabet::Symbol, grammar::NonterminalAlphabet >::valid ( const alphabet::Symbol & symbol ) const {
	const grammar::ContextPreservingUnrestrictedGrammar * grammar = static_cast < const grammar::ContextPreservingUnrestrictedGrammar * > ( this );

	if ( grammar->accessComponent < grammar::TerminalAlphabet > ( ).get ( ).count ( symbol ) )
		throw grammar::GrammarException ( "Symbol " + ( std::string ) symbol + "cannot be in nonterminal alphabet since it is already in terminal alphabet" );
}

template < >
bool grammar::ContextPreservingUnrestrictedGrammar::Element < grammar::ContextPreservingUnrestrictedGrammar, alphabet::Symbol, grammar::InitialSymbol >::available ( const alphabet::Symbol & symbol ) const {
	const grammar::ContextPreservingUnrestrictedGrammar * grammar = static_cast < const grammar::ContextPreservingUnrestrictedGrammar * > ( this );

	return grammar->accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol );
}

template < >
void grammar::ContextPreservingUnrestrictedGrammar::Element < grammar::ContextPreservingUnrestrictedGrammar, alphabet::Symbol, grammar::InitialSymbol >::valid ( const alphabet::Symbol & ) const {
}

} /* namespace std */

namespace alib {

auto ContextPreservingUnrestrictedGrammarParserRegister = xmlApi < grammar::Grammar >::ParserRegister < grammar::ContextPreservingUnrestrictedGrammar > ( );
auto ContextPreservingUnrestrictedGrammarParserRegister2 = xmlApi < alib::Object >::ParserRegister < grammar::ContextPreservingUnrestrictedGrammar > ( );

} /* namespace alib */
