/*
 * UnrestrictedGrammar.h
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#ifndef UNRESTRICTED_GRAMMAR_H_
#define UNRESTRICTED_GRAMMAR_H_

#include "../GrammarBase.h"
#include <map>
#include <vector>
#include <core/components.hpp>
#include "../../alphabet/Symbol.h"

namespace grammar {

/**
 * Unrestricted grammar. Type 0 in Chomsky hierarchy. Produces recursively enumerable language.
 */
class TerminalAlphabet;
class NonterminalAlphabet;
class InitialSymbol;

class UnrestrictedGrammar : public GrammarBase, public std::Components < UnrestrictedGrammar, alphabet::Symbol, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > {
	std::map < std::vector < alphabet::Symbol >, std::set < std::vector < alphabet::Symbol > > > rules;

public:
	explicit UnrestrictedGrammar ( alphabet::Symbol initialSymbol );

	explicit UnrestrictedGrammar ( std::set < alphabet::Symbol > nonTerminalSymbols, std::set < alphabet::Symbol > terminalSymbols, alphabet::Symbol initialSymbol );

	virtual GrammarBase * clone ( ) const;

	virtual GrammarBase * plunder ( ) &&;

	bool addRule ( std::vector < alphabet::Symbol > leftHandSide, std::vector < alphabet::Symbol > rightHandSide );

	const std::map < std::vector < alphabet::Symbol >, std::set < std::vector < alphabet::Symbol > > > & getRules ( ) const;

	bool removeRule ( const std::vector < alphabet::Symbol > & leftHandSide, const std::vector < alphabet::Symbol > & rightHandSide );

	const alphabet::Symbol & getInitialSymbol ( ) const {
		return accessElement < InitialSymbol > ( ).get ( );
	}

	bool setInitialSymbol ( alphabet::Symbol symbol ) {
		return accessElement < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < alphabet::Symbol > & getNonterminalAlphabet ( ) const {
		return accessComponent < NonterminalAlphabet > ( ).get ( );
	}

	bool addNonterminalSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < NonterminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setNonterminalAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < NonterminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	const std::set < alphabet::Symbol > & getTerminalAlphabet ( ) const {
		return accessComponent < TerminalAlphabet > ( ).get ( );
	}

	bool addTerminalSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < TerminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setTerminalAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < TerminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const UnrestrictedGrammar & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	const static std::string XML_TAG_NAME;

	static UnrestrictedGrammar parse ( std::deque < sax::Token >::iterator & input );
	static void parseRule ( std::deque < sax::Token >::iterator & input, UnrestrictedGrammar & grammar );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeRules ( std::deque < sax::Token > & out ) const;
};

} /* namespace grammar */

#endif /* UNRESTRICTED_GRAMMAR_H_ */
