/*
 * FormalRegExp.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include "FormalRegExp.h"
#include <exception/CommonException.h>
#include "FormalRegExpEmpty.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../unbounded/UnboundedRegExp.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/RegExpFromXMLParser.h"
#include "../common/RegExpToXMLComposer.h"
#include "../RegExp.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace regexp {

FormalRegExp::FormalRegExp ( std::set < alphabet::Symbol > alphabet, FormalRegExpElement && regExp ) : std::Components < FormalRegExp, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ) ), std::tuple < > ( ) ), regExp ( NULL ) {
	setRegExp ( std::move ( regExp ) );
}

FormalRegExp::FormalRegExp ( std::set < alphabet::Symbol > alphabet, const FormalRegExpElement & regExp ) : FormalRegExp ( std::move ( alphabet ), std::move_copy ( regExp ) ) {
}

FormalRegExp::FormalRegExp ( ) : FormalRegExp ( std::set < alphabet::Symbol > ( ), FormalRegExpEmpty ( ) ) {
}

FormalRegExp::FormalRegExp ( FormalRegExpElement && regExp ) : FormalRegExp ( regExp.computeMinimalAlphabet ( ), regExp ) {
}

FormalRegExp::FormalRegExp ( const FormalRegExpElement & regExp ) : FormalRegExp ( regExp.computeMinimalAlphabet ( ), std::move_copy ( regExp ) ) {
}

FormalRegExp::FormalRegExp ( const UnboundedRegExp & other ) : FormalRegExp ( other.getAlphabet ( ), std::move ( * other.getRegExp ( ).asFormal ( ) ) ) {
}

FormalRegExp::FormalRegExp ( const FormalRegExp & other ) : std::Components < FormalRegExp, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( other.getAlphabet ( ) ), std::tuple < > ( ) ), regExp ( other.regExp ) {
	this->regExp->attachRegExp ( this );
}

FormalRegExp::FormalRegExp ( FormalRegExp && other ) noexcept : std::Components < FormalRegExp, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( other.accessComponent < GeneralAlphabet > ( ).get ( ) ) ), std::tuple < > ( ) ), regExp ( std::move ( other.regExp ) ) {
	this->regExp->attachRegExp ( this );
}

RegExpBase * FormalRegExp::clone ( ) const {
	return new FormalRegExp ( * this );
}

RegExpBase * FormalRegExp::plunder ( ) && {
	return new FormalRegExp ( std::move ( * this ) );
}

FormalRegExp & FormalRegExp::operator =( const FormalRegExp & other ) {
	if ( this == & other )
		return * this;

	* this = FormalRegExp ( other );

	return * this;
}

FormalRegExp & FormalRegExp::operator =( FormalRegExp && other ) noexcept {
	std::swap ( this->regExp, other.regExp );
	std::swap ( accessComponent < GeneralAlphabet > ( ).get ( ), other.accessComponent < GeneralAlphabet > ( ).get ( ) );

	return * this;
}

FormalRegExp::~FormalRegExp ( ) noexcept {
}

const FormalRegExpElement & FormalRegExp::getRegExp ( ) const {
	return * regExp;
}

FormalRegExpElement & FormalRegExp::getRegExp ( ) {
	return * regExp;
}

void FormalRegExp::setRegExp ( const FormalRegExpElement & regExp ) {
	setRegExp ( std::move_copy ( regExp ) );
}

void FormalRegExp::setRegExp ( FormalRegExpElement && param ) {
	this->regExp = std::smart_ptr < FormalRegExpElement > ( std::move ( param ).plunder ( ) );

	if ( !regExp->attachRegExp ( this ) )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

void FormalRegExp::operator >>( std::ostream & out ) const {
	out << "(FormalRegExp " << * ( this->regExp ) << ")";
}

int FormalRegExp::compare ( const FormalRegExp & other ) const {
	auto first = std::tie ( regExp, getAlphabet ( ) );
	auto second = std::tie ( other.regExp, other.getAlphabet ( ) );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

FormalRegExp::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

const std::string FormalRegExp::XML_TAG_NAME = "FormalRegExp";

FormalRegExp FormalRegExp::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, FormalRegExp::XML_TAG_NAME );

	std::set < alphabet::Symbol > alphabet = RegExpFromXMLParser::parseAlphabet ( input );

	std::rvalue_ref < FormalRegExpElement > element ( alib::xmlApi < regexp::FormalRegExpElement >::parse ( input ) );
	FormalRegExp regexp ( std::move ( alphabet ), std::move ( element ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, FormalRegExp::XML_TAG_NAME );

	return regexp;
}

void FormalRegExp::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( FormalRegExp::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT );
	RegExpToXMLComposer::composeAlphabet ( out, getAlphabet ( ) );
	alib::xmlApi < FormalRegExpElement >::compose ( out, this->getRegExp ( ) );
	out.emplace_back ( FormalRegExp::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT );
}

} /* namespace regexp */

namespace std {

template < >
bool regexp::FormalRegExp::Component < regexp::FormalRegExp, alphabet::Symbol, regexp::GeneralAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	return static_cast < const regexp::FormalRegExp * > ( this )->getRegExp ( ).testSymbol ( symbol );
}

template < >
bool regexp::FormalRegExp::Component < regexp::FormalRegExp, alphabet::Symbol, regexp::GeneralAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void regexp::FormalRegExp::Component < regexp::FormalRegExp, alphabet::Symbol, regexp::GeneralAlphabet >::valid ( const alphabet::Symbol & ) const {
}

} /* namespace std */

namespace alib {

auto formalRegExpParserRegister = xmlApi < regexp::RegExp >::ParserRegister < regexp::FormalRegExp > ( );
auto formalRegExpParserRegister2 = xmlApi < alib::Object >::ParserRegister < regexp::FormalRegExp > ( );

auto FormalRegExpFromUnboundedRegExp = castApi::CastRegister < regexp::FormalRegExp, regexp::UnboundedRegExp > ( );
auto FormalRegExpCastBinder = castApi::CastPoolStringBinder < regexp::FormalRegExp > ( regexp::FormalRegExp::XML_TAG_NAME );

} /* namespace alib */
