/*
 * FormalRegExp.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef FORMAL_REG_EXP_H_
#define FORMAL_REG_EXP_H_

#include <string>
#include <set>
#include <core/components.hpp>

#include "../RegExpBase.h"
#include "../../alphabet/Symbol.h"

namespace regexp {

class FormalRegExpElement;
class UnboundedRegExp;
class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
class FormalRegExp : public RegExpBase, public std::Components < FormalRegExp, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < > > {
protected:
	std::smart_ptr < FormalRegExpElement > regExp;

public:
	/**
	 * @copydoc FormalRegExpElement::clone() const
	 */
	virtual RegExpBase * clone ( ) const;

	/**
	 * @copydoc FormalRegExpElement::plunder() const
	 */
	virtual RegExpBase * plunder ( ) &&;

	explicit FormalRegExp ( );

	explicit FormalRegExp ( std::set < alphabet::Symbol > alphabet, FormalRegExpElement && regExp );
	explicit FormalRegExp ( std::set < alphabet::Symbol > alphabet, const FormalRegExpElement & regExp );
	explicit FormalRegExp ( FormalRegExpElement && regExp );
	explicit FormalRegExp ( const FormalRegExpElement & regExp );
	explicit FormalRegExp ( const UnboundedRegExp & other );

	/**
	 * Copy constructor.
	 * @param other RegExp to copy
	 */
	FormalRegExp ( const FormalRegExp & other );
	FormalRegExp ( FormalRegExp && other ) noexcept;
	FormalRegExp & operator =( const FormalRegExp & other );
	FormalRegExp & operator =( FormalRegExp && other ) noexcept;
	~FormalRegExp ( ) noexcept;

	/**
	 * @return Root node of the regular expression tree
	 */
	const FormalRegExpElement & getRegExp ( ) const;

	/**
	 * @return Root node of the regular expression tree
	 */
	FormalRegExpElement & getRegExp ( );

	/**
	 * Sets the root node of the regular expression tree
	 * @param regExp root node to set
	 */
	void setRegExp ( FormalRegExpElement && regExp );
	void setRegExp ( const FormalRegExpElement & regExp );

	virtual const std::set < alphabet::Symbol > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	/**
	 * Prints XML representation of the RegExp to the output stream.
	 * @param out output stream to which print the RegExp
	 * @param regexp RegExp to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const FormalRegExp & other ) const;

	virtual explicit operator std::string ( ) const;

	const static std::string XML_TAG_NAME;

	static FormalRegExp parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace regexp */

#endif /* FORMAL_REG_EXP_H_ */
