/*
 * LinearString.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "LinearString.h"
#include "Epsilon.h"
#include "../tree/ranked/PrefixRankedTree.h"
#include "../tree/ranked/PrefixRankedPattern.h"
#include "../tree/ranked/PrefixRankedBarTree.h"
#include "../tree/ranked/PrefixRankedBarPattern.h"
#include "../tree/unranked/PrefixBarTree.h"
#include "../tree/common/TreeAuxiliary.h"
#include <exception/CommonException.h>

#include <sstream>
#include <algorithm>

#include <sax/FromXMLParserHelper.h>
#include "common/StringFromXMLParser.h"
#include "common/StringToXMLComposer.h"
#include "common/StringAuxiliary.h"
#include "String.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace string {

LinearString::LinearString(std::set<alphabet::Symbol> alphabet, std::vector<alphabet::Symbol> data) : std::Components < LinearString, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ) ), std::tuple < > ( ) ) {
	setContent(std::move(data));
}

LinearString::LinearString() : LinearString ( std::set < alphabet::Symbol > ( ), std::vector < alphabet::Symbol > ( ) ) {
}

LinearString::LinearString(std::vector<alphabet::Symbol> data) : LinearString ( std::set < alphabet::Symbol> ( data.begin(), data.end() ), data ) {
}

LinearString::LinearString(const std::string& string) : LinearString ( StringAuxiliary::toInternal ( string ) ) {
}

LinearString::LinearString(const char* string) : LinearString((std::string) string) {
}

LinearString::LinearString(const Epsilon& epsilon) : LinearString ( epsilon.getAlphabet( ), std::vector < alphabet::Symbol > ( ) ) {
}

LinearString::LinearString ( const tree::PrefixRankedTree & tree ) : LinearString ( StringAuxiliary::wrapSymbols ( tree.getAlphabet ( ) ), StringAuxiliary::wrapSymbols ( tree.getContent ( ) ) ) {
}

LinearString::LinearString ( const tree::PrefixRankedPattern & tree ) : LinearString ( StringAuxiliary::wrapSymbols ( tree.getAlphabet ( ) ), StringAuxiliary::wrapSymbols ( tree.getContent ( ) ) ) {
}

LinearString::LinearString ( const tree::PrefixRankedBarTree & tree ) : LinearString ( StringAuxiliary::wrapSymbols ( tree.getAlphabet ( ) ), StringAuxiliary::wrapSymbols ( tree.getContent ( ) ) ) {
}

LinearString::LinearString ( const tree::PrefixRankedBarPattern & tree ) : LinearString ( StringAuxiliary::wrapSymbols ( tree.getAlphabet ( ) ), StringAuxiliary::wrapSymbols ( tree.getContent ( ) ) ) {
}

LinearString::LinearString ( const tree::PrefixBarTree & tree ) : LinearString ( tree.getAlphabet ( ), tree.getContent ( ) ) {
}

StringBase * LinearString::clone ( ) const {
	return new LinearString ( * this );
}

StringBase * LinearString::plunder ( ) && {
	return new LinearString ( std::move ( * this ) );
}

void LinearString::appendSymbol ( alphabet::Symbol symbol ) {
	if ( getAlphabet().count ( symbol ) == 0 )
		throw exception::CommonException ( "Input symbol \"" + ( std::string ) symbol + "\" not in the alphabet." );

	m_Data.push_back ( std::move ( symbol ) );
}

const std::vector < alphabet::Symbol > & LinearString::getContent ( ) const {
	return this->m_Data;
}

void LinearString::setContent ( std::vector < alphabet::Symbol > data ) {
	std::set < alphabet::Symbol > minimalAlphabet ( data.begin ( ), data.end ( ) );
	std::set < alphabet::Symbol > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::CommonException ( "Input symbols not in the alphabet." );

	this->m_Data = std::move ( data );
}

bool LinearString::isEmpty ( ) const {
	return this->m_Data.size ( ) == 0;
}

int LinearString::compare ( const LinearString & other ) const {
	auto first = std::tie ( m_Data, getAlphabet() );
	auto second = std::tie ( other.m_Data, other.getAlphabet() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void LinearString::operator >>( std::ostream & out ) const {
	out << "(LinearString ";

	for ( const alphabet::Symbol & symbol : this->m_Data )
		out << symbol;

	out << ")";
}

LinearString::operator std::string ( ) const {
	std::stringstream ss;
	ss << "\"";

	for ( const alphabet::Symbol & symbol : this->m_Data )
		ss << ( std::string ) symbol;

	ss << "\"";
	return std::move ( ss ).str ( );
}

const std::string LinearString::XML_TAG_NAME = "LinearString";

LinearString LinearString::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, LinearString::XML_TAG_NAME );
	std::set < alphabet::Symbol > alphabet = StringFromXMLParser::parseAlphabet ( input );
	std::vector < alphabet::Symbol > content = StringFromXMLParser::parseContent ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, LinearString::XML_TAG_NAME );

	return LinearString ( alphabet, content );
}

void LinearString::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( LinearString::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT );
	StringToXMLComposer::compose ( out, getAlphabet() );
	StringToXMLComposer::compose ( out, m_Data );
	out.emplace_back ( LinearString::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT );
}

} /* namespace string */

namespace std {

template < >
bool ::string::LinearString::Component < ::string::LinearString, alphabet::Symbol, ::string::GeneralAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	const std::vector<alphabet::Symbol>& content = static_cast < const ::string::LinearString * > ( this )->getContent ( );
	return std::find(content.begin(), content.end(), symbol) != content.end();
}

template < >
bool ::string::LinearString::Component < ::string::LinearString, alphabet::Symbol, ::string::GeneralAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void ::string::LinearString::Component < ::string::LinearString, alphabet::Symbol, ::string::GeneralAlphabet >::valid ( const alphabet::Symbol & ) const {
}

} /* namespace std */

namespace alib {

auto linearStringParserRegister = xmlApi < string::String >::ParserRegister < string::LinearString > ();
auto linearStringParserRegister2  = xmlApi < alib::Object >::ParserRegister < string::LinearString > ();

auto LinearStringFromEpsilon = castApi::CastRegister < string::LinearString, string::Epsilon > ( );
auto LinearStringFromPrefixRankedTree = castApi::CastRegister < string::LinearString, tree::PrefixRankedTree > ( );
auto LinearStringFromPrefixRankedPattern = castApi::CastRegister < string::LinearString, tree::PrefixRankedPattern > ( );
auto LinearStringFromPrefixRankedBarTree = castApi::CastRegister < string::LinearString, tree::PrefixRankedBarTree > ( );
auto LinearStringFromPrefixRankedBarPattern = castApi::CastRegister < string::LinearString, tree::PrefixRankedBarPattern > ( );
auto LinearStringFromPrefixBarTree = castApi::CastRegister < string::LinearString, tree::PrefixBarTree > ( );
auto LinearStringCastBinder = castApi::CastPoolStringBinder < string::LinearString > ( string::LinearString::XML_TAG_NAME );

} /* namespace alib */
