/*
 * RankedNonlinearPattern.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef RANKED_NONLINEAR_PATTERN_H_
#define RANKED_NONLINEAR_PATTERN_H_

#include <string>
#include <set>
#include <core/components.hpp>

#include "RankedNode.h"
#include "../RankedTreeBase.h"

namespace tree {

class RankedNode;
class UnrankedPattern;
class GeneralAlphabet;
class SubtreeWildcard;
class NonlinearAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a pattern of RegExpElement.
 */
class RankedNonlinearPattern : public RankedTreeBase, public std::Components < RankedNonlinearPattern, alphabet::RankedSymbol, std::tuple < GeneralAlphabet, NonlinearAlphabet >, std::tuple < SubtreeWildcard > > {
	std::smart_ptr < RankedNode > pattern;

public:
	/**
	 * @copydoc RankedNode::clone() const
	 */
	virtual RankedTreeBase * clone ( ) const;

	/**
	 * @copydoc RankedNode::plunder() const
	 */
	virtual RankedTreeBase * plunder ( ) &&;

	explicit RankedNonlinearPattern ( alphabet::RankedSymbol subtreeWildcard, std::set < alphabet::RankedSymbol > nonlinearVariables, std::set < alphabet::RankedSymbol > alphabet, RankedNode pattern );
	explicit RankedNonlinearPattern ( alphabet::RankedSymbol subtreeWildcard, std::set < alphabet::RankedSymbol > nonlinearVariables, RankedNode pattern );
	explicit RankedNonlinearPattern ( alphabet::RankedSymbol subtreeWildcard, RankedNode pattern );

	/**
	 * Copy constructor.
	 * @param other pattern to copy
	 */
	RankedNonlinearPattern ( const RankedNonlinearPattern & other );
	RankedNonlinearPattern ( RankedNonlinearPattern && other ) noexcept;
	RankedNonlinearPattern & operator =( const RankedNonlinearPattern & other );
	RankedNonlinearPattern & operator =( RankedNonlinearPattern && other ) noexcept;
	~RankedNonlinearPattern ( ) noexcept;

	/**
	 * @return Root node of the regular expression pattern
	 */
	const RankedNode & getRoot ( ) const;

	const std::set < alphabet::RankedSymbol > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < alphabet::RankedSymbol > & symbols ) {
		accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	const alphabet::RankedSymbol & getSubtreeWildcard ( ) const {
		return accessElement < SubtreeWildcard > ( ).get ( );
	}

	const std::set < alphabet::RankedSymbol > & getNonlinearVariables ( ) const {
		return accessComponent < NonlinearAlphabet > ( ).get ( );
	}

	/**
	 * @return Root node of the regular expression pattern
	 */
	RankedNode & getRoot ( );

	/**
	 * Sets the root node of the regular expression pattern
	 * @param pattern root node to set
	 */
	void setTree ( RankedNode pattern );

	/**
	 * Prints XML representation of the pattern to the output stream.
	 * @param out output stream to which print the pattern
	 * @param pattern pattern to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const RankedNonlinearPattern & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	const static std::string XML_TAG_NAME;

	static RankedNonlinearPattern parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace tree */

#endif /* RANKED_NONLINEAR_PATTERN_H_ */
