/*
 * RankedTree.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#ifndef RANKED_TREE_H_
#define RANKED_TREE_H_

#include <string>
#include <set>
#include <core/components.hpp>

#include "RankedNode.h"
#include "../RankedTreeBase.h"

namespace tree {

class RankedNode;
class UnrankedTree;
class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
class RankedTree : public RankedTreeBase, public std::Components < RankedTree, alphabet::RankedSymbol, std::tuple < GeneralAlphabet >, std::tuple < > > {
	std::smart_ptr < RankedNode > tree;

public:
	/**
	 * @copydoc RankedNode::clone() const
	 */
	virtual RankedTreeBase * clone ( ) const;

	/**
	 * @copydoc RankedNode::plunder() const
	 */
	virtual RankedTreeBase * plunder ( ) &&;

	explicit RankedTree ( std::set < alphabet::RankedSymbol > alphabet, RankedNode tree );
	explicit RankedTree ( RankedNode tree );
	explicit RankedTree ( const UnrankedTree & other );

	/**
	 * Copy constructor.
	 * @param other tree to copy
	 */
	RankedTree ( const RankedTree & other );
	RankedTree ( RankedTree && other ) noexcept;
	RankedTree & operator =( const RankedTree & other );
	RankedTree & operator =( RankedTree && other ) noexcept;
	~RankedTree ( ) noexcept;

	/**
	 * @return Root node of the regular expression tree
	 */
	const RankedNode & getRoot ( ) const;

	const std::set < alphabet::RankedSymbol > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < alphabet::RankedSymbol > & symbols ) {
		accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	/**
	 * @return Root node of the regular expression tree
	 */
	RankedNode & getRoot ( );

	/**
	 * Sets the root node of the regular expression tree
	 * @param tree root node to set
	 */
	void setTree ( RankedNode tree );

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const RankedTree & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	const static std::string XML_TAG_NAME;

	static RankedTree parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace tree */

#endif /* RANKED_TREE_H_ */
