/*
 * PrefixBarTree.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "PrefixBarTree.h"
#include "../TreeException.h"

#include <sstream>
#include <algorithm>
#include <deque>

#include "UnrankedTree.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../Tree.h"
#include "../UnrankedTreeWrapper.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

#include "../../alphabet/BarSymbol.h"

namespace tree {

PrefixBarTree::PrefixBarTree ( alphabet::Symbol bar, std::set < alphabet::Symbol > alphabet, std::vector < alphabet::Symbol > data ) : std::Components < PrefixBarTree, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < BarSymbol > > ( std::make_tuple ( std::move ( alphabet ) ), std::make_tuple ( std::move ( bar ) ) ) {
	setContent ( std::move ( data ) );
}

PrefixBarTree::PrefixBarTree ( alphabet::Symbol bar, std::vector < alphabet::Symbol > data ) : PrefixBarTree ( std::move ( bar ), std::set < alphabet::Symbol > ( data.begin ( ), data.end ( ) ) + std::set < alphabet::Symbol > { bar }, std::move ( data ) ) {
}

PrefixBarTree::PrefixBarTree ( alphabet::Symbol bar, const UnrankedTree & tree ) : PrefixBarTree ( bar, tree.getAlphabet ( ) + std::set < alphabet::Symbol > { bar }, toPrefixBar ( tree.getRoot ( ), bar ) ) {
}

PrefixBarTree::PrefixBarTree ( const UnrankedTree & tree ) : PrefixBarTree ( alphabet::BarSymbol::BAR_SYMBOL, tree ) {
}

std::vector < alphabet::Symbol > PrefixBarTree::toPrefixBar ( const UnrankedNode & node, const alphabet::Symbol & bar ) {
	std::vector < alphabet::Symbol > res;
	toPrefixBar ( node, bar, res );
	return res;
}

void PrefixBarTree::toPrefixBar ( const UnrankedNode & node, const alphabet::Symbol & bar, std::vector < alphabet::Symbol > & linearNotation ) {
	linearNotation.push_back ( node.getSymbol ( ) );

	for ( const std::smart_ptr < const UnrankedNode > & child : node.getChildren ( ) )
		toPrefixBar ( * child, bar, linearNotation );

	linearNotation.push_back ( bar );
}

UnrankedTreeBase * PrefixBarTree::clone ( ) const {
	return new PrefixBarTree ( * this );
}

UnrankedTreeBase * PrefixBarTree::plunder ( ) && {
	return new PrefixBarTree ( std::move ( * this ) );
}

const std::vector < alphabet::Symbol > & PrefixBarTree::getContent ( ) const {
	return this->m_Data;
}

void PrefixBarTree::setContent ( std::vector < alphabet::Symbol > data ) {
	arityChecksum ( data );

	std::set < alphabet::Symbol > minimalAlphabet ( data.begin ( ), data.end ( ) );
	std::set < alphabet::Symbol > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet ( ).begin ( ), getAlphabet ( ).end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw TreeException ( "Input symbols not in the alphabet." );

	this->m_Data = std::move ( data );
}

void PrefixBarTree::arityChecksum ( const std::vector < alphabet::Symbol > & data ) {
	int arityChecksumTypes = 0;

	for ( const alphabet::Symbol & symbol : data ) {
		if ( symbol == getBar ( ) )
			arityChecksumTypes -= 1;
		else
			arityChecksumTypes += 1;
	}

	if ( arityChecksumTypes != 0 ) throw TreeException ( "The string does not form a tree" );
}

bool PrefixBarTree::isEmpty ( ) const {
	return this->m_Data.size ( ) == 0;
}

int PrefixBarTree::compare ( const PrefixBarTree & other ) const {
	auto first = std::tie ( m_Data, getAlphabet ( ), getBar ( ) );
	auto second = std::tie ( other.m_Data, other.getAlphabet ( ), other.getBar ( ) );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void PrefixBarTree::operator >>( std::ostream & out ) const {
	out << "(PrefixBarTree ";

	for ( const alphabet::Symbol & symbol : this->m_Data )
		out << symbol;

	out << ")";
}

PrefixBarTree::operator std::string ( ) const {
	std::stringstream ss;
	ss << "\"";

	for ( const alphabet::Symbol & symbol : this->m_Data )
		ss << ( std::string ) symbol;

	ss << "\"";
	return std::move ( ss ).str ( );
}

const std::string PrefixBarTree::XML_TAG_NAME = "PrefixBarTree";

PrefixBarTree PrefixBarTree::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, PrefixBarTree::XML_TAG_NAME );
	alphabet::Symbol bar = TreeFromXMLParser::parseBar ( input );
	std::set < alphabet::Symbol > unrankedAlphabet = TreeFromXMLParser::parseUnrankedAlphabet ( input );
	std::vector < alphabet::Symbol > data = TreeFromXMLParser::parseUnrankedContent ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, PrefixBarTree::XML_TAG_NAME );

	return PrefixBarTree ( std::move ( bar ), std::move ( unrankedAlphabet ), std::move ( data ) );
}

void PrefixBarTree::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( PrefixBarTree::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeBar ( out, getBar ( ) );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet ( ) );
	TreeToXMLComposer::composeContent ( out, m_Data );
	out.emplace_back ( PrefixBarTree::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT );
}

} /* namespace tree */

namespace std {

template < >
bool tree::PrefixBarTree::Component < tree::PrefixBarTree, alphabet::Symbol, tree::GeneralAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	const std::vector < alphabet::Symbol > & content = static_cast < const tree::PrefixBarTree * > ( this )->getContent ( );

	return std::find ( content.begin ( ), content.end ( ), symbol ) != content.end ( );
}

template < >
bool tree::PrefixBarTree::Component < tree::PrefixBarTree, alphabet::Symbol, tree::GeneralAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void tree::PrefixBarTree::Component < tree::PrefixBarTree, alphabet::Symbol, tree::GeneralAlphabet >::valid ( const alphabet::Symbol & ) const {
}

template < >
bool tree::PrefixBarTree::Element < tree::PrefixBarTree, alphabet::Symbol, tree::BarSymbol >::available ( const alphabet::Symbol & symbol ) const {
	return static_cast < const tree::PrefixBarTree * > ( this )->accessComponent < tree::GeneralAlphabet > ( ).get ( ).count ( symbol );
}

template < >
void tree::PrefixBarTree::Element < tree::PrefixBarTree, alphabet::Symbol, tree::BarSymbol >::valid ( const alphabet::Symbol & ) const {
}

} /* namespace std */

namespace alib {

auto prefixBarTreeParserRegister = xmlApi < tree::Tree >::ParserRegister < tree::PrefixBarTree > ( );
auto prefixBarTreeParserRegister2 = xmlApi < tree::UnrankedTreeWrapper >::ParserRegister < tree::PrefixBarTree > ( );
auto prefixBarTreeParserRegister3 = xmlApi < alib::Object >::ParserRegister < tree::PrefixBarTree > ( );

auto PrefixBarTreeFromUnrankedTree = castApi::CastRegister < tree::PrefixBarTree, tree::UnrankedTree > ( );
auto PrefixBarTreeCastBinder = castApi::CastPoolStringBinder < tree::PrefixBarTree > ( tree::PrefixBarTree::XML_TAG_NAME );

} /* namespace alib */
