/*
 * UnrankedNonlinearPattern.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef UNRANKED_NONLINEAR_PATTERN_H_
#define UNRANKED_NONLINEAR_PATTERN_H_

#include <string>
#include <set>
#include <core/components.hpp>

#include "UnrankedNode.h"
#include "../UnrankedTreeBase.h"

namespace tree {

class UnrankedNode;
class UnrankedPattern;
class GeneralAlphabet;
class SubtreeWildcard;
class NonlinearAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a pattern of RegExpElement.
 */
class UnrankedNonlinearPattern : public UnrankedTreeBase, public std::Components < UnrankedNonlinearPattern, alphabet::Symbol, std::tuple < GeneralAlphabet, NonlinearAlphabet >, std::tuple < SubtreeWildcard > > {
	std::smart_ptr < UnrankedNode > pattern;

public:
	/**
	 * @copydoc UnrankedNode::clone() const
	 */
	virtual UnrankedTreeBase * clone ( ) const;

	/**
	 * @copydoc UnrankedNode::plunder() const
	 */
	virtual UnrankedTreeBase * plunder ( ) &&;

	explicit UnrankedNonlinearPattern ( alphabet::Symbol subtreeWildcard, std::set < alphabet::Symbol > nonlinearVariables, std::set < alphabet::Symbol > alphabet, UnrankedNode pattern );
	explicit UnrankedNonlinearPattern ( alphabet::Symbol subtreeWildcard, std::set < alphabet::Symbol > nonlinearVariables, UnrankedNode pattern );
	explicit UnrankedNonlinearPattern ( alphabet::Symbol subtreeWildcard, UnrankedNode pattern );

	/**
	 * Copy constructor.
	 * @param other pattern to copy
	 */
	UnrankedNonlinearPattern ( const UnrankedNonlinearPattern & other );
	UnrankedNonlinearPattern ( UnrankedNonlinearPattern && other ) noexcept;
	UnrankedNonlinearPattern & operator =( const UnrankedNonlinearPattern & other );
	UnrankedNonlinearPattern & operator =( UnrankedNonlinearPattern && other ) noexcept;
	~UnrankedNonlinearPattern ( ) noexcept;

	/**
	 * @return Root node of the regular expression pattern
	 */
	const UnrankedNode & getRoot ( ) const;

	const std::set < alphabet::Symbol > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < alphabet::Symbol > & symbols ) {
		accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	const alphabet::Symbol & getSubtreeWildcard ( ) const {
		return accessElement < SubtreeWildcard > ( ).get ( );
	}

	const std::set < alphabet::Symbol > & getNonlinearVariables ( ) const {
		return accessComponent < NonlinearAlphabet > ( ).get ( );
	}

	/**
	 * @return Root node of the regular expression pattern
	 */
	UnrankedNode & getRoot ( );

	/**
	 * Sets the root node of the regular expression pattern
	 * @param pattern root node to set
	 */
	void setTree ( UnrankedNode pattern );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const alphabet::Symbol & symbol );

	/**
	 * Prints XML representation of the pattern to the output stream.
	 * @param out output stream to which print the pattern
	 * @param pattern pattern to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const UnrankedNonlinearPattern & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	const static std::string XML_TAG_NAME;

	static UnrankedNonlinearPattern parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace tree */

#endif /* UNRANKED_NONLINEAR_PATTERN_H_ */
