/*
 * UnrankedPattern.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "UnrankedPattern.h"
#include "../TreeException.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../ranked/RankedPattern.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../common/TreeAuxiliary.h"
#include "../Tree.h"
#include "../UnrankedTreeWrapper.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace tree {

UnrankedPattern::UnrankedPattern ( alphabet::Symbol subtreeWildcard, std::set < alphabet::Symbol > alphabet, UnrankedNode pattern ) : std::Components < UnrankedPattern, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < SubtreeWildcard > > ( std::make_tuple ( std::move ( alphabet ) ), std::make_tuple ( std::move ( subtreeWildcard ) ) ), pattern ( nullptr ) {
	setTree ( std::move ( pattern ) );
}

UnrankedPattern::UnrankedPattern ( alphabet::Symbol subtreeWildcard, UnrankedNode pattern ) : UnrankedPattern ( subtreeWildcard, pattern.computeMinimalAlphabet() + std::set < alphabet::Symbol > { subtreeWildcard }, pattern ) {
}

UnrankedPattern::UnrankedPattern ( const RankedPattern & other ) : UnrankedPattern ( other.getSubtreeWildcard ( ).getSymbol ( ), TreeAuxiliary::unrankSymbols ( other.getAlphabet ( ) ), other.getRoot ( ).asUnranked ( ) ) {
}

UnrankedPattern::UnrankedPattern ( const UnrankedPattern & other ) : std::Components < UnrankedPattern, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < SubtreeWildcard > > ( std::make_tuple ( other.getAlphabet ( ) ), std::make_tuple ( other.getSubtreeWildcard ( ) ) ), pattern ( other.pattern ) {
	this->pattern->attachAlphabet ( & ( this->getAlphabet ( ) ) );
}

UnrankedPattern::UnrankedPattern ( UnrankedPattern && other ) noexcept : std::Components < UnrankedPattern, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < SubtreeWildcard > > ( std::make_tuple ( std::move ( other.accessComponent < GeneralAlphabet > ( ).get ( ) ) ), std::make_tuple ( std::move ( other.accessElement < SubtreeWildcard > ( ).get ( ) ) ) ), pattern ( std::move ( other.pattern ) ) {
	this->pattern->attachAlphabet ( & ( this->getAlphabet ( ) ) );
}

UnrankedTreeBase * UnrankedPattern::clone ( ) const {
	return new UnrankedPattern ( * this );
}

UnrankedTreeBase * UnrankedPattern::plunder ( ) && {
	return new UnrankedPattern ( std::move ( * this ) );
}

UnrankedPattern & UnrankedPattern::operator =( const UnrankedPattern & other ) {
	if ( this == & other )
		return * this;

	* this = UnrankedPattern ( other );

	return * this;
}

UnrankedPattern & UnrankedPattern::operator =( UnrankedPattern && other ) noexcept {
	std::swap ( this->pattern, other.pattern );
	std::swap ( accessComponent < GeneralAlphabet > ( ).get ( ), other.accessComponent < GeneralAlphabet > ( ).get ( ) );
	std::swap ( accessElement < SubtreeWildcard > ( ).get ( ), other.accessElement < SubtreeWildcard > ( ).get ( ) );

	return * this;
}

UnrankedPattern::~UnrankedPattern ( ) noexcept {
}

const UnrankedNode & UnrankedPattern::getRoot ( ) const {
	return * pattern;
}

UnrankedNode & UnrankedPattern::getRoot ( ) {
	return * pattern;
}

void UnrankedPattern::setTree ( UnrankedNode pattern ) {
	this->pattern = std::make_smart < UnrankedNode > ( std::move ( pattern ) );

	if ( !this->pattern->attachAlphabet ( & ( this->getAlphabet ( ) ) ) )
		throw TreeException ( "Input symbols not in the alphabet." );
}

void UnrankedPattern::operator >>( std::ostream & out ) const {
	out << "(UnrankedPattern " << * ( this->pattern ) << ")";
}

int UnrankedPattern::compare ( const UnrankedPattern & other ) const {
	auto first = std::tie ( * pattern, getAlphabet(), getSubtreeWildcard() );
	auto second = std::tie ( * other.pattern, other.getAlphabet(), other.getSubtreeWildcard() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void UnrankedPattern::nicePrint ( std::ostream & os ) const {
	pattern->nicePrint ( os );
}

UnrankedPattern::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

const std::string UnrankedPattern::XML_TAG_NAME = "UnrankedPattern";

UnrankedPattern UnrankedPattern::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, UnrankedPattern::XML_TAG_NAME );
	alphabet::Symbol subtreeWildcardSymbol = TreeFromXMLParser::parseSubtreeWildcardSymbol ( input );
	std::set < alphabet::Symbol > rankedAlphabet = TreeFromXMLParser::parseUnrankedAlphabet ( input );
	UnrankedNode root = alib::xmlApi < UnrankedNode >::parse ( input );
	UnrankedPattern tree ( std::move ( subtreeWildcardSymbol ), std::move ( rankedAlphabet ), std::move ( root ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, UnrankedPattern::XML_TAG_NAME );
	return tree;
}

void UnrankedPattern::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( UnrankedPattern::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeSubtreeWildcard ( out, getSubtreeWildcard ( ) );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet ( ) );
	alib::xmlApi < UnrankedNode >::compose ( out, * pattern );
	out.emplace_back ( UnrankedPattern::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT );
}

} /* namespace tree */

namespace std {

template < >
bool tree::UnrankedPattern::Component < tree::UnrankedPattern, alphabet::Symbol, tree::GeneralAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	const tree::UnrankedPattern * pattern = static_cast < const tree::UnrankedPattern * > ( this );

	return pattern->getRoot ( ).testSymbol ( symbol ) || pattern->accessElement < tree::SubtreeWildcard > ( ).get ( ) == symbol;
}

template < >
bool tree::UnrankedPattern::Component < tree::UnrankedPattern, alphabet::Symbol, tree::GeneralAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void tree::UnrankedPattern::Component < tree::UnrankedPattern, alphabet::Symbol, tree::GeneralAlphabet >::valid ( const alphabet::Symbol & ) const {
}

template < >
bool tree::UnrankedPattern::Element < tree::UnrankedPattern, alphabet::Symbol, tree::SubtreeWildcard >::available ( const alphabet::Symbol & symbol ) const {
	return static_cast < const tree::UnrankedPattern * > ( this )->accessComponent < tree::GeneralAlphabet > ( ).get ( ).count ( symbol );
}

template < >
void tree::UnrankedPattern::Element < tree::UnrankedPattern, alphabet::Symbol, tree::SubtreeWildcard >::valid ( const alphabet::Symbol & ) const {
}

} /* namespace std */

namespace alib {

auto unrankedPatternParserRegister = xmlApi < tree::Tree >::ParserRegister < tree::UnrankedPattern > ( );
auto unrankedPatternParserRegister2 = xmlApi < tree::UnrankedTreeWrapper >::ParserRegister < tree::UnrankedPattern > ( );
auto unrankedPatternParserRegister3 = xmlApi < alib::Object >::ParserRegister < tree::UnrankedPattern > ( );

auto UnrankedPatternFromRankedPattern = castApi::CastRegister < tree::UnrankedPattern, tree::RankedPattern > ( );
auto UnrankedPatternCastBinder = castApi::CastPoolStringBinder < tree::UnrankedPattern > ( tree::UnrankedPattern::XML_TAG_NAME );

} /* namespace alib */
