/*
 * UnrankedTree.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#ifndef UNRANKED_TREE_H_
#define UNRANKED_TREE_H_

#include <string>
#include <set>
#include <core/components.hpp>

#include "UnrankedNode.h"
#include "../UnrankedTreeBase.h"

#include "../../alphabet/RankedSymbol.h"

namespace tree {

class UnrankedNode;
class RankedTree;
class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
class UnrankedTree : public UnrankedTreeBase, public std::Components < UnrankedTree, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < > > {
	std::smart_ptr < UnrankedNode > tree;

	static std::set < alphabet::Symbol > unrankSymbols ( const std::set < alphabet::RankedSymbol > & alphabet );

public:
	/**
	 * @copydoc UnrankedNode::clone() const
	 */
	virtual UnrankedTreeBase * clone ( ) const;

	/**
	 * @copydoc UnrankedNode::plunder() const
	 */
	virtual UnrankedTreeBase * plunder ( ) &&;

	explicit UnrankedTree ( std::set < alphabet::Symbol > alphabet, UnrankedNode tree );
	explicit UnrankedTree ( UnrankedNode tree );
	explicit UnrankedTree ( const RankedTree & other );

	/**
	 * Copy constructor.
	 * @param other tree to copy
	 */
	UnrankedTree ( const UnrankedTree & other );
	UnrankedTree ( UnrankedTree && other ) noexcept;
	UnrankedTree & operator =( const UnrankedTree & other );
	UnrankedTree & operator =( UnrankedTree && other ) noexcept;
	~UnrankedTree ( ) noexcept;

	/**
	 * @return Root node of the regular expression tree
	 */
	const UnrankedNode & getRoot ( ) const;

	const std::set < alphabet::Symbol > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < alphabet::Symbol > & symbols ) {
		accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	/**
	 * @return Root node of the regular expression tree
	 */
	UnrankedNode & getRoot ( );

	/**
	 * Sets the root node of the regular expression tree
	 * @param tree root node to set
	 */
	void setTree ( UnrankedNode tree );

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const UnrankedTree & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	const static std::string XML_TAG_NAME;

	static UnrankedTree parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace tree */

#endif /* UNRANKED_TREE_H_ */
